#include "crrc_dialog.h"

#include <iostream>

/**
 *  Constants for unified dialog style.
 */
const int DLG_DEF_SPACE         = 10;
const int DLG_DEF_BUTTON_WIDTH  = 70;
const int DLG_DEF_BUTTON_HEIGHT = 25;
const int DLG_CHECK_W = 15;
const int DLG_CHECK_H = 15;


/** \brief A list of all currently existing CRRCDialogs.
 *
 *  This vector keeps a pointer to each instance of CRRCDialog.
 *  We need it to determine the dialog which shall receive
 *  a click after a button callback was activated. PLIB only
 *  provides a pointer to the actual button, but we need
 *  a pointer to the dialog which contains the button.
 */
std::vector<CRRCDialog*> CRRCDialog::instances;


// Prototypes for this file
static void CRRCDialogOKCallback(puObject* obj);
static void CRRCDialogCancelCallback(puObject* obj);
puObject *findDialog(puObject *obj);


/** \brief Construct the dialog frame and up to two buttons.
 *
 *  The created dialog will be at least big enough to hold
 *  two buttons and can be enlarged by providing the
 *  width and height arguments.
 *  \param width Dialog's width (must be greater than the space
 *               occupied by the buttons to take effect)
 *  \param height Dialog's height (must be greater than the space
 *                occupied by the buttons to take effect)
 *  \param style Can be CRRC_DIALOG_OK (dialog has an OK button),
 *               CRRC_DIALOG_CANCEL (dialog has a Cancel button)
 *               or both (bitwise OR'ed, which also
 *               is the default).
 */
CRRCDialog::CRRCDialog(int width, int height, int style)
    : puDialogBox(10, 10),
    butCancel(NULL), butOK(NULL)
{
  // calculate sizes and position
  int min_width, min_height;
  int window_width, window_height;

  // Calculate minimum sizes:
  // one button and two spaces high
  min_height = 2 * DLG_DEF_SPACE + DLG_DEF_BUTTON_HEIGHT;
  // two buttons and 3 spaces wide
  min_width = 2 * DLG_DEF_BUTTON_WIDTH + 3 * DLG_DEF_SPACE;

  width = (min_width > width) ? min_width : width;
  height = (min_height > height) ? min_height : height;

  // Place message box on screen: horiz. centered, 1/3 from top
  puGetWindowSize(&window_width, &window_height);
  setPosition(window_width/2 - width/2,
              window_height*2/3 - height/2);

  // Create the dialog's frame
  dlgFrame = new puFrame(0, 0, width, height);

  // Just make sure nothing weird happens if anyone calls this
  // ctor with an invalid style parameter --> fall back to default
  if (!(style & CRRC_DIALOG_OK) && !(style & CRRC_DIALOG_CANCEL))
  {
    style = CRRC_DIALOG_OK | CRRC_DIALOG_CANCEL;
  }

  // Create buttons if wanted
  if (style & CRRC_DIALOG_OK)
  {
    butOK = new puOneShot(0, DLG_DEF_SPACE,
                          DLG_DEF_BUTTON_WIDTH,
                          DLG_DEF_SPACE + DLG_DEF_BUTTON_HEIGHT);
    butOK->setCallback(CRRCDialogOKCallback);
    butOK->setLegend("OK");
    // makeReturnDefault lets the button look ugly in SMALL_BEVELLED...
    //butOK->makeReturnDefault(true);
  }
  if (style & CRRC_DIALOG_CANCEL)
  {
    butCancel = new puOneShot(0, DLG_DEF_SPACE,
                              DLG_DEF_BUTTON_WIDTH,
                              DLG_DEF_SPACE + DLG_DEF_BUTTON_HEIGHT);
    butCancel->setCallback(CRRCDialogCancelCallback);
    butCancel->setLegend("Cancel");
  }
  // Place the buttons.
  centerButtons();

  CRRCDialog::instances.push_back(this);
  reveal();
}


/** \brief Destroy the object.
 *
 *  Deallocates the object's dynamic memory and removes
 *  the dialog from the internal list of all dialog instances.
 */
CRRCDialog::~CRRCDialog()
{
  // Delete dialog from internal list of instances
  std::vector<CRRCDialog*>::iterator it = CRRCDialog::instances.begin();
  while(it != CRRCDialog::instances.end())
  {
    if (*it == this)
    {
      instances.erase(it);
      break;
    }
    it++;
  }

  puDeleteObject(butOK);
  puDeleteObject(butCancel);
}


/** \brief Set the dialog's size.
 *
 *  This method overloads the inherited puObject::setSize() method.
 *  In addition to resizing the base object it resizes the
 *  underlying puFrame.
 *  \param w  The dialog's new width.
 *  \param h  The dialog's new height.
 */
void CRRCDialog::setSize(int w, int h)
{
  puDialogBox::setSize(w, h);
  dlgFrame->setSize(w, h);
  centerButtons();
}


/** \brief Re-center the dialog's buttons.
 *
 *  If both buttons are activated, the following layout is applied:
 *  The "OK" button will be placed DLG_DEF_SPACE/2 pixels
 *  left of the dialog's horizontal center. The dialog's
 *  Cancel button will be placed DLG_DEF_SPACE/2 pixels
 *  right of the center.
 *
 *  If only one button is activated it will be centered.
 *
 *  In both cases, the buttons will be placed DLG_DEF_SPACE
 *  pixels from the dialog's bottom edge.
 */
void CRRCDialog::centerButtons()
{
  int w, h, ok_x, ok_y = 0, can_x, can_y = 0;
  int buttons = 0;

  puDialogBox::getSize(&w, &h);
  if (butOK != NULL)
  {
    buttons++;
    butOK->getPosition(&ok_x, &ok_y);
  }
  if (butCancel != NULL)
  {
    buttons++;
    butCancel->getPosition(&can_x, &can_y);
  }

  switch (buttons)
  {
    case 1:
      // only one button --> centered
      ok_x = can_x = (w - DLG_DEF_BUTTON_WIDTH) / 2;
      break;

    case 2:
    default:
      // two buttons
      ok_x  = (w - DLG_DEF_SPACE)/2 - DLG_DEF_BUTTON_WIDTH;
      can_x = (w + DLG_DEF_SPACE)/2;
      break;
  }

  if (butOK != NULL)
  {
    butOK->setPosition(ok_x, ok_y);
  }
  if (butCancel != NULL)
  {
    butCancel->setPosition(can_x, can_y);
  }
}


/** \brief Find the dialog which contains a specific button.
 *
 *  As the button's callback only gets a pointer to the
 *  button and not to the dialog itself, we have to search
 *  for the dialog in the list of instances.
 *
 *  This function is a friend to CRRCDialog as it needs
 *  access to the private list of instances.
 * \param obj Pointer to a dialog's OK or Cancel button.
 * \return Pointer to dialog or NULL if not found.
 */
puObject *findDialog(puObject *obj)
{
  std::vector<CRRCDialog*>::iterator it = CRRCDialog::instances.begin();
  puObject *dlg = NULL;

  // search for the dialog
  while((it != CRRCDialog::instances.end()) && (dlg == NULL))
  {
    if (((*it)->butOK == obj) || ((*it)->butCancel == obj))
    {
      dlg = *it;
    }
    it++;
  }
  return dlg;
}


/** \brief The generic OK callback.
 *
 *  This callback is invoked if the dialog's OK button is
 *  activated. It searches for the dialog the button belongs
 *  to, sets the dialog's value to CRRC_DIALOG_OK and
 *  invokes the dialog's callback.
 */
void CRRCDialogOKCallback(puObject* obj)
{
  puObject *dlg = findDialog(obj);
  // found dialog object?
  if (dlg != NULL)
  {
    dlg->setValue(CRRC_DIALOG_OK);
    dlg->invokeCallback();
  }
}


/** \brief The generic Cancel callback.
 *
 *  This callback is invoked if the dialog's OK button is
 *  activated. It searches for the dialog the button belongs
 *  to, sets the dialog's value to CRRC_DIALOG_OK and
 *  invokes the dialog's callback.
 */
void CRRCDialogCancelCallback(puObject* obj)
{
  puObject *dlg = findDialog(obj);
  // found dialog object?
  if (dlg != NULL)
  {
    dlg->setValue(CRRC_DIALOG_CANCEL);
    dlg->invokeCallback();
  }
}


/** \brief Set the OK button legend to a different text.
 *
 *  This method is useful if the default "OK"/"Cancel"
 *  button pair does not suit your needs, e.g. when you
 *  need "Yes"/"No" buttons.
 */
void CRRCDialog::setOKButtonLegend(const char *text)
{
  butOK->setLegend(text);
}


/** \brief Set the Cancel button legend to a different text.
 *
 *  This method is useful if the default "OK"/"Cancel"
 *  button pair does not suit your needs, e.g. when you
 *  need "Yes"/"No" buttons.
 */
void CRRCDialog::setCancelButtonLegend(const char *text)
{
  butCancel->setLegend(text);
}


/** \brief Get a pointer to the top-most dialog on screen.
 *
 *  This method returns a pointer to the top-most dialog
 *  on screen, or NULL if no dialog is visible.
 */
CRRCDialog* CRRCDialog::getToplevel()
{
  CRRCDialog *ret = NULL;

  if (instances.size() > 0)
  {
    ret = instances.back();
  }

  return ret;
}


/** \brief Determine if the dialog has an OK button.
 *
 *
 */
bool CRRCDialog::hasOKButton()
{
  bool ret = false;
  
  if (butOK != NULL)
  {
    ret = true;
  }
  
  return ret;
}

/** \brief Determine if the dialog has a Cancel button.
 *
 *
 */
bool CRRCDialog::hasCancelButton()
{
  bool ret = false;
  
  if (butCancel != NULL)
  {
    ret = true;
  }
  
  return ret;
}

/** \brief Lock the dialog.
 *
 *  Greys out the buttons so the user can't close the
 *  dialog.
 */
void CRRCDialog::lock()
{
  if (butOK != NULL)
  {
    butOK->greyOut();
  }
  if (butCancel != NULL)
  {
    butCancel->greyOut();
  }
}

/** \brief Unlock the dialog.
 *
 *  Reactivates the dialog's buttons.
 */
void CRRCDialog::unlock()
{
  if (butOK != NULL)
  {
    butOK->activate();
  }
  if (butCancel != NULL)
  {
    butCancel->activate();
  }
}
