/** \file SimStateHandler.cpp
 *
 *  This file includes the definition of the class
 *  SimStateHandler, which takes care of the current
 *  simulation state (running, paused, ...) and provides
 *  some handy time values.
 *
 *  \author J. Reucker
 */

#ifndef SIMSTATEHANDLER_H
#define SIMSTATEHANDLER_H

#include "crrc_inputs.h"
#include <SDL.h>

// Typedef section :

/// Function pointer to the "idle" function
typedef void (*TIdleFuncPtr)(TSimInputs*);

/// simulation state definitions
typedef enum
{
  STATE_RESUMING,     // the next iteration will be resuming from a pause
  STATE_PAUSED,       // the simulation is paused
  STATE_RUN,          // the simulation is running
  STATE_EXIT          // exit after next iteration
} T_SimState;  


/*****************************************************************************/
// Classes section :

/** \brief The simulation state handler
 *
 *  This class represents the simulation's state machine.
 *  It also provides some handy time values (since startup,
 *  in-game-time, ...).
 *  
 */
class SimStateHandler
{
  private:
    T_SimState    nState;         ///< state of the simulation
    TIdleFuncPtr  IdleFunc;       ///< the currently used "idle function"
    TIdleFuncPtr  OldIdleFunc;    ///< stores idle pointer for temporary switching
    unsigned long int sim_steps;  ///< number of simulation steps performed
    unsigned long int pause_time; ///< time when pause mode was entered
    unsigned long int accum_pause_time; ///< pause time since last reset
    unsigned long int reset_time; ///< time of the last reset
  
  public:
    /// The constructor
    SimStateHandler();
    
    /// The destructor
    ~SimStateHandler();
  
    /// pause the simulation
    void pause();
  
    /// resume from pause mode
    void resume();

    /// reset the simulation
    void reset();
  
    /// quit the simulation
    void quit();
  
    /// query the current state
    T_SimState  getState()  const {return nState;};
    
    /// set state machine to a new state
    void setState(T_SimState new_state) {nState = new_state;};
    
    /// run the current "idle" function
    void doIdle(TSimInputs* in);
    
    /// temporarily switch to a different idle function
    void setNewIdle(TIdleFuncPtr new_idle);
    
    /// fall back to previous idle function
    void resetIdle();

    /// increase number of simulation steps
    void incSimSteps(int multiloop) {sim_steps += multiloop;};
    
    /// get the total time since the sim was launched (in ms)
    unsigned long int getTotalTime() const {return SDL_GetTicks();};
    
    /// get the time since the last reset (including pause time, in ms)
    unsigned long int getTotalTimeSinceReset();
    
    /// get the time since the last reset (not counting pause time, in ms)
    unsigned long int getGameTimeSinceReset();

    /// get the simulation time since the last reset (number of sim steps * dt, in ms)
    unsigned long int getSimulationTimeSinceReset();
};

#endif  // SIMSTATEHANDLER_H

