#ifndef T_TX_INTERFACE_SERIAL
#define T_TX_INTERFACE_SERIAL

#include "tx_interface.h"
#include "SimpleXMLTransfer.h"
#include "chardevice/serialchardevice.h"
#include <string>

using namespace std;

//#define READ_BUFFER_SIZE  (256)   /* max. line length for config files  */
//#define SYNC_TIMEOUT      (500)   /* try to sync for approx. 5 seconds  */

/** \brief An interface using the serial port
 *
 * This is an abstract base class for interfaces which use a transmitter
 * connected to a serial port.
 *
 * There are multiple interfaces using the serial port, which leads to code
 * duplication. What's worse, some of them might not implement configuration or
 * support for a particular OS.
 *
 * So this base class can be used for all interfaces using the serial port. It
 * was written for the Zhen Hua interface. I suggest modifying all serial input
 * classes to inherit from this  class.
 *
 * Implementations of this class need to:
 *   - implement processDataByte and call setRawData from there
 *   - implement getInterfaceName (needed for storing settings)
 *   - override inputMethod to return the appropriate constant from T_TX_Interface::eIM_...
 * Implementations should:
 *   - override getNumAxes
 * Implementations may:
 *   - override the init method to do things like setting DTR and RTS after
 *     calling the original
 *
 * This class uses SerialCharDevice.
 */
class T_TX_InterfaceSerial: public T_TX_Interface
{
	public:
		// Construction
		T_TX_InterfaceSerial ();
		virtual ~T_TX_InterfaceSerial ();

		// Configuration
		virtual int init (SimpleXMLTransfer *config); // Initialize the interface, read configuration data
		virtual void putBackIntoCfg (SimpleXMLTransfer *config); // Write configuration back

		virtual int getDeviceNumber (const char *list[]);
		virtual void setDeviceName (std::string devname); // ttyS0, ttyUSB0, COM1...

		virtual int getDeviceSpeed ();
		virtual void setDeviceSpeed (int speed);

		// Data access
		virtual void getInputData (TSimInputs *inputs); // Called by the simulator to get the input data
		virtual void getRawData(float* target); // Needed for calibration

		// I/O
		void readSerialData ();
		virtual void setRawData (int num, float data);
		static char *baudRates[];

		// Interface
		virtual string getErrMsg () { return errorMessage; };

	protected:
		virtual std::string getInterfaceName ()=0;	// Implementations need to implement this
		virtual std::string getXmlChildName () { return "inputMethod."+getInterfaceName (); }
		virtual int getDefaultBaudRate () { return 19200; }
		virtual void processDataByte (unsigned char byte)=0; // Implementations need to implement this
		virtual void setErrMsg (string msg) { errorMessage=msg; }
		virtual int baudRateToSpeed (int baudRate);

		virtual void setDtr (bool dtr);
		virtual void setRts (bool rts);



	private:
		int getSerialByte ();
		SerialCharDevice *charDevice;
		string portName;
		unsigned int baudRate;
    
    static const int nNumBaudrates;
    static const int anBaudRates[];
		
    float rawData[TX_MAXAXIS];
		double calibrateValue (int function);

		string errorMessage;

#if DEBUG_TX_INTERFACE>0
#define DBGBUFFERSIZE (256)
		unsigned char dbgbuffer[DBGBUFFERSIZE];
		int dbgbufferidx;
    void debugHandler(unsigned char ucByte);
#endif
};

#endif	// TX_INTERFACE_SERIAL

