/**
 *  \file tx_audio.cpp
 *
 *  Definition of the AUDIO interface for CRRCsim.
 *
 *  This class will read signals from a PPM transmitter
 *  using a sound input device.
 */

#include "tx_audio.h"

#include <stdio.h>
#include <cstdlib>


static paAudioData  AudioData;
static int          synchro_index = 0;


// --------------------------------------------------------------------------
/** This routine will be called by the PortAudio engine when audio is needed.
 *  It may be called at interrupt level on some machines so don't do anything
 *  that could mess up the system like calling malloc() or free().
 *  Purpose of this routine is to update a rotating buffer
 *  index of buffer is AudioData.frameIndex
 *  buffer size is AudioData.frameSize
 *  number of samples stored since audio_rc_open() called
 */
#if PORTAUDIO > 0
#if PORTAUDIO == 18
static int recordCallback( void *inputBuffer, void *outputBuffer,
                           unsigned long framesPerBuffer,
                           PaTimestamp outTime, void *userData )
#else
static int recordCallback( const void *inputBuffer, void *outputBuffer,
                           unsigned long framesPerBuffer,
                           const PaStreamCallbackTimeInfo *outTime,
                           PaStreamCallbackFlags b, void *userData )
#endif
{
  paAudioData *data = (paAudioData*)userData;
  SAMPLE *rptr = (float*)inputBuffer;
  SAMPLE *wptr;
  unsigned long i;
  unsigned long framesLeft = data->frameSize - data->frameIndex;

  /* Prevent unused variable warnings. */
  outputBuffer = NULL;
  outTime = 0;

#if PORTAUDIO != 18
  b = 0;
#endif

  data->SamplesCounter += framesPerBuffer;

  wptr = &data->recordedSamples[data->frameIndex];
  if( framesPerBuffer <= framesLeft )
  {
    for( i=0; i<framesPerBuffer; i++ )
    {
      *wptr++ = *rptr++;
      rptr++;
    }  // assume NUM_CHANNELS = 2
  }
  else
  {
    for( i=0; i<framesLeft; i++ )
    {
      *wptr++=*rptr++;
      rptr++;
    } // assume NUM_CHANNELS = 2

    wptr = &data->recordedSamples[0];
    for( i=0; i<framesPerBuffer-framesLeft; i++ )
    {
      *wptr++=*rptr++;
      rptr++;
    }  // assume NUM_CHANNELS = 2
  }

  data->frameIndex=(data->frameIndex+framesPerBuffer)%data->frameSize;

  return 0;
}
#endif

// --------------------------------------------------------------------------

/**
 *  Create an audio interface.
 */
T_TX_InterfaceAudio::T_TX_InterfaceAudio()
  : cname("audio"), sc_num(0), sDevName("")
{
#if DEBUG_TX_INTERFACE > 0
  printf("T_TX_InterfaceAudio::T_TX_InterfaceAudio()\n");
#endif
}


/**
 *  Delete an audio interface and close the input device.
 *
 */
T_TX_InterfaceAudio::~T_TX_InterfaceAudio()
{
#if DEBUG_TX_INTERFACE > 0
  printf("T_TX_InterfaceAudio::~T_TX_InterfaceAudio()\n");
#endif
  audio_rc_close();
  delete mixer;
}


/**
 *  Initialize an audio interface from the configuration file
 *  and try to open the input device.
 *
 *  \param config Pointer to the global configuration structure
 */
int T_TX_InterfaceAudio::init(SimpleXMLTransfer* config)
{
#if DEBUG_TX_INTERFACE > 0
  printf("int T_TX_InterfaceAudio::init(SimpleXMLTransfer* config)\n");
#endif
  T_TX_InterfacePPM::init(config);

  mixer = new T_TX_Mixer(this, config, "inputMethod.audio");
  
  SimpleXMLTransfer *tag = config->getChild("inputMethod.audio", true);
  sc_num = tag->attributeAsInt("soundcard", 0);
  sDevName = tag->attribute("device_name", "NOT_SPECIFIED");
  maxInputCh = tag->attributeAsInt("max_in", -1);
  maxOutputCh = tag->attributeAsInt("max_out", -1);
  
  if (audio_rc_open() == 0)
    return(0);
  else
  {
    errMsg = "Error opening audio interface";
    return(-1);
  }
}


/**
 *  Get the current channel values from the interface.
 *
 *  \param inputs Pointer to a structure that stores the input values.
 */
void T_TX_InterfaceAudio::getInputData(TSimInputs* inputs)
{
#if DEBUG_TX_INTERFACE > 1
  printf("void T_TX_InterfaceAudio::getInputData(TSimInputs* inputs)\n");
#endif
  
  get_data_from_audio_interface(rc_channel_values);
  T_TX_InterfacePPM::getInputData(inputs);
}


/**
 *  Write all configuration data back into the config structure.
 *
 *  \param config Pointer to the global configuration structure.
 */
void T_TX_InterfaceAudio::putBackIntoCfg(SimpleXMLTransfer* config)
{
#if DEBUG_TX_INTERFACE > 0
  printf("int T_TX_InterfaceAudio::putBackIntoCfg(SimpleXMLTransfer* config)\n");
#endif

  SimpleXMLTransfer *tag = config->getChild("inputMethod.audio", true);
  tag->setAttributeOverwrite("soundcard", sc_num);
  tag->setAttributeOverwrite("device_name", sDevName);
  tag->setAttributeOverwrite("max_in", maxInputCh);
  tag->setAttributeOverwrite("max_out",maxOutputCh);

  T_TX_InterfacePPM::putBackIntoCfg(config);
  mixer->putBackIntoCfg(config);
}


/**
 *  Close the audio input device.
 *
 *  \retval  0  success
 *  \retval -1  failure
 */
int T_TX_InterfaceAudio::audio_rc_close()
{
  int iRetCode = 0;

#if PORTAUDIO > 0
  PaError    err = paNoError;

  if (stream != NULL)
  {
    err = Pa_CloseStream( stream );
  }
  if( err == paNoError )
  {
    if (AudioData.recordedSamples!=NULL)
      free( AudioData.recordedSamples );
  
    Pa_Terminate();
    
    printf("audio device closed\n");
  }
  else
  {
    Pa_Terminate();
    fprintf( stderr, "An error occured while using the portaudio stream in audio_rc_close\n" );
    fprintf( stderr, "Error number: %d\n", err );
    fprintf( stderr, "Error message: %s\n", Pa_GetErrorText( err ) );
  
    iRetCode = -1;
  }
#endif
  
  return iRetCode;
}


int T_TX_InterfaceAudio::get_data_from_audio_interface(float *values)
{
  static int nvals=0;

#if PORTAUDIO > 0
  static float vals[11];
  int i,n;
  
  /*
   * there is no need to update values more than 50 times a second
   * then lest check here every (SAMPLE_RATE/50) recorded samples
   */

  Pa_Sleep (1); // not sure this is usefull

  n=AudioData.SamplesCounter;
  if (n-AudioData.LastSamplesCounter>SAMPLE_RATE/50)
  {

    AudioData.LastSamplesCounter = n;                         // lets update values
    if(get_data(vals,&n))
    {
      for(i=0;i<11;i++)
        values[i]=vals[i];
      nvals = n;
    }
  }
  else
    /*
     * LastSamplesCounter & SamplesCounter saturation management
     */
    if (n<AudioData.LastSamplesCounter)
      AudioData.LastSamplesCounter = n;

#endif

  return nvals;
}


/**
 *  Open the audio input device.
 */
int T_TX_InterfaceAudio::audio_rc_open()
{
#if PORTAUDIO > 0
  int     i, numDevices;
  PaError err;
  const PaDeviceInfo *pdi;

  #if PORTAUDIO != 18
  PaStreamParameters inStreamParm;
  #endif

  err = Pa_Initialize();
  if( err != paNoError )
    goto error;

  /* List all available audio input and output devices */

  #if PORTAUDIO == 18
  numDevices = Pa_CountDevices();
  #else
  numDevices = Pa_GetDeviceCount();
  #endif
  printf(    "Number of total audio input/output devices : %d.\n", numDevices );
  for( i = 0; i < numDevices; i++ )
  {
    pdi = Pa_GetDeviceInfo( i );
    printf("Device(%d) name  \t: %s.\n", i, pdi->name);
    printf("\tMax Inputs   = %d\n", pdi->maxInputChannels  );
    printf("\tMax Outputs  = %d\n", pdi->maxOutputChannels  );
  }

  inputDevice = -1;   // mark as invalid
  
  /* Find the correct input device.
   *
   * The following strategy is applied:
   * (1a) If only an ID was specified, assume that we're reading an old
   *      config file and try to use the ID
   * (1b) Else, if the device specified in the config file is found at the same
   *      device ID, it will be used.
   * (2)  Else, if the device specified in the config file is found at a
   *      different ID, the ID is adjusted and the device will be used.
   * (3)  Else, try to use the default input device recommended by Portaudio.
   */
  
  /* (1) */
  if (sc_num < numDevices)
  {
    pdi = Pa_GetDeviceInfo(sc_num);
    
    if (sDevName == "NOT_SPECIFIED")
    {
      /* (1a) */
      // maybe old config file, no "device_name" attribute,
      // try to use the specified device ID
      printf("Name of audio device not specified in configuration file, using ID %d\n", sc_num);
      inputDevice = sc_num;
    }
    else
    {
      /* (1b) */
      if ((sDevName == pdi->name)
            &&
          (maxInputCh == pdi->maxInputChannels)
            &&
          (maxOutputCh == pdi->maxOutputChannels))
      {
        /* found device from configuration file at the same ID, use it */
        printf("Found audio device specified in configuration file (%s) at old ID (%d)\n", pdi->name, sc_num);
        inputDevice = sc_num;
      }
    }
  }
  
  /* (2) */
  if (inputDevice < 0)
  {
    // no match found yet, browse all devices
    for( i = 0; i < numDevices; i++ )
    {
      pdi = Pa_GetDeviceInfo( i );
      
      if ((sDevName == pdi->name)
            &&
          (maxInputCh == pdi->maxInputChannels)
            &&
          (maxOutputCh == pdi->maxOutputChannels))
      {
        /* found device from configuration file at a different ID, use it */
        printf("Audio device specified in configuration file (%s) has moved (ID %d --> ID %d)\n", pdi->name, sc_num, i);
        inputDevice = i;
        break;
      }
    }
  }
  
  /* (3) */
  if (inputDevice < 0)
  {
    // still no match found, use default input device
    #if PORTAUDIO == 18
    inputDevice = Pa_GetDefaultInputDeviceID();
    #else
    inputDevice = Pa_GetDefaultInputDevice();
    #endif
    fprintf(stderr, "Soundcard specified in configuration file (%s) not found,\n", sDevName.c_str());
    fprintf(stderr, "falling back to default (ID %d)\n", inputDevice);
  }
  
  /* now we should have a valid device ID, try to configure the device */
  if ((inputDevice >=0) && (inputDevice < numDevices))
  {
    pdi = Pa_GetDeviceInfo( inputDevice );
    printf("Using input device with ID %d (%s)\n", inputDevice, pdi->name );

    /*  check configuration */
    if (pdi->maxInputChannels < NUM_CHANNELS)
    {
      printf("no audio input device available, please configure your computer to support an input audio device.\n");
      Pa_Terminate();
      return (-1);
    }

    /*  configure device */

    AudioData.frameSize = NUM_SAMPLES; /* Record for a few samples. */
    AudioData.frameIndex = 0;
    AudioData.LastSamplesCounter = 0;
    AudioData.SamplesCounter = 0;

    AudioData.recordedSamples = (SAMPLE *) malloc( AudioData.frameSize * sizeof(SAMPLE) );
    if( AudioData.recordedSamples == NULL )
    {
      printf("Could not allocate record array.\n");
      exit(1);
    }
    for( i=0; i<AudioData.frameSize; i++ )
      AudioData.recordedSamples[i] = 0;

    #if PORTAUDIO != 18
    inStreamParm.device = inputDevice;
    inStreamParm.channelCount = NUM_CHANNELS;
    inStreamParm.hostApiSpecificStreamInfo = NULL;
    inStreamParm.sampleFormat = paFloat32;
    inStreamParm.suggestedLatency = pdi->defaultHighInputLatency;
    #endif

    /* Record some audio. -------------------------------------------- */

    #if PORTAUDIO == 18
    err = Pa_OpenStream(&stream,
                        inputDevice,
                        NUM_CHANNELS,  /* NUM_CHANNELS : stereo input */
                        paFloat32,
                        NULL,
                        paNoDevice,
                        0,
                        paFloat32,
                        NULL,
                        SAMPLE_RATE,
                        1024,          /* frames per buffer */
                        0,             /* number of buffers, if zero then use default minimum */
                        0, //paDitherOff,    /* flags */
                        recordCallback,
                        &AudioData );
    #else
    err = Pa_OpenStream(&stream,
                        &inStreamParm,
                        NULL,
                        SAMPLE_RATE,
                        1024,
                        0,
                        recordCallback,
                        &AudioData);
    #endif

    if( err != paNoError )
      goto error;

    err = Pa_StartStream( stream );
    if( err != paNoError )
      goto error;

    // save the interface device data for the configuration
    sc_num      = inputDevice;
    sDevName    = pdi->name;
    maxInputCh  = pdi->maxInputChannels;
    maxOutputCh = pdi->maxOutputChannels;
    
    return 0;
  }
  

error:
  Pa_Terminate();
  fprintf( stderr, "An error occured while using the portaudio stream in audio_rc_open\n" );
  fprintf( stderr, "Error number: %d\n", err );
  fprintf( stderr, "Error message: %s\n", Pa_GetErrorText( err ) );
  return -1;

#else  // PORTAUDIO == 0
  return -1;    // make it impossible to use this interface
#endif
}


/**
 * Process the recorded audio data and determine the channel values.
 *
 * \param values    Pointer to array that stores the channel values
 * \param nvalues   Number of channels detected will be stored here
 * \return          Always 1
 */
int T_TX_InterfaceAudio::get_data(float *values,int *nvalues)
{
  float x, px, max, min, moy;
  int nval = AudioData.frameSize;
  int i, j, chanel=0;
  float time, dt;
  float *sig=AudioData.recordedSamples;

  max= -100000;
  min=  100000;
  
  for(i = 0; i < nval; i++)
  {
    x = sig[i];
    if (x > max)
      max = x;
    if (x < min)
      min = x;
  }
  moy = (max + min) / 2;
  x = 0;
  chanel = -1;
  time = 0;
  j = (AudioData.frameIndex + 1024) % AudioData.frameSize;
  for(i = 0; i < nval; i++)
  {
    time++;
    if (time > 100)
    {                   // Synchro detection
      if (chanel < 0)
        chanel = 0; //start
      else if (chanel > 0)
        break; //end
    }
    px = x;
    x = sig[j++];
    if (j == AudioData.frameSize)
    {
      j = 0;
    }
    if ((x > moy) && (px < moy))
    {
      dt = (moy - px) / (x - px);
      if (chanel >= 0)
      {
        if (chanel == 0 && i < (nval / 2))
        {  
          synchro_index = j;  //for OSCILLO  synchro
        }
        if (chanel > 0)
        {
          values[chanel-1] = (time + dt) / 44.1 - 1.5;
        }
        chanel ++;
        if (chanel > 10)
          break;
      }
      time = -dt;
    }
  }
  *nvalues = chanel - 1;

  return 1;
}


/**
 * Switch to a new audio device.
 */
void T_TX_InterfaceAudio::setNewDevice(int iDevID, std::string sName, int iMaxIn, int iMaxOut)
{
  sc_num = iDevID;
  sDevName = sName;
  maxInputCh = iMaxIn;
  maxOutputCh = iMaxOut;
}


/**
 * \brief Re-open the audio interface
 *
 * Close the currently open audio-interface device, and open
 * the audio device specified in the internal variables
 */
void T_TX_InterfaceAudio::reopen()
{
  /// \todo Error handling?
  audio_rc_close();
  audio_rc_open();
}


/**
 * Return a single sample of the synchronised audio signal.
 *
 * This function provides an interface to the audio data for
 * external functions like the oscilloscope.
 *
 * \param  i  sample number
 * \return    sampled value at index i
 */
float get_audio_signal(int i)
{
  float *sig;
  int j;
  sig = AudioData.recordedSamples;
  j = synchro_index + i;
  if ( j >=  AudioData.frameSize)
  {
    j -=  AudioData.frameSize;
  }
  return sig[j];                     
}



