/**
 *  \file tx_audio.h
 *
 *  Declaration of the AUDIO interface for CRRCsim.
 *
 *  This class will read signals from a PPM transmitter
 *  using a sound input device.
 */
#ifndef TX_INTERFACE_AUDIO_H
#define TX_INTERFACE_AUDIO_H

#include <string>

#include "../tx_interfacePPM.h"
#include "../SimpleXMLTransfer.h"
#include <crrc_config.h>

#if PORTAUDIO > 0
#include <portaudio.h>
#endif

#define SAMPLE_RATE  (44100)
#define NUM_SAMPLES   (3000)
#define NUM_CHANNELS     (2)


typedef float SAMPLE;

typedef struct
{
  int     frameIndex;  /* Index into sample array. */
  int     frameSize;
  int     SamplesCounter;
  int     LastSamplesCounter;
  SAMPLE  *recordedSamples;
} paAudioData;


#if PORTAUDIO != 18
typedef void PortAudioStream;
#define PaStream PortAudioStream
#endif


float get_audio_signal(int i);


/** \brief The AUDIO interface
 *
 *  A PPM interface which acquires the Tx's signal through the
 *  sound card.
 */
class T_TX_InterfaceAudio : public T_TX_InterfacePPM
{
  public:
   T_TX_InterfaceAudio();
   virtual ~T_TX_InterfaceAudio();
   
   /**
    * Initialize interface. Read further config data from a file, if necessary.
    */
   int init(SimpleXMLTransfer* config);
   
   /**
    * Set current input data. If some value is not available, the value 
    * is not overwritten.
    */
   void getInputData(TSimInputs* inputs);
   
   /**
    * Write configuration back
    */
   virtual void putBackIntoCfg(SimpleXMLTransfer* config);

   /**
    * Get input method
    */
   int inputMethod() { return(T_TX_Interface::eIM_audio); };
   
   /** \brief Get the name of the interface for the config file
    *
    *  Returns the name of the interface. It should be a single
    *  word, only lower-case characters.
    */
   const char * getConfigName() {return cname;};
   
   void setNewDevice(int iDevID, std::string sName, int iMaxIn, int iMaxOut);

   /**
    * Re-open the audio device
    */
   void reopen();
   
  private:
    const char *cname;        ///< name in the config file
  
    // These members identify the input device. The ID "sc_num"
    // may change when new devices are added to a system (e.g.
    // by hotplugging USB devices, so additional information
    // should be stored in the config file to find the right
    // device after CRRCsim is restarted.
    int         sc_num;       ///< number of the sound card
    std::string sDevName;     ///< name of the device as reported by Portaudio
    int         maxInputCh;   ///< max number of input channels
    int         maxOutputCh;  ///< max number of output channels
    
    int         inputDevice;

    #if PORTAUDIO > 0
    PortAudioStream   *stream;
    #endif

    int   audio_rc_open();
    int   audio_rc_close();
    int   get_data_from_audio_interface(float *values);
    int   get_data(float *values, int *nvalues);
};

#endif

