#include <fstream>

#include "global.h"
#include "config.h"
#include "lib_conversions.h"

#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <errno.h>
#include "defines.h"
#include "zoom.h"
#include "crrc_main_cpp.h"
#include "crrc_scenery.h"
#include "windfield.h"

#ifdef linux
# include <stdlib.h>     // getenv()
# include <sys/stat.h>   // mkdir
# include <sys/types.h>  // mkdir
#endif

#if defined(__APPLE__) || defined(MACOSX)
# include <stdlib.h>     // getenv()
# include <sys/stat.h>   // mkdir
# include <sys/types.h>  // mkdir
#endif

#ifdef WIN32
# include <direct.h>   // mkdir
#endif

/* ******************************************************************************* */

// Pointer to the new configuration file. This should be passed
// to every module/object which wants to read something from the
// config (crrcsim.xml).
SimpleXMLTransfer* cfgfile = (SimpleXMLTransfer*)0;

// 
T_Config*       cfg;


/* ******************************************************************************* */

/**
 * Utility function which makes sure some path exists -- parts of it are created,
 * if needed.
 */
void makeSurePathExists(std::string path)
{
  if (path.length() == 0)
    return;
  
  unsigned int   pos0 = 0;
  
  if (path[pos0] == '/')
    pos0++;
  
  // last character should be '/'
  if (path[path.length()-1] != '/')
    path.append("/");
  
  while (pos0 < path.length())
  {    
    while (path[pos0] != '/')
      pos0++;

#if linux
    mkdir(path.substr(0, pos0).c_str(), 0766);
#endif
#if defined(__APPLE__) || defined(MACOSX) 
    mkdir(path.substr(0, pos0).c_str(), 0700);
#endif
#ifdef WIN32
    mkdir(path.substr(0, pos0).c_str());
#endif

    
    pos0++;
  }
}

// description: see header file
int options_saveToFile()
{
  options_saveToConfig();
  
  {
    std::ofstream outfile;
    std::string   path = T_Config::getConfigFilePath();

    if (path.find('/') != std::string::npos)
      makeSurePathExists(path.substr(0, path.rfind('/')));

    outfile.open(path.c_str());
    if (!outfile)
      return(-1);
    
    cfgfile->print(outfile);
    
    outfile.close();
  }
  return(0);
}

// description: see header file
int options_saveToConfig()
{
  // Some options need to be put back into the SimpleXMLTransfer
  zoom_putBackIntoCfg();
  write_globals_into_config();
  cfg->putBackIntoCfg(cfgfile);
  TXInterface->putBackIntoCfg(cfgfile);
  inputDev.putBackIntoCfg(cfgfile);

  return(0);
}

// description: see header file
bool options_changed()
{
  bool fEqual;
  
  options_saveToConfig();
  
  // Read config file
  SimpleXMLTransfer* cfgfileold = (SimpleXMLTransfer*)0;
  cfg->provideConfigAndPath(cfgfileold);
  
  fEqual = cfgfile->equalsOrdered(cfgfileold);

  delete cfgfileold;
  
  if (fEqual)
  {
    printf("options are unchanged\n");
    return(false);
  }
  else
  {
    printf("options have been changed\n");
    // cfgfile->print(std::cout); // test
    return(true);
  }
}

/* ******************************************************************************* */

T_Thermal::T_Thermal()
{
}

void T_Thermal::read(SimpleXMLTransfer* cfgfile,
                     T_Config*          cfg)
{
  SimpleXMLTransfer* logcfgptr = cfg->getCurLocCfgPtr(cfgfile);
  
  if (logcfgptr->indexOfChild("thermal") < 0)
    logcfgptr->addChild(GetDefaultConf_Thermal());
  
  // Read buffer data
  SimpleXMLTransfer* el = logcfgptr->getChild("thermal");  
  strength_mean   = el->getDouble("strength_mean");
  strength_sigma  = el->getDouble("strength_sigma");
  radius_mean     = el->getDouble("radius_mean");
  radius_sigma    = el->getDouble("radius_sigma");
  density         = el->getDouble("density");
  lifetime_mean   = el->getDouble("lifetime_mean");
  lifetime_sigma  = el->getDouble("lifetime_sigma");

  // show values:
  printf("Thermals: strength_mean=%f strength_sigma=%f radius_mean=%f radius_sigma=%f\n",
         strength_mean, strength_sigma,
         radius_mean, radius_sigma);
  printf("Thermals: density=%f lifetime_mean=%f lifetime_sigma=%f\n",
         density, 
         lifetime_mean, lifetime_sigma);  
}

int T_Thermal::putBackIntoCfg(SimpleXMLTransfer* cfgfile)
{
  // This is not necessary anymore as there is no internal data.
  return(0);
}

/* ******************************************************************************* */

void T_Wind::setDefaults()
{
  velocity  =   7;
  direction = 270;
}

void T_Wind::setVelocity(double val)
{
  velocity = val;  
}

void T_Wind::setDirection(double val, T_Config* cfg)
{
  // Sanity check, see description in header file
  if (cfg->wind == this)
  {  
    direction = val;
        
    cfg->checkDynamicSoaring();
  }
}

T_Wind::T_Wind()
{
  setDefaults();
}

/**
 *  Read the values for the current wind configuration from the
 *  main config file. If no entry exists for the current location,
 *  try to get a sensible default from the current location object
 *  itself. If currently no scenery is loaded, use a global default.
 *
 *  In any case, don't write anything back into the configuration or
 *  strange things might happen...
 */
void T_Wind::read(SimpleXMLTransfer* cfgfile, T_Config* cfg)
{
  // the global defaults
  float flDirection = 270;
  float flSpeed = 7;
  
  // try to override defaults with values from the current scenery object
  if (Global::scenery != NULL)
  {
    flDirection = Global::scenery->getDefaultWindDirection();
    flSpeed = Global::scenery->getDefaultWindSpeed();
    printf("Default wind from scenery: %.2f ft/s at %.2f deg\n", flSpeed, flDirection);
  }
  
  // now try to override these defaults with values from the config file
  try
  {
    SimpleXMLTransfer* el = cfg->getCurLocCfgPtr(cfgfile)->getChild("wind");
    flDirection = el->getDouble("direction");
    flSpeed     = el->getDouble("velocity");
  }
  catch (XMLException e)
  {
    // nothing in the config file...
  }
  setVelocity (flSpeed);
  setDirection(flDirection, cfg);
    
  // show values:
  printf("Wind: velocity=%f ft/sec   direction=%f\n",  velocity, direction);
}

int T_Wind::putBackIntoCfg(SimpleXMLTransfer* cfgfile)
{
  SimpleXMLTransfer* el = cfg->getCurLocCfgPtr(cfgfile)->getChild("wind", true);

  el->setAttributeOverwrite("velocity",  doubleToString(velocity));
  el->setAttributeOverwrite("direction", doubleToString(direction));

  return(0);
}

/* ******************************************************************************* */

/** static members */
std::string T_Config::ConfigFilePath;
std::string T_Config::soundfilePath;

T_Config::T_Config(SimpleXMLTransfer*& acfgfile)
{
  // Defaults
  nLocation       = Scenery::DAVIS;
  thermal         = new T_Thermal();
  wind            = new T_Wind();
  dynamic_soaring = FALSE;

  provideConfigAndPath(acfgfile);
}

void T_Config::provideConfigAndPath(SimpleXMLTransfer*& acfgfile)
{
  // determine path and open configuration file: depends on operating system
  acfgfile = (SimpleXMLTransfer*)0;
  
  if (ConfigFilePath.length() <= 0)
  {
#ifdef linux
  // configuration file
  // If there is a crrcsim.xml in the working directory, take this one. If not,
  // ~/.crrcsim/crrcsim.xml is the one to choose.
  {
    ConfigFilePath = "crrcsim.xml";
    try
    {
      acfgfile = new SimpleXMLTransfer(ConfigFilePath);
    }
    catch (XMLException e)
    {
      char* homepath = getenv("HOME");
      if (homepath != (char*)0)
      {
        ConfigFilePath  = homepath;
        ConfigFilePath += "/.crrcsim/crrcsim.xml";        
      }
    }
  }
#elif defined(__APPLE__) || defined(MACOSX)
  // configuration file
  // If there is a crrcsim.xml in the working directory, take this one. If not,
  // ~/.crrcsim/crrcsim.xml is the one to choose.
  {
    ConfigFilePath = "crrcsim.xml";
    try
    {
      acfgfile = new SimpleXMLTransfer(ConfigFilePath);
    }
    catch (XMLException e)
    {
      char* homepath = getenv("HOME");
      if (homepath != (char*)0)
      {
        ConfigFilePath  = homepath;
        ConfigFilePath += "/Library/Preferences/crrcsim.xml";        
      }
    }
  }
#elif defined(WIN32)
  // configuration file
  // If there is a crrcsim.xml in the working directory, take this one. If not,
  // ~/.crrcsim/crrcsim.xml is the one to choose.
  {
    ConfigFilePath = "crrcsim.xml";
    try
    {
      acfgfile = new SimpleXMLTransfer(ConfigFilePath);
    }
    catch (XMLException e)
    {
      char* homepath = getenv("USERPROFILE");
      if (homepath != (char*)0)
      {
        ConfigFilePath  = homepath;
        ConfigFilePath += "/.crrcsim/crrcsim.xml";        
      }
    }
  }
# else
  ConfigFilePath = "crrcsim.xml";
# endif
  }
  
  printf("Configuration file is %s\n", ConfigFilePath.c_str());
  
  if (acfgfile == (SimpleXMLTransfer*)0)
  {
    try
    {
      acfgfile = new SimpleXMLTransfer(ConfigFilePath);
    }
    catch (XMLException e)
    {
      // So there is no configuration file to load. Create default values.
      createDefaultConfig(acfgfile);
    }
  }
  
  
  // Take a look at the version number of the config file. Maybe
  // it is an old, incompatible one and there is something we need to
  // do about it?
  {
    int nVer = acfgfile->attributeAsInt("version", 1);

    printf("Configuration version is %i\n", nVer);
        
    if (nVer < 2)
    {
      // Coordinate system has changed (initial airplane position, wind 
      // direction).
      // Things which need to be changed:
      //   -wind presets
      //   -initial airplane position for every location
      //   -wind direction for every location
      // This is not impossible, but I am too lazy now, so I just go on as if there had been no config file.
      delete acfgfile;      
      acfgfile = (SimpleXMLTransfer*)0;
      createDefaultConfig(acfgfile);      
    }
  }    
}

void T_Config::createDefaultConfig(SimpleXMLTransfer*& acfgfile)
{
  acfgfile = new SimpleXMLTransfer();
  acfgfile->setName("crrcsimConfig");
  
  // fill with necessary data
  acfgfile->makeSureAttributeExists("version", "2");
  acfgfile->makeSureAttributeExists("video.skybox.texture_offset", 
                                    ftoStr(DEFAULT_SKYBOX_TEXTURE_OFFSET, 1, 5).c_str());
  
  SimpleXMLTransfer* tex;

  // things which are not really necessary, but useful:
  //   -launch presets  
  tex = acfgfile->getChild("launch.preset", true);
  tex->makeSureAttributeExists("name_en",      "hand");
  tex->makeSureAttributeExists("altitude",     "10");
  tex->makeSureAttributeExists("velocity_rel", "1");
  tex->makeSureAttributeExists("angle",        "0");
  
  tex = new SimpleXMLTransfer();
  tex->setName("preset");
  tex->addAttribute("name_en",      "winch");
  tex->addAttribute("altitude",     "300");
  tex->addAttribute("velocity_rel", "1");
  tex->addAttribute("angle",        "0");
  tex->addAttribute("sal",          "0");
  acfgfile->getChild("launch")->addChild(tex);
  
  tex = new SimpleXMLTransfer();
  tex->setName("preset");
  tex->addAttribute("name_en",      "throw");
  tex->addAttribute("altitude",     "10");
  tex->addAttribute("velocity_rel", "2");
  tex->addAttribute("angle",        "0");
  tex->addAttribute("sal",          "0");
  acfgfile->getChild("launch")->addChild(tex);
  
  tex = new SimpleXMLTransfer();
  tex->setName("preset");
  tex->addAttribute("name_en",      "hlg");
  tex->addAttribute("altitude",     "10");
  tex->addAttribute("velocity_rel", "5");
  tex->addAttribute("angle",        "0.38");
  tex->addAttribute("sal",          "0");
  acfgfile->getChild("launch")->addChild(tex);
  
  tex = new SimpleXMLTransfer();
  tex->setName("preset");
  tex->addAttribute("name_en",      "SAL");
  tex->addAttribute("altitude",     "5");
  tex->addAttribute("velocity_rel", "7");
  tex->addAttribute("angle",        "0.20");
  tex->addAttribute("sal",          "1");
  acfgfile->getChild("launch")->addChild(tex);
  
  
  //   -thermal presets
  tex = GetDefaultConf_Thermal();
  tex->addAttribute("name_en", "Default (v3)");
  acfgfile->getChild("presets.thermal", true)->addChild(tex);
  
  tex = new SimpleXMLTransfer();
  tex->setName("thermal");
  tex->addAttribute("name_en",      "F3F (no thermals)");
  tex->addAttribute("strength_mean",      "0");
  tex->addAttribute("strength_sigma",     "0");
  tex->addAttribute("radius_mean",        "0");
  tex->addAttribute("radius_sigma",       "0");
  tex->addAttribute("lifetime_mean",      "0");
  tex->addAttribute("lifetime_sigma",     "0");
  tex->addAttribute("density",            "0");
  acfgfile->getChild("presets.thermal", true)->addChild(tex);
  
  //   -wind presets
  //~ acfgfile->makeSureAttributeExists("presets.wind.wind.name_en",   "Dynamic Soaring Cape Cod");
  //~ acfgfile->makeSureAttributeExists("presets.wind.wind.velocity",  "13");
  //~ acfgfile->makeSureAttributeExists("presets.wind.wind.direction", "90");
  
  tex = new SimpleXMLTransfer();
  tex->setName("wind");
  tex->addAttribute("name_en",      "Dynamic soaring Cape Cod");
  tex->addAttribute("velocity",     "13");
  tex->addAttribute("direction",    "90");
  acfgfile->getChild("presets.wind", true)->addChild(tex);
  
  tex = new SimpleXMLTransfer();
  tex->setName("wind");
  tex->addAttribute("name_en",      "Soaring slow Cape Cod");
  tex->addAttribute("velocity",     "10");
  tex->addAttribute("direction",    "270");
  acfgfile->getChild("presets.wind")->addChild(tex);
  
  tex = new SimpleXMLTransfer();
  tex->setName("wind");
  tex->addAttribute("name_en",      "Soaring ruff Cape Cod");
  tex->addAttribute("velocity",     "18");
  tex->addAttribute("direction",    "270");
  acfgfile->getChild("presets.wind")->addChild(tex);
  
  tex = new SimpleXMLTransfer();
  tex->setName("wind");
  tex->addAttribute("name_en",      "F3F competition at Cape Cod");
  tex->addAttribute("velocity",     "30");
  tex->addAttribute("direction",    "270");
  acfgfile->getChild("presets.wind")->addChild(tex);
}

void T_Config::read(SimpleXMLTransfer* cfgfile)
{
  cfgfile->makeSureAttributeExists("location.name", "scenery/davis-orig.xml");
  setLocation(cfgfile->getString("location.name").c_str(), cfgfile, false);

  checkLocationBasedPar(cfgfile);
  
  thermal->read(cfgfile, this);
  wind->read(cfgfile, this);
}

/// Checks for CAPE_COD and dynamic soaring conditions
void T_Config::checkDynamicSoaring()
{
  // if we have already loaded a valid scenery, update the
  // internal location ID
  if (Global::scenery != NULL)
  {
    nLocation = Global::scenery->getID();
  }
  
  if (nLocation == Scenery::CAPE_COD)
  {
    if ((wind->getDirection() < 360) && (wind->getDirection() > 180))
    {
      if (wind->getDirection() != 270)
      {
        printf("Off normal winds not supported yet.\n");
        // This results in me calling myself once again -- but
        // the next time this 'if' isn't true, so this is no problem.
        wind->setDirection(270, cfg);
      }
      printf("Sloping.\n");
      dynamic_soaring = FALSE;
    } 
    else 
    {
      if (wind->getDirection() != 90)
      {
        printf("Off normal winds not supported yet.\n");
        // This results in me calling myself once again -- but
        // the next time this 'if' isn't true, so this is no problem.
        wind->setDirection(90, cfg);
      }
      printf("Dynamic soaring.\n");
      dynamic_soaring = TRUE;
    }
  }      
  else
    dynamic_soaring = FALSE;
}

void T_Config::setLocation(const char*         locstr,
                           SimpleXMLTransfer*  cfgfile)
{
  setLocation(locstr, cfgfile, true);
}

void T_Config::setLocation(const char*         locstr,
                           SimpleXMLTransfer*  cfgfile,
                           bool                fReadAgain)
{
  std::string tmp = trim(locstr);
  if (tmp.find("./") == 0)
    tmp = tmp.substr(2);

  printf("T_Config: Setting location to %s\n", tmp.c_str());
  scenery_filename  = tmp;

  // Set 
  cfgfile->setAttributeOverwrite("location.name", getLocationName());
  
  // reread config
  if (fReadAgain)
    read(cfgfile);
  
  // Location possibly changed, so this check is needed.
  // JR: this should be done from initialize_windfield()
  // checkDynamicSoaring();
}

const char* T_Config::getLocationName()
{
  return(scenery_filename.c_str());
}

int T_Config::putBackIntoCfg(SimpleXMLTransfer* cfgfile)
{
  thermal->putBackIntoCfg(cfgfile);
  wind->putBackIntoCfg(cfgfile);
  
  return(0);
}

SimpleXMLTransfer* T_Config::getCurLocCfgPtr(SimpleXMLTransfer* cfgfile)
{
  SimpleXMLTransfer* pgr;
  SimpleXMLTransfer* gr;
  
  // Now I can use the values specific to the chosen location
  std::string location = getLocationName();
  
  try
  {
    pgr = cfgfile->getChild("locations", true);
    int size = pgr->getChildCount();
    for (int m=0; m<size; m++)
    {
      gr = pgr->getChildAt(m);
      if (gr->getName().compare("location") == 0 && gr->attribute("name").compare(location) == 0)
        return(gr);
    }
    // So there is no entry for the current location. Let's create one:
    gr = new SimpleXMLTransfer();
    gr->setName("location");
    gr->addAttribute("name", location);
    pgr->addChild(gr);
    return(gr);
  }
  catch (XMLException e)
  {
  }
  
  return((SimpleXMLTransfer*)0);
}

void T_Config::checkLocationBasedPar(SimpleXMLTransfer* cfgfile)
{ 
  SimpleXMLTransfer* ptr;

  // make sure location entry is there
  ptr = getCurLocCfgPtr(cfgfile);

  // also check values read elsewhere
  /// \todo these defaults should already come from the scenery file!?!
  if (nLocation == Scenery::CAPE_COD)    
  {
    ptr->makeSureAttributeExists("position.airplane.x",  "-42");
    ptr->makeSureAttributeExists("position.airplane.y",  "10.5");
  }
  else
  {
    ptr->makeSureAttributeExists("position.airplane.x",  "-20.93");
    ptr->makeSureAttributeExists("position.airplane.y",  "0");
  }
}


/**
 *  Test if a file exists.
 *
 *  \param  path   File name and path to test.
 *  \retval true   if file exists
 *  \retval false  if file does not exist
 */
bool T_Config::fileExists(std::string path)
{
  std::ifstream testDat;
  
  testDat.open(path.c_str());
  
  if (!testDat) 
    return(false);
  else 
  { 
    testDat.close();
    return(true);
  }  
}


void T_Config::putConfigFilePath(std::string _ConfigFilePath)
{
  ConfigFilePath = _ConfigFilePath;
}


std::string T_Config::getConfigFilePath()
{
  return(ConfigFilePath);
}


/**
 *  Fill the given pathlist with paths that point to the directory dirname
 *
 *  This method creates a list of paths depending on the platform CRRCsim
 *  is compiled for. The directory dirname is appended to each of the
 *  paths.
 *  
 *  \param pathlist   reference to a vector of strings that will be filled with the paths
 *  \param dirname    directory name to be appended to each path.
 */
void T_Config::getSearchPathList(std::vector<std::string>& pathlist, std::string dirname)
{
  // Search in:
  // current working directory (.)        (all)
  // $HOME/Documents/crrcsim              (APPLE/MACOSX)
  // /Library/Application Support/crrcsim (APPLE/MACOSX)
  // $USERPROFILE/.crrcsim                (WIN32)
  // $HOME/.crrcsim                       (LINUX)
  // CRRC_DATA_PATH                       (LINUX)
  // /usr/local/share/games/crrcsim       (LINUX)
  // /usr/share/games/crrcsim             (LINUX)

  // cwd
  if (dirname != "")
  {
    pathlist.push_back(dirname);
  }
  else
  {
    pathlist.push_back(".");
  }
  
  // home directories
  std::string homepath_str = getHomePath();
  if (homepath_str != "")
  {
    if (dirname != "")
    {
      homepath_str += '/';
      homepath_str.append(dirname);
    }
    pathlist.push_back(homepath_str);
  }

  // System-wide installation paths (Linux and Mac OS X only)
  #if defined(linux)
  std::string data_path;
  #ifdef CRRC_DATA_PATH
  data_path = CRRC_DATA_PATH;
  #endif

  if (dirname != "")
  {
    std::string s;
    data_path += '/';
    #ifdef CRRC_DATA_PATH
    s = data_path;
    s.append(dirname);
    pathlist.push_back(s);
    #endif
    s = "/usr/local/share/games/crrcsim/";
    if (s != data_path)    // avoid adding this path twice
    {
      s.append(dirname);
      pathlist.push_back(s);
    }
    s = "/usr/share/games/crrcsim/";
    if (s != data_path)    // avoid adding this path twice
    {
      s.append(dirname);
      pathlist.push_back(s);
    }
  }
  else
  {
    #ifdef CRRC_DATA_PATH
    pathlist.push_back(data_path);
    #endif
    if (data_path != "/usr/local/share/games/crrcsim")  // avoid adding this path twice
    {
      pathlist.push_back("/usr/local/share/games/crrcsim");
    }
    if (data_path != "/usr/share/games/crrcsim")  // avoid adding this path twice
    {
      pathlist.push_back("/usr/share/games/crrcsim");
    }
  }
  #endif
  #if defined(__APPLE__) || defined(MACOSX)
  {
    std::string s = "/Library/Application Support/crrcsim";
    if (dirname != "")
    {
      s.append("/");
      s.append(dirname);
    }
    pathlist.push_back(s);
  }
  #endif
  #if 0
  std::cout << "T_Config::getSearchPathList():" << std::endl;
  for (std::vector<std::string>::size_type i = 0; i < pathlist.size(); i++)
  {
    std::cout << pathlist[i] << std::endl;
  }
  #endif
}


/** returns the platform-dependend crrcsim directory
 *  in the user's home directory
 */
std::string T_Config::getHomePath()
{
  std::string homepath_str = "";
  char        *homepath = NULL;
  
#ifdef WIN32
  homepath = getenv("USERPROFILE");
#else
  homepath = getenv("HOME");
#endif

  if (homepath != NULL)
  {
    homepath_str  = homepath;
    #if defined(__APPLE__) || defined(MACOSX)
    homepath_str.append("/Documents/crrcsim");
    #else
    homepath_str.append("/.crrcsim");
    #endif
  }
  
  return homepath_str;
}


/**
 *  Get the full path to the given data item
 *
 *  Pass the filename and relative path to this function and it
 *  will search for a suitable file in the search path. The
 *  absolute path to this file, including the filename,
 *  will be returned.
 *
 *  \param item       data item to search for (filename and relative path)
 *  \param fThrowEx   if set to true, an exception will be thrown if no
 *                    matching file was found
 *
 *  \return absolute path to the file, empty string on error
 */
std::string T_Config::getDataPath(std::string item, bool fThrowEx)
{
  std::string path = "";
  
  std::vector<std::string> possible_paths;
  T_Config::getSearchPathList(possible_paths, item);

  //~ std::cout << "T_Config::getDataPath(): " << std::endl;
  //~ for (std::vector<std::string>::size_type i = 0; i < possible_paths.size(); i++)
  //~ {
    //~ std::cout << possible_paths[i] << std::endl;
  //~ }

  for (std::vector<std::string>::size_type i = 0; i < possible_paths.size(); i++)
  {
    if (T_Config::fileExists(possible_paths[i]))
    {
      path = possible_paths[i];
      break;
    }
  }
  if ((path == "") && (fThrowEx))
  {
    throw XMLException("Unable to find a matching file: " + item);
  }

  return(path);
}


void T_Config::getModelDirs(std::vector<std::string>& dirlist)
{
  T_Config::getSearchPathList(dirlist, "models");
}


/**
 *  Create a list of F3F sound directories
 *
 *  F3F sound directories are all directories that reside in "sounds/f3f"
 *  somewhere in the search path and contain .wav files.
 *
 *  \param dirlist  This list of strings will be set to the retrieved paths.
 */
void T_Config::getF3FSoundDirs(std::vector<std::string>& dirlist)
{
  std::vector<std::string> paths;
  T_Config::getSearchPathList(paths, "sounds/f3f");

  // search the list of paths for subdirectories possibly containing
  // F3F wav files
  for (std::vector<std::string>::size_type i = 0; i < paths.size(); i++)
  {
    DIR *dir;
    if ((dir = opendir(paths[i].c_str())) == NULL)
    {
      std::string s = "T_Config::getF3FSoundDirs(): unable to open ";
      s += paths[i];
      perror(s.c_str());
    }
    else
    {
      struct dirent *ent;
      while ((ent = readdir(dir)) != NULL)
      {
        // test if this entity is really a directory
        std::string entity_fullpath = paths[i];
        entity_fullpath += "/";
        entity_fullpath += ent->d_name;
        
        DIR *subdir = opendir(entity_fullpath.c_str());
        if (subdir != NULL)
        {
          // Yes, this is a directory. We may immediately close it again
          // and see if we can add it to the list.
          closedir(subdir);
          
          // exclude some special directories
          if  ( (strcmp(ent->d_name, ".") != 0)       // exclude current dir
                  && 
                (strcmp(ent->d_name, "..") != 0)      // exclude parent dir
                  && 
                (strcmp(ent->d_name, "CVS") != 0)     // exclude CVS dir (mostly useless)
              )
          {
            // We've found a sound directory! Add it to the list.
            /// \todo Maybe we should check if this directory contains at least
            ///       one .wav file?
            dirlist.push_back(entity_fullpath);
          }
        }
        else
        {
          std::string s = "T_Config::getF3FSoundDirs(): ";
          s += entity_fullpath;
          perror(s.c_str());
        }
      }
      closedir(dir);
    }
  }
  
  std::cout << "F3F sound directories:" << std::endl;
  for (std::vector<std::string>::size_type i = 0; i < dirlist.size(); i++)
  {
    std::cout << "  " << dirlist[i] << std::endl;
  }
}

