/** \file crrc_animation.h
 *
 *  Classes for animated 3D model parts.
 */

#include "crrc_animation.h"

// --- non-class-functions ------------------------------------------

/**
 *  \brief Locate a named SSG node in a branch.
 *
 *  This method locates a named node in an SSG scenegraph branch. The
 *  node name usually corresponds to the name of an object in the 3D model.
 *
 *  This method was taken from SimGear, credits go to the original
 *  authors.
 *
 *  \param  node    Pointer to the scenegraph branch
 *  \param  name    Name of the object to search for
 *  \return Pointer to the node or NULL if no node with this name was found
 */
static ssgEntity* find_named_node (ssgEntity* node, const char* name)
{
  char *node_name = node->getName();
  if (node_name != 0 && !strcmp(name, node_name))
  {
    return node;
  }
  else if (node->isAKindOf(ssgTypeBranch()))
  {
    int nKids = node->getNumKids();
    for (int i = 0; i < nKids; i++)
    {
      ssgEntity * result =
        find_named_node(((ssgBranch*)node)->getKid(i), name);
      if (result != 0)
      {
        return result;
      }
    }
  } 
  return NULL;
}


/**
 *  Splice a branch in between all child nodes and their parents.
 *
 *  Inserts the given branch in the scenegraph above the
 *  entity \c child.
 *
 *  This method was taken from SimGear, credits go to the original
 *  authors.
 *
 *  \param  branch    the branch to be inserted
 *  \param  child     the location where the branch shall be inserted
 */
static void splice_branch (ssgBranch * branch, ssgEntity * child)
{
  int nParents = child->getNumParents();
  branch->addKid(child);
  for (int i = 0; i < nParents; i++)
  {
    ssgBranch * parent = child->getParent(i);
    parent->replaceKid(child, branch);
  }
}


/**
 *  Callback to update an animation.
 *
 *  This callback is activated whenever an animation node is
 *  traversed while rendering the scenegraph. It simply calls
 *  the animation's \c update() method which should know how
 *  to update itself.
 *
 *  This method was taken from SimGear, credits go to the original
 *  authors.
 *
 *  \param entity   the entity for which the callback was activated
 *  \param mask     dunno
 *  \return 1 if the animation was updated
 */
static int animation_callback (ssgEntity * entity, int mask)
{
    return ((CRRCAnimation *)entity->getUserData())->update();
}


/**
 * Callback to restore the state after an animation.
 */
//~ static int
//~ restore_callback (ssgEntity * entity, int mask)
//~ {
    //~ ((SGAnimation *)entity->getUserData())->restore();
    //~ return 1;
//~ }


/** \brief Add animations to a model
 *
 *  This method reads animation description tags from a model file
 *  and tries to add the corresponding animations to the 3D model.
 *
 *  \todo Right now there's only one type of animation: movable control
 *  surfaces. Therefore this method receives a pointer to the control
 *  input class. If animations are added that need a different kind of
 *  input for their update() method, we need to decide how to pass all
 *  this stuff to initAnimations().
 *
 *  \param  model_file    XML model description file
 *  \param  model         scenegraph of the 3D model
 *  \param  fInputs       pointer to the control input class
 */
void initAnimations(SimpleXMLTransfer *model_file, ssgEntity* model, TSimInputs *fInput)
{
  SimpleXMLTransfer *animations = model_file->getChild("animations", true);
  int num_anims = animations->getChildCount();
  fprintf(stdout, "initAnimations: found %d children\n", num_anims);
  
  for (int i = 0; i < num_anims; i++)
  {
    SimpleXMLTransfer *animation = animations->getChildAt(i);
    ssgEntity *node;
    
    if (animation->getName() != "animation")
    {
      fprintf(stderr, "initAnimations: invalid child <%s>\n", animation->getName().c_str());
    }
    else
    {
      std::string node_name = animation->getString("object.name", "default");
      std::string type      = animation->getString("type", "default");

      node = find_named_node(model, node_name.c_str());
      if (node != NULL)
      {
        CRRCAnimation *anim = NULL;
        printf("initAnimations: found animation node %s, type %s\n", 
                node_name.c_str(), type.c_str());
        
        if (type == "ControlSurface")
        {
          anim = new CRRCControlSurfaceAnimation(animation, fInput);
        }
        else
        {
          fprintf(stderr, "initAnimations: unknown animation type '%s'\n", type.c_str());
        }
        
        if (anim->getBranch() == NULL)
        {
          fprintf(stderr, "initAnimations: defunct animation class (animation branch is <NULL>)\n");
          exit(0);
        }
        else if (anim != NULL)
        {
          splice_branch(anim->getBranch(), node);
          anim->init();
          anim->setName("Animation");
          anim->getBranch()->setUserData(anim);
          anim->getBranch()->setTravCallback(SSG_CALLBACK_PRETRAV, animation_callback);

        }
        
      }
      else
      {
        fprintf(stderr, "initAnimations: node '%s' not found in 3D model\n", node_name.c_str());
      }
    }
  }
}



// --- Base class -----------------------------------------

/**
 *  Create a CRRCAnimation object.
 *
 */
CRRCAnimation::CRRCAnimation(ssgBranch *branch)
  : anim_branch(branch)
{
  init();
  branch->recalcBSphere();
}


/**
 *  Delete a CRRCAnimation object.
 *
 */
CRRCAnimation::~CRRCAnimation()
{
}


/**
 *  Initialize a CRRCAnimation object
 */
void CRRCAnimation::init()
{
}




// --- Control surface animation ------------------------------------

/**
 *  Create a CRRCControlSurfaceAnimation object
 *
 *  Initialize the animation from an 
 *  <animation type="ControlSurface"> tag
 */
CRRCControlSurfaceAnimation::CRRCControlSurfaceAnimation(SimpleXMLTransfer *xml, TSimInputs *inputs)
 : CRRCAnimation(new ssgTransform()), fallback_data(0.0f)
{
  bool failed = false;
  
  // evaluate <object> tag
  SimpleXMLTransfer *map = xml->getChild("object", true);
  max_angle = (float)(map->getDouble("max_angle", 0.0) * SG_RADIANS_TO_DEGREES);
  abs_max_angle = fabs(max_angle);

  // find hinges and evaluate all <control> tags
  int num_controls = 0;
  int num_hinges = 0;
  for (int i = 0; i < xml->getChildCount(); i++)
  {
    SimpleXMLTransfer *child = xml->getChildAt(i);
    if (child->getName() == "hinge")
    {
      // found a <hinge> child
      sgVec3 pos;
      //~ pos[SG_X] = child->getDouble("x", 0.0f);
      //~ pos[SG_Y] = child->getDouble("y", 0.0f);
      //~ pos[SG_Z] = child->getDouble("z", 0.0f);
      pos[SG_X] = -1 * child->getDouble("y", 0.0f);
      pos[SG_Y] = -1 * child->getDouble("x", 0.0f);
      pos[SG_Z] = -1 * child->getDouble("z", 0.0f);
      if (num_hinges == 0)
      {
        sgCopyVec3(hinge_1, pos);
      }
      else if (num_hinges == 1)
      {
        sgCopyVec3(hinge_2, pos);
      }
      num_hinges++;
    }
    else if (child->getName() == "control")
    {
      // found a <control> child
      // The "*2" factor for each gain value scales the control input
      // values from -0.5...+0.5 to -1.0...+1.0. This saves one
      // float multiplication per mapping in the runtime update() routine.
      std::string mapping = child->getString("mapping", "NOTHING");
      float gain = (float)child->getDouble("gain", 1.0);
      if (mapping == "ELEVATOR")
      {
        datasource.push_back(&(inputs->elevator));
        source_gain.push_back(gain * 2);
        num_controls++;
      }
      else if (mapping == "AILERON")
      {
        datasource.push_back(&(inputs->aileron));
        source_gain.push_back(gain * 2);
        num_controls++;
      }
      else if (mapping == "THROTTLE")
      {
        datasource.push_back(&(inputs->throttle));
        source_gain.push_back(gain * 2);
        num_controls++;
      }
      else if (mapping == "RUDDER")
      {
        datasource.push_back(&(inputs->rudder));
        source_gain.push_back(gain * 2);
        num_controls++;
      }
      else
      {
        fprintf(stderr, "ControlSurfaceAnimation: ignoring <control> tag without mapping.\n");
      }
      
    }
  }

  if (num_controls < 1)
  {
    fprintf(stderr, "ControlSurfaceAnimation: found animation without proper <control> tag. Animation disabled.\n");
    failed = true;
  }

  if (num_hinges < 2)
  {
    fprintf(stderr, "ControlSurfaceAnimation: Must specify exactly two hinges!\n");
    failed = true;
  }
  else
  {
    if (num_hinges > 2)
    {
      fprintf(stderr, "ControlSurfaceAnimation: Must specify exactly two hinges!\n");
      fprintf(stderr, "ControlSurfaceAnimation: Ignoring excessive hinge tag(s).\n");
    }
    sgSubVec3(axis, hinge_2, hinge_1);
    if (sgLengthVec3(axis) < 0.001)
    {
      fprintf(stderr, "ControlSurfaceAnimation: Insufficient spacing between hinges!\n");
      failed = true;
    }
  }

  if (failed)
  {
    fprintf(stderr, "ControlSurfaceAnimation: Animation setup failed.\n");
    // set to some non-critical defaults
    datasource.resize(1);
    datasource[0] = &fallback_data;
    source_gain.resize(1);
    source_gain[0] = 1.0;
    sgSetVec3(hinge_1, 0.0f, 0.0f, 0.0f);
    sgSetVec3(hinge_2, 1.0f, 0.0f, 0.0f);
    sgSubVec3(axis, hinge_2, hinge_1);
  }
  
  sgMakeIdentMat4(move_to_origin);
  move_to_origin[3][0] = -hinge_1[0];
  move_to_origin[3][1] = -hinge_1[1];
  move_to_origin[3][2] = -hinge_1[2];

  sgMakeTransMat4(move_back, hinge_1);

  init();
}


/**
 *  Delete a CRRCControlSurfaceAnimation object
 */
CRRCControlSurfaceAnimation::~CRRCControlSurfaceAnimation()
{
}


/**
 *  Initialize a CRRCControlSurfaceAnimation object
 */
void CRRCControlSurfaceAnimation::init()
{
  sgQuat q;
  sgMat4 m;
  sgMat4 qmat;
  sgAngleAxisToQuat(q, 0.0, axis);
  sgQuatToMatrix(qmat, q);

  ((ssgTransform*)anim_branch)->setTransform(m);
}


/**
 *  Update the animation. This method will be called by the
 *  scenegraph rendering function just before the animated
 *  branch will be rendered.
 */
int CRRCControlSurfaceAnimation::update()
{
  // calculate the deflection angle by summing up the contribution
  // of all mapped control inputs times the individual gain
  float angle = 0.0f;
  for (int i = 0; i < (int)datasource.size(); i++)
  {
    angle += *datasource[i] * source_gain[i] * max_angle;
  }
  // limit to max_angle (caution: max_angle could be < 0.0, so we use the abs value!)
  if (angle > abs_max_angle)
  {
    angle = abs_max_angle;
  }
  else if (angle < -abs_max_angle)
  {
    angle = -abs_max_angle;
  }

  // calculate transformation matrix for surface rotation
  sgQuat q;
  sgMat4 m;
  sgMat4 qmat;
  sgAngleAxisToQuat(q, angle, axis);
  sgQuatToMatrix(qmat, q);

  sgMultMat4(m, qmat, move_to_origin);
  sgPostMultMat4(m, move_back);
  
  ((ssgTransform*)anim_branch)->setTransform(m);
  return 1;
}
