/*****************************************************************************
Title: CRRCsim, the Charles River Radio Control Club Flight Simulator Project
Authors:
Jan Kansky:  Programming
Mark Drela:  Aerodynamics
Jan Reucker
Jens Wilhelm Wulf
 
Purpose:  The idea for CRRCsim is to take away any last excuse you might
have for not using a flight simulator to keep your thumbs certified for RC
flying.   It's free, it works, enjoy.  This is an open source project, so if
you don't like the way something works, help us fix it!
 
Thanks:
  Bruce Jackson for the LaRCsim framework.
  Flight Gear project for the sky sphere.
 
Contacts:
If you'd like to help with CRRCSIM, then send me an email!
    email                : kansky@ll.mit.edu
 
File:   crrc_loadair.h
        Routines dealing with .air file loading
 
*****************************************************************************/

#ifndef CRRC_LOADAIR_H
# define CRRC_LOADAIR_H

# include <string>
# include <vector>
# include <stdexcept>

# include <plib/ssg.h>

# include "global.h"
# include "include_gl.h"
# include "fdm.h"
# include "crrc_sound.h"
# include "SimpleXMLTransfer.h"



/** \brief A very basic airplane description class.
 *
 *  This is a first attempt to organize all airplane-related
 *  stuff into a class. At this time it's nothing more than
 *  a wrapper for the existing functions.
 * 
 *  Update: this class should contain everything but the simulation of
 *  the EOM (all the physics stuff). So this is about visualization.
 */
class CRRCAirplane
{
  public:
   
   /** \brief The constructor.
    *
    *  
    */
   CRRCAirplane();
   
   /** \brief The destructor.
    *
    *  Clean up.
    */
   virtual ~CRRCAirplane();
   
   /** \brief Draw the precompiled airplane shadow.
    *
    *  Applies some material properties and calls the
    *  shadow's display list.
    */
   virtual void draw_shadow(CRRC_AirplaneSim* airplane,
                            float             shadow_matrix[4][4]) = 0;
   
   /** \brief Draw the airplane
    *
    *
    */
   virtual void draw(CRRC_AirplaneSim* airplane) = 0;

   virtual int  getNumSounds()  {return (sound.size());};
   //~ virtual int  getSoundType(int soundnum)  {return sound[soundnum]->type;};
   //~ virtual double getSoundPitchFactor(int soundnum) { return(sound[soundnum]->dPitchFactor); };
  
  protected:
   std::vector<T_AirplaneSound*> sound;
};


/**
 * Reads geometric data directly from old-style .air-files.
 */
class CRRCAirplaneLaRCSim : public CRRCAirplane
{
  public:
   CRRCAirplaneLaRCSim();
   CRRCAirplaneLaRCSim(SimpleXMLTransfer* xml);
   virtual ~CRRCAirplaneLaRCSim();

  protected:
   
  /** \brief Initialize the airplane's sound.
    *
    *  Reads all sound related parameters from an xml description.
    *  \todo Make this method search all possible paths on Linux!
    */
   virtual void  initSound(SimpleXMLTransfer* xml);

   float max_thrust;
  
};


/** \brief Display airplane model using SSG
 *
 *  This airplane class also uses old-style .air-files,
 *  but does not use any of the geometry descriptions in
 *  the file. Instead it loads the model specified by
 *  the "model" parameter in the file.
 */
class CRRCAirplaneLaRCSimSSG : public CRRCAirplaneLaRCSim
{
  public:
    CRRCAirplaneLaRCSimSSG(const char*        filename, ssgBranch* graph);
    CRRCAirplaneLaRCSimSSG(SimpleXMLTransfer* xml,      ssgBranch* graph);
   
    virtual ~CRRCAirplaneLaRCSimSSG();
    
    virtual void draw(CRRC_AirplaneSim* airplane);
    virtual void draw_shadow(CRRC_AirplaneSim* airplane,
                            float             shadow_matrix[4][4]);

  
  private:
    ssgTransform  *initial_trans;
    ssgTransform  *model_trans;
    ssgEntity     *model;
    ssgEntity     *shadow;
    ssgTransform  *shadow_trans;
};


#endif // CRRC_LOADAIR_H
