/*****************************************************************************
Title: CRRCsim, the Charles River Radio Control Club Flight Simulator Project
Authors:
Jan Kansky:  Programming
Mark Drela:  Aerodynamics

Purpose:  The idea for CRRCsim is to take away any last excuse you might
have for not using a flight simulator to keep your thumbs certified for RC
flying.   It's free, it works, enjoy.  This is an open source project, so if
you don't like the way something works, help us fix it!

Thanks:
  Bruce Jackson for the LaRCsim framework.
  Flight Gear project for the sky sphere.

Contacts:
If you'd like to help with CRRCSIM, then send me an email!
    email                : kansky@ll.mit.edu
*****************************************************************************/

#include "global.h"
#include "defines.h"
#include "crrc_loadair.h"
#include "crrc_main.h"
#include "crrc_system.h"
#include "crrc_main_cpp.h"
#include "crrc_config.h"
#include "crrc_sound.h"
#include "crrc_scenery.h"
#include "SimStateHandler.h"
#include "crrc_graphics.h"
#include "windfield.h"
#include "GUI/crrc_gui_main.h"
#include "GUI/crrc_joy.h"
#include "SimpleXMLTransfer.h"
#include "lib_conversions.h"
#include "config.h"
#include "zoom.h"
#include "CTime.h"
#include "T_GameHandler.h"
#include "F3F/handlerF3F.h"
#include "airtoxml.h"
#include "xmlmodelfile.h"
#include "crrc_rand.h"
#include "glconsole.h"
#include "crrc_fdm.h"

#include "interface_serial2/tx_serial2.h"
#include "interface_mnav/tx_mnav.h"
#include "audio_interface/tx_audio.h"
#include "interface_parallel/parallel.h"
#include "interface_software/tx_software.h"
#include "interface_serpic/tx_interface_serpic.h"
#include "T_TX_InterfaceSerial.h"
#include "interface_zhenhua/tx_interface_zhenhua.h"

#ifdef linux
#include "interface_rctran/tx_rctran.h"
#include "interface_rctran2/tx_rctran2.h"
#endif

#include <math.h>
#include <string>
#include <time.h>

//#include <plib/ssg.h>

//#define LOG_FRAMES

/*****************************************************************************/
//Configuration Data
TInputDev  inputDev;

CTime   *crrc_time;

// Functions defined outside crrc_main.c and used inside crrc_main.c :
void key_down(SDL_keysym *keysym);

// Functions defined and used in crrc_main.cpp
void idle(TSimInputs* inputs);

// Interface to TX
T_TX_Interface* TXInterface = (T_TX_Interface*)0;

// Pointer to scenery object and sky sphere
CVector *player_pos;

// Variometer sound
T_VariometerSound *vario_sound = NULL;
int vario_sound_channel = -1;

// Enviroment interface to FDM
FDMEnviroment* fdmenv = 0;


/*****************************************************************************/
void activate_test_mode()
{
  double dDist;

  TSimInputs inp  = TSimInputs();
  
  // store pointer to real fdm
  Global::testmode.fdm_backup = Global::fdm;
  // 
  Global::testmode.test_mode  = TRUE;
  
  // Set distance to plane to make it visible in a good size, set 
  // fixed flFieldOfView.
  {            
    double dAircraftSize;
    float  flFieldOfView; // degrees
   
    Global::testmode.flAutozoom = flAutozoom;
    flAutozoom          = 0; // autozoom disabled
    
    dAircraftSize = Global::fdm->airplane->getAircraftSize();
    // this is just to read flFieldOfView
    flFieldOfView = zoom_calc(dAircraftSize);
    
    // 0.42 * 2* tan(M_PI * flFieldOfView /2 / 180) * dDist  = 2*dAircraftSize
    dDist = dAircraftSize / ( 0.42 * tan(M_PI * flFieldOfView / 2 / 180) );
  }
  
  // create testmode fdm
  CVector ppos = Global::scenery->getPlayerPosition();
  Global::fdm = new CRRC_FDM();
  Global::fdm->loadAirplaneTestmode(-ppos.z, ppos.x - dDist, -ppos.y);
  initialize_flight_model();
  Global::fdm->update(&inp, 0, 0);
}

void leave_test_mode()
{
  //
  Global::testmode.test_mode  = FALSE;
  // delete testmode fdm
  delete Global::fdm;
  // restore real fdm
  Global::fdm = Global::testmode.fdm_backup;
  // restore autozoom
  flAutozoom = Global::testmode.flAutozoom;
}

/*****************************************************************************/

/** \brief Calculate Z rotation for simulated SAL
 *
 *  The angular velocity is calculated using an estimated
 *  rotation radius for the C-of-G of 0.8 m (stretched arm)
 *  plus half the wingspan. The value is negative, so a right-hand
 *  SAL is simulated.
 *
 *  \param vel relative launch velocity
 *
 *  \return body rotation around Z axis in rad/s
 */
double calculate_z_rotation(double vel)
{
  double radius = (0.8 / 0.3048) + (Global::fdm->airplane->getWingspan() / 2.0);
  double velocity = vel * Global::fdm->airplane->getTrimmedFlightVelocity(); // ft/s
  return (-1 * velocity / radius);
}


void initialize_flight_model()
{
  float Altitude;

  Altitude = cfgfile->getDouble("launch.altitude", 6)
    + Global::fdm->airplane->getZLow()
    + Global::scenery->getHeight(player_pos->x, player_pos->z); // height of CG above reference ellipsoid [ft]

  double velocity_rel = cfgfile->getDouble("launch.velocity_rel", 1);
  double dZRot = 0.0;
  
  if (cfgfile->getInt("launch.sal", 0) == 1)
  {
    dZRot = calculate_z_rotation(velocity_rel);
  }
  
  Global::fdm->initAirplaneState(velocity_rel,
                                 cfgfile->getDouble("launch.angle", 0),
                                 (cfg->wind->getDirection()*M_PI/180),
                                 cfg->getCurLocCfgPtr(cfgfile)->getDouble("position.airplane.x"),
                                 cfg->getCurLocCfgPtr(cfgfile)->getDouble("position.airplane.y"),
                                 -1*Altitude,
                                 0.0,
                                 0.0,
                                 dZRot);
  
  Global::Simulation->resume();
}

/*****************************************************************************/

/**
 *  Handle mouse buttons
 *
 *  This method evaluates any SDL mouse button events and dispatches them
 *  to the configured handler functions.
 *
 *  Note that mouse wheel movement is also reported as button presses.
 *  Scrolling the wheel up generates an SDL_MOUSEBUTTONDOWN event for
 *  SDL_BUTTON_WHEELUP, immediately followed by an SDL_MOUSEBUTTONDOWN event.
 *
 *  \todo Instead of handling the mouse wheel events like buttons, we could
 *        in general handle it like a third axis for the mouse interface
 *        and let the user configure the axis mapping to his liking.
 *
 *  \param event  The pending SDL_MOUSEBUTTONDOWN event.
 */
void mouse_button(SDL_MouseButtonEvent *event)
{
  if (event->state != SDL_PRESSED)
    return;
  
  int mouse_binding = NOTHING;

  switch (event->button)
  {
    case SDL_BUTTON_LEFT:
      mouse_binding = inputDev.mouse_bind_l;
      break;
      
    case SDL_BUTTON_RIGHT:
      mouse_binding = inputDev.mouse_bind_r;
      break;      
      
    case SDL_BUTTON_MIDDLE:
      mouse_binding = inputDev.mouse_bind_m;
      break;
    
#ifndef SDL_WITHOUT_MOUSEWHEEL
    case SDL_BUTTON_WHEELUP:
      mouse_binding = inputDev.mouse_bind_u;
      break;

    case SDL_BUTTON_WHEELDOWN:
      mouse_binding = inputDev.mouse_bind_d;
      break;
#endif
    
    default:
      break;
  }
      
  switch (mouse_binding)
  {
    case RESUME:
      Global::Simulation->resume();
      break;
    
    case RESET:
      Global::Simulation->reset();
      break;
    
    case PAUSE:
      Global::Simulation->pause();
      LOG("Press <u> to resume.");
      break;
    
    case INCTHROTTLE:
      TXInterface->increase_throttle();
      break;
    
    case DECTHROTTLE:
      TXInterface->decrease_throttle();
      break;
    
    default:
      break;
  }
}


/***************************************************************************/
#ifndef NORM_MOUSE_X
# define NORM_MOUSE_X(X) ((float)X / window_xsize - 0.5)
# define NORM_MOUSE_Y(Y) ((float)Y / window_ysize - 0.5)
#endif
void mouse_motion(SDL_MouseMotionEvent *event)
{
  int x, y;

  // Not returning in test_mode makes test_mode work with mouse input,
  // but messes up input values a little.
  //~ if (nState != STATE_RUN && !(test_mode && nState == STATE_RESUMING))
  //~ return;

  x = event->x;
  y = event->y;

  if ((inputDev.zoom_control==MOUSE) && !(Global::gui && Global::gui->isVisible()))
  {
    zoom_set(y);
    return;
  }

  if (TXInterface->inputMethod() == T_TX_Interface::eIM_mouse)
  {
    TXInterface->setAxis(0, NORM_MOUSE_X(x));
    TXInterface->setAxis(1, NORM_MOUSE_Y(y));
  }
}

/*****************************************************************************/
#ifndef NORM_JOYSTICK
# define NORM_JOYSTICK(X) ((float)(X) / 65536.0)
//#define NORM_JOYSTICK(X) (((float)(X) -50 )/ 200 ) // for realflight joystick&linux

#endif
void joystick_motion(SDL_JoyAxisEvent *event)
{
  if (event->axis > MAXJOYAXIS)
    return;

  TXInterface->setAxis(event->axis, NORM_JOYSTICK(event->value));
}

void joystick_button(SDL_JoyButtonEvent *event)
{
  if (event->state != SDL_PRESSED || event->button > MAXJOYBUTTON)
    return;

  switch (inputDev.joystick_bind_b[event->button])
  {
   case RESUME:
    Global::Simulation->resume();
    break;
   case RESET:
    Global::Simulation->reset();
    break;
   case PAUSE:
    Global::Simulation->pause();
    LOG("Press <u> to resume.");
    break;
   case INCTHROTTLE:
    TXInterface->increase_throttle();
    break;
   case DECTHROTTLE:
    TXInterface->decrease_throttle();
    break;
   case ZOOMIN:
    zoom_in();
    break;
   case ZOOMOUT:
    zoom_out();
    break;
   default:
    break;
  }
}

/*****************************************************************************/
void calc_fps()
{
  static int last_time = 0;
  int cur_time, time_diff;
  static int avg_diff = 10;
  int avg_fps;

  cur_time = SDL_GetTicks();
  time_diff = (cur_time - last_time);

  if (time_diff < 1)
    time_diff = 10;

  avg_diff += ((time_diff << 8) - avg_diff) >> 5;

  avg_fps = (int)rintf(1000.0 / (avg_diff >> 8));
  Global::nFPS = avg_fps;

  last_time = cur_time;
}

void handle_events()
{
  SDL_Event event;
  while (SDL_PollEvent(&event))
  {
    switch (event.type)
    {
     case SDL_VIDEORESIZE:
      resize_window(event.resize.w, event.resize.h);
      break;

     case SDL_JOYAXISMOTION:
      if (nJoystickDlg)
        joystickDlg->joystickDlgMotion(&event.jaxis);
      else
        joystick_motion(&event.jaxis);
      break;

     case SDL_JOYBUTTONDOWN:
      if (nJoystickDlg)
        joystickDlg->joystickDlgButton(&event.jbutton);
      else
        joystick_button(&event.jbutton);
      break;

     case SDL_MOUSEMOTION:
      /**
       * GUI_MOUSE_MOTION_WORKAROUND
       * The GUI sometimes returns that it used the event, although it is not visible anymore!
       * Therefore we need to check whether it is visible at all.
       * Where it works:
       *   -ESC to show GUI
       *   -open dialog (for example video)
       *   -close dialog
       *   -ESC to hide GUI
       * Everything is fine, the mouse-motions are not used by the GUI.
       *   -ESC to show GUI
       *   -toggle something from the menu bar (for example Verbosity)
       *   -ESC to hide GUI
       * Now the GUI still uses the mouse motions, although it is invisible!
       */
      if (TXInterface->inputMethod() == T_TX_Interface::eIM_mouse)
      {
        // always update the mouse interface
        mouse_motion(&event.motion);
        if (Global::gui)
          Global::gui->mouseMotionHandler(event.motion.x, event.motion.y);
      }
      else if (!Global::gui || !Global::gui->isVisible()
               ||
               !Global::gui->mouseMotionHandler(event.motion.x, event.motion.y))
      {
        // GUI did not use the event, pass it to the simulation itself
        mouse_motion(&event.motion);
      }
      break;

     case SDL_MOUSEBUTTONDOWN:
#if TEST_WITHOUT_JOYSTICK > 0
      if (nJoystickDlg && event.button.button != SDL_BUTTON_LEFT)
      {
        SDL_JoyButtonEvent jevent;
        jevent.type   = SDL_JOYBUTTONDOWN;
        jevent.state  = SDL_PRESSED;
        jevent.which  = 0;
        switch (event.button.button)
        {
         case SDL_BUTTON_MIDDLE:
          jevent.button = 0;
          break;
         case SDL_BUTTON_RIGHT:
          jevent.button = 1;
          break;
         default:
          jevent.button = 2;
          break;
        }
        joystickDlg->joystickDlgButton(&jevent);
      }
#endif
      if (!Global::gui || !Global::gui->mouseButtonDownHandler(event.button.button, event.button.x, event.button.y))
      {
        // GUI did not use the event, pass it to the simulation itself
        mouse_button(&event.button);
      }
      break;

     case SDL_MOUSEBUTTONUP:
      if (Global::gui)
        Global::gui->mouseButtonUpHandler(event.button.button, event.button.x, event.button.y);
      break;

     case SDL_KEYDOWN:
      if (!Global::gui || !Global::gui->keyDownEventHandler(event.key.keysym))
      {
        // GUI did not use the event, pass it to the simulation itself
        key_down(&event.key.keysym);
      }
      break;

     case SDL_KEYUP:
      if (Global::gui)
        Global::gui->keyUpEventHandler(event.key.keysym);
      break;

     case SDL_QUIT:
      if (Global::gui)
        Global::gui->doQuitDialog();
      else
        Global::Simulation->quit();
      break;
    }
  }
}

/**
 * Reverts to mouse input.
 */
std::string input_method_failed(std::string msg)
{
  if (msg.length())
    msg += "\n";

  msg += "Failed to initialize selected input method.\nReverting to mouse input.";

  // close previous interface
  if (TXInterface != (T_TX_Interface*)0)
  {
    delete TXInterface;
    TXInterface = (T_TX_Interface*)0;
  }

  // select and open mouse interface
  cfgfile->setAttributeOverwrite("inputMethod.method", InputMethodStrings[T_TX_Interface::eIM_mouse]);

  TXInterface  = new T_TX_InterfaceSoftware(T_TX_Interface::eIM_mouse);
  if (TXInterface->init(cfgfile))
    msg += "\nEven mouse input failed:\n" + TXInterface->getErrMsg();

  return(msg);
}

/**
 * (Re)configures the input method according to the configuration
 * in <tt>cfgfile</tt>.
 * Returns a message, if something (bad) happened.
 */
std::string reconfigureInputMethod()
{
  printf("std::string reconfigureInputMethod()\n");

  std::string method = strU(cfgfile->getString("inputMethod.method", InputMethodStrings[T_TX_Interface::eIM_mouse]));
  std::string msg;

  printf("New input method: %s\n", method.c_str());

  // stop old input method
  if (TXInterface != (T_TX_Interface*)0)
  {
    inputDev.closeJoystick();
    delete TXInterface;
    TXInterface = (T_TX_Interface*)0;
  }

  // create new one
  if (method.compare(strU(InputMethodStrings[T_TX_Interface::eIM_keyboard])) == 0)
  {
    TXInterface  = new T_TX_InterfaceSoftware(T_TX_Interface::eIM_keyboard);
    TXInterface->init(cfgfile);
  }
  else if (method.compare(strU(InputMethodStrings[T_TX_Interface::eIM_mouse])) == 0)
  {
    TXInterface  = new T_TX_InterfaceSoftware(T_TX_Interface::eIM_mouse);
    TXInterface->init(cfgfile);
  }
  else if (method.compare(strU(InputMethodStrings[T_TX_Interface::eIM_joystick])) == 0)
  {
    TXInterface  = new T_TX_InterfaceSoftware(T_TX_Interface::eIM_joystick);
    TXInterface->init(cfgfile);

#if TEST_WITHOUT_JOYSTICK == 0
    msg = inputDev.openJoystick();
    if (msg.length())
      return(input_method_failed(msg));
#endif
  }
  else if (method.compare(strU(InputMethodStrings[T_TX_Interface::eIM_audio])) == 0)
  {
    #if PORTAUDIO > 0
    TXInterface  = new T_TX_InterfaceAudio();
    if (TXInterface->init(cfgfile))
      return(input_method_failed(TXInterface->getErrMsg()));
    #else
      return(input_method_failed("This version of CRRCsim was built without support for the AUDIO interface.\n"));
    #endif
  }
#ifdef linux
  else if (method.compare(strU(InputMethodStrings[T_TX_Interface::eIM_rctran])) == 0)
  {
    TXInterface  = new T_TX_InterfaceRCTRAN();
    if (TXInterface->init(cfgfile))
      return(input_method_failed(TXInterface->getErrMsg()));
  }
  else if (method.compare(strU(InputMethodStrings[T_TX_Interface::eIM_rctran2])) == 0)
  {
    TXInterface  = new T_TX_Interface_RCTran2();
    if (TXInterface->init(cfgfile))
      return(input_method_failed(TXInterface->getErrMsg()));
  }
#endif
  else if (method.compare(strU(InputMethodStrings[T_TX_Interface::eIM_parallel]))  == 0)
  {
    printf("Trying to init parallel\n");
    TXInterface  = new T_TX_InterfaceParallel();
    if (TXInterface->init(cfgfile))
      return(input_method_failed(TXInterface->getErrMsg()));
  }
  else if (method.compare(strU(InputMethodStrings[T_TX_Interface::eIM_serial2])) == 0)
  {
    TXInterface  = new T_TX_InterfaceSerial2();
    if (TXInterface->init(cfgfile))
      return(input_method_failed(TXInterface->getErrMsg()));
  }
  else if (method.compare(strU(InputMethodStrings[T_TX_Interface::eIM_mnav])) == 0)
  {
    TXInterface  = new T_TX_InterfaceMNAV();
    if (TXInterface->init(cfgfile))
      return(input_method_failed(TXInterface->getErrMsg()));
  }
  else if (method.compare(strU(InputMethodStrings[T_TX_Interface::eIM_serpic])) == 0)
  {
    TXInterface  = new T_TX_InterfaceSerPIC();
    if (TXInterface->init(cfgfile))
      return(input_method_failed(TXInterface->getErrMsg()));
  }
  else if (method.compare(strU(InputMethodStrings[T_TX_Interface::eIM_zhenhua])) == 0)
  {
    TXInterface  = new T_TX_InterfaceZhenHua();
    if (TXInterface->init(cfgfile))
      return(input_method_failed(TXInterface->getErrMsg()));
  }
  else
    return(input_method_failed(""));

  return("");
}

/**
 * Reading of (mostly initial) values from the xml config into
 * global variables (at least global to this file).
 * Maybe those variables can be encapsulated somewhere else...
 */
void read_config_into_globals()
{
  zoom_reset();
  Global::training_mode = cfgfile->getInt("training_mode.fUse", 0);
  Global::dt = cfgfile->getDouble("simulation.flightModel.dt", 0.002777);
  graphics_read_config(cfgfile);
}

void write_globals_into_config()
{
  cfgfile->setAttributeOverwrite("training_mode.fUse", Global::training_mode);
  cfgfile->setAttributeOverwrite("simulation.flightModel.dt",
                                 doubleToString(Global::dt));
}

/*****************************************************************************/
void initializeRandomNumberGenerator()
{
  time_t sometime = time(0);
  srand((unsigned int) sometime);
  CRRC_Random::insertData(rand());
  std::cout << "RAND_MAX = " << RAND_MAX << "\n";
}

/*****************************************************************************/
/** 
 *  This function tries to perform some cleanup when
 *  exiting. If errmsg != NULL, a system message box
 *  will be opened with the message string.
 *
 *  \param exit_code should be CRRC_EXIT_SUCCESS or CRRC_EXIT_FAILURE
 *  \param errmsg    optional message to be displayed on exit
 */
void crrc_exit(int exit_code, const char *errmsg)
{
  // ToDo: clean up allocated objects
  SDL_Quit();
  
  if ((errmsg != NULL) && (*errmsg != '\0'))
  {
    if (exit_code == CRRC_EXIT_SUCCESS)
    {
      SystemMessage(errmsg, SM_INFO);
    }
    else
    {
      SystemMessage(errmsg, SM_ERROR);
    }
  }
  
  exit(exit_code);
}

/**
 * Description: see header file
 */
void loadAirplane()
{
  std::string filename = cfgfile->getString("airplane.file", "models/allegro.xml");
  filename = air_to_xml_file_load(filename);

  try
  {
    SimpleXMLTransfer* xml = new SimpleXMLTransfer(filename);

    SimpleXMLTransfer* ap = cfgfile->getChild("airplane");

    // Here we copy graphics and config preferences from crrcsim's config file
    // into the in-memory-copy of the airplane. This is because an airplane file 
    // should not be altered by user preferences.
    XMLModelFile::SetGraphics(xml, ap->attributeAsInt("graphics", 0));
    XMLModelFile::SetConfig  (xml, ap->attributeAsInt("config",   0));

    Global::fdm->loadAirplane(xml, fdmenv, cfgfile);
    if (cfgfile->getInt("video.enabled", 1))
      Global::airplane = new CRRCAirplaneLaRCSimSSG(xml, scene);
    else
      Global::airplane = (CRRCAirplane*)0;

    delete xml;
  }
  catch (XMLException e)
  {
    std::string msg = "Error opening airplane specification file: ";
    msg += filename;
    msg += ": ";
    msg += e.what();

    throw std::runtime_error(msg);    
  }  
}


void set_aux(int aux_num, int setting)
{
  if (aux_num <= 0 || aux_num > NUM_AUX_INPUTS)
    return;
  switch(setting)
  {
    case 1:
      Global::inputs.aux[aux_num - 1] = -0.5;
      break;
    case 2:
      Global::inputs.aux[aux_num - 1] =  0.0;
      break;
    case 3:
      Global::inputs.aux[aux_num - 1] =  0.5;
      break;
    default:
      break;
  }
}



/*****************************************************************************/
int main(int argc,char **argv)
{
  float        field_of_view;
  
  Global::testmode.test_mode = FALSE;

  Global::Simulation = new SimStateHandler();
  
  Global::fdm = new CRRC_FDM();
  fdmenv = new CRRC_FDM_Env();
  
  std::string startup_time;
  getSystemTimeString(startup_time);
  printf("CRRCsim %s started at %s\n", PACKAGE_VERSION, startup_time.c_str());
  printf("Running on %s\n", getOSVersionString());
  printf("Using plib version %d.%d.%d\n", PLIB_MAJOR_VERSION,
                                          PLIB_MINOR_VERSION,
                                          PLIB_TINY_VERSION);
  printf("Compiled with SDL version %d.%d.%d\n",  SDL_MAJOR_VERSION,
                                                  SDL_MINOR_VERSION,
                                                  SDL_PATCHLEVEL);
  printf("(Linked SDL version is %d.%d.%d)\n",  SDL_Linked_Version()->major,
                                                SDL_Linked_Version()->minor,
                                                SDL_Linked_Version()->patch);
#ifdef CRRC_DATA_PATH
  printf("Configured data path: %s\n", CRRC_DATA_PATH);
#endif

  {
    std::vector<std::string> search_path;
    
    T_Config::getSearchPathList(search_path, "");
    std::cout << "Data file search path:" << std::endl;
    for (std::vector<std::string>::size_type i = 0; i < search_path.size(); i++)
    {
      std::cout << "  " << search_path[i] << std::endl;
    }
  }

  initializeRandomNumberGenerator();

  {
    unsigned int SDLFlags = SDL_INIT_JOYSTICK;
    int nRetCodeCmdline;
    int i;

    try
    {
      // ***** Read configuration, parse commandline... ***********************
      for (i = 1; i < argc - 1; i++)
      {
        if (!strcmp(argv[i], "-g"))
          T_Config::putConfigFilePath(argv[i+1]);
      }
      
      cfg = new T_Config(cfgfile); // This will also set up cfgfile
      cfg->read(cfgfile);
      
      {
        int nNum = air_to_xml();
        if (nNum != 0)
        {
          // todo: files have been converted; issue message.
        }
      }
                  
      inputDev.init(cfgfile);
      
      // command line options override settings read from the config file
      nRetCodeCmdline = crrc_checkopts(argc, argv, cfgfile, cfg);

      if (nRetCodeCmdline)
        crrc_exit(CRRC_EXIT_FAILURE);

      // must be after crrc_checkopts because crrc_checkopts can change
      //   video.enabled and sound.enabled based on command line options
      if (cfgfile->getInt("video.enabled", 1))
        SDLFlags |= SDL_INIT_VIDEO;
      if (cfgfile->getInt("sound.enabled", 1))
        SDLFlags |= SDL_INIT_AUDIO;
      SDL_Init(SDLFlags);
      SDL_EnableUNICODE(1); // We need this to pass keys to pui
      SDL_EnableKeyRepeat(50, 150);

      read_config_into_globals();

      std::string msg = reconfigureInputMethod();
      if (msg.length())
        printf("%s", msg.c_str());

      // ***** Video setup ****************************************************
      if (cfgfile->getInt("video.enabled", 1))
        video_setup(0, 0, 0);

      // ***** Setting window caption *****************************************
      if (cfgfile->getInt("video.enabled", 1))
        setWindowTitleString();

      // ***** Sound **********************************************************
      if (cfgfile->getInt("sound.enabled", 1))
      {
        try
        {
          Global::soundserver = new CRRCAudioServer(cfgfile);
          // init the variometer sound
          cfgfile->makeSureAttributeExists("sound.variometer.vol", "0");
          int vario_sound_volume = cfgfile->getInt("sound.variometer.vol") << 3;
          vario_sound = new T_VariometerSound(Global::soundserver->getAudioSpec());
          vario_sound_channel = Global::soundserver->playSample((T_SoundSample*)vario_sound, vario_sound_volume);
          soundUpdate3D(0.0, 0.0, 0.0, 0.0);

          Global::soundserver->pause();
        }
        catch (std::runtime_error& e)
        {
          fprintf(stderr, "%s\n", e.what());
          Global::soundserver = (CRRCAudioServer*)0;
        }
      }
      else
        Global::soundserver = (CRRCAudioServer*)0;

      // ***** Video **********************************************************

      initialize_scenegraph();
  
      std::string sceneryfile = cfg->getLocationName();
      Global::scenery = loadScenery(T_Config::getDataPath(sceneryfile).c_str());
      if (Global::scenery == NULL)
      {
        fprintf(stderr, "Unable to initialize scenery from file %s,\n",
                        sceneryfile.c_str());
        fprintf(stderr, "reverting to default scenery \"scenery/davis-orig.xml\"\n");
        
        // try the default scenery
        Global::scenery = loadScenery(T_Config::getDataPath("scenery/davis-orig.xml").c_str());
        if (Global::scenery == NULL)
        {
          std::string s;
          s = "Unable to initialize default scenery from file \"scenery/davis-orig.xml\"";
          crrc_exit(CRRC_EXIT_FAILURE, s.c_str());
        }
        else
        {
          // set configuration file back to default
          cfg->setLocation("scenery/davis-orig.xml", cfgfile);
        }
      }
      // re-read the wind configuration to activate any scenery defaults if
      // no wind config is present in the cfgfile for this location
      cfg->wind->read(cfgfile, cfg);

      if (cfgfile->getInt("video.enabled", 1))
        initialize_window(cfg);
      initialize_wind_field();

      player_pos = new CVector(Global::scenery->getPlayerPosition());

      // Create invisible GUI
      if (cfgfile->getInt("video.enabled", 1))
        Global::gui = new CGUIMain(false);
      else
        Global::gui = NULL;

      // load airplane
      {
        bool airplane_failed = false;
        try
        {
          // load the airplane specified in the config file
          loadAirplane();
          initialize_flight_model();
        }
        catch (std::runtime_error& e)
        {
          fprintf(stderr, "%s\n", e.what());
          airplane_failed = true;
        }
        if (airplane_failed)
        {
          // Failed to load airplane file.
          // Using some fallback.
          cfgfile->setAttributeOverwrite("airplane.file", T_Config::getDataPath("models/allegro.xml"));
          try
          {
            loadAirplane();
            initialize_flight_model();
          }
          catch (std::runtime_error& e)
          {
            std::string s = "Unable to load airplane file:\n";
            s += e.what();
            fprintf(stderr, "%s\n", s.c_str());
            crrc_exit(CRRC_EXIT_FAILURE, s.c_str());
          }
        }
      }

      //initialise game mode
      if (cfgfile->getInt("game.f3f.enabled",0))
        Global::gameHandler= new HandlerF3F();
      else
        Global::gameHandler= new T_GameHandler();
    }
    catch (XMLException e)
    {
      std::string s = "XMLException: ";
      s += e.what();
      fprintf(stderr, "%s\n", s.c_str());
      crrc_exit(CRRC_EXIT_FAILURE, s.c_str());
    }
  }

  if (Global::soundserver != (CRRCAudioServer*)0)
    Global::soundserver->pause(false);
  Global::Simulation->reset();
  
  //~ nVerbosity = 3;
  crrc_time = new CTime(cfgfile);
  
  Global::console = new GlConsole(400, 125, 15, 15);
  if (Global::console == NULL)
  {
    crrc_exit(CRRC_EXIT_FAILURE, "Unable to initialize console.");
  }
  Global::console->setAutoHide(4000, 1000);
  LOG("CRRCsim successfully started!");
  LOG("Press <ESC> to show the setup menu.");
  
#ifdef LOG_FRAMES
  FILE* fp = fopen("frames.dat", "w");
  if (fp == NULL)
  {
    crrc_exit(CRRC_EXIT_FAILURE, "Failed.");
  }
#endif
  
  while (Global::Simulation->getState() != STATE_EXIT)
  {
    crrc_time->update();
    handle_events();

    TXInterface->getInputData(&Global::inputs);

    Global::Simulation->doIdle(&Global::inputs);
    
    // random data
    {
      CRRC_Random::insertData(SDL_GetTicks());
      CRRC_Random::insertData(Global::inputs.getRandNum());                   
    }

    float distance_to_model;
    {
      double X_cg_rwy =    Global::fdm->airplane->getPos().r[0];
      double Y_cg_rwy =    Global::fdm->airplane->getPos().r[1];
      double H_cg_rwy = -1*Global::fdm->airplane->getPos().r[2];
      
      distance_to_model = player_pos->distance(X_cg_rwy, H_cg_rwy, Y_cg_rwy);
    }
    
    field_of_view = zoom_calc(distance_to_model);
    if (Global::gui)
      adjust_zoom(field_of_view);
    calc_fps();
    
    #if 0
    Global::verboseString += " X: " + ftoStr(Global::fdm->airplane->getPos().r[0], 2, 2, true, false);
    Global::verboseString += " Y: " + ftoStr(Global::fdm->airplane->getPos().r[1], 2, 2, true, false);
    Global::verboseString += " Z: " + ftoStr(Global::fdm->airplane->getPos().r[2], 2, 2, true, false);
    Global::verboseString += " Phi: " + ftoStr(Global::fdm->airplane->getPhi() * SG_RADIANS_TO_DEGREES, 2, 2, true, false);
    Global::verboseString += " Theta: " + ftoStr(Global::fdm->airplane->getTheta() * SG_RADIANS_TO_DEGREES, 2, 2, true, false);
    Global::verboseString += " Psi: " + ftoStr(Global::fdm->airplane->getPsi() * SG_RADIANS_TO_DEGREES, 2, 2, true, false);
    if (Global::gui)
      Global::gui->setVerboseText(Global::verboseString.c_str());
    #else
    switch (Global::nVerbosity)
    {
     case 3:
      Global::verboseString += "FPS: " + itoStr(Global::nFPS, ' ', 1);
      //fallthrough
     case 2:
      Global::verboseString += " FoV: " + ftoStr(field_of_view, 2, 1, false, false);
      //fallthrough
     case 1:
      Global::verboseString += " Ail: " + ftoStr(Global::inputs.aileron, 2, 2, true, false)
        + " Ele: " + ftoStr(Global::inputs.elevator, 2, 2, true, false)
          + " Rud: "   + ftoStr(Global::inputs.rudder,   2, 2, true, false)
            + " Thr: " + ftoStr(Global::inputs.throttle, 2, 2, true, false);
      if (Global::gui)
        Global::gui->setVerboseText(Global::verboseString.c_str());
      else
      {
        static int verbose_print_c = 0;
        if (++verbose_print_c >= 30)
        {
          Global::verboseString += "\n";
          printf(Global::verboseString.c_str());
          verbose_print_c = 0;
        }
      }
      break;

     default:
      if (Global::gui)
        Global::gui->setVerboseText("");
      break;
    }
    #endif
    
    if (Global::gui)
      display();
    Global::verboseString = "";

#ifdef LOG_FRAMES
    fprintf(fp, "%lu\n", (unsigned long)SDL_GetTicks());
#endif
    
    // sound calculations
    if (Global::soundserver != (CRRCAudioServer*)0)
      soundUpdate3D(distance_to_model,
                    Global::fdm->airplane->getPropFreq(),
                    -1*Global::fdm->airplane->getPos().r[2],
                    Global::fdm->airplane->getVRelAirmass()/Global::fdm->airplane->getTrimmedFlightVelocity());
  }
#ifdef LOG_FRAMES
  fclose(fp);
#endif

  if (Global::airplane != (CRRCAirplane*)0)
    delete Global::airplane;
  delete Global::fdm;
  if (Global::soundserver != (CRRCAudioServer*)0)
    delete Global::soundserver;
  if (vario_sound != (T_VariometerSound*)0)
    delete vario_sound;
  delete Global::scenery;
  graphics_cleanup();
  delete Global::Simulation;

  crrc_exit(CRRC_EXIT_SUCCESS, NULL);

  // crrc_exit() will never return, keep the compiler happy anyway:
  return 0;
}

