#ifndef CRRC_RAND
#define CRRC_RAND

#include <stdlib.h>

/**
 * This is just a wrapper around standard functions rand and srand which
 * allows for easy re-seeding.
 * 
 * From libc manual:
 * 
 * All functions have in common that they use the same congruential formula 
 * with the same constants. The formula is
 * 
 *    Y = (a * X + c) mod m
 *
 * where X is the state of the generator at the beginning and Y the state at
 * the end. a and c are constants determining the way the generator works.
 * By default they are
 *
 *     a = 0x5DEECE66D = 25214903917
 *     c = 0xb = 11
 *
 * From man rand:
 *  POSIX 1003.1?2003 gives the following example of an implementation of 
 *  rand() and srand(), possibly useful when one needs the same sequence
 *  on two different machines.
 *
 *       static unsigned long next = 1;
 *       * 
 *       // RAND_MAX assumed to be 32767 
 *       int myrand(void) {
 *       next = next * 1103515245 + 12345;
 *       return((unsigned)(next/65536) % 32768);
 *       }
 * 
 * @author Jens Wilhelm Wulf
 */
class CRRC_Random
{
  public:
   
   /**
    * Returns a random number between 0 and max().
    */
   static inline int rand() { return(::rand()); };

   static inline int max() { return(RAND_MAX); };
   
   /**
    * Call this using some random data you have, anytime you want to.
    */
   static void insertData(int nData);
   
  private:
   
   /**
    * random state
    */
   static unsigned int uRandState16;
   static unsigned int uRandState32;
   
};

#endif

