/** \file crrc_sky.h
 *
 *  Declaration of a set of sky rendering classes.
 */
#ifndef _CRRC_SKY_H
#define _CRRC_SKY_H

#include <plib/sg.h>
#include <plib/ssg.h>
#include <plib/ssgaSky.h>

#include <string>

#if defined(__APPLE__) || defined(MACOSX)
#define DEFAULT_SKYBOX_TEXTURE_OFFSET (0.0009f)
#else
#define DEFAULT_SKYBOX_TEXTURE_OFFSET (0.0f)
#endif

/** \brief A base class for sky rendering classes.
 *
 *  This base class defines the common interface for
 *  all sky rendering classes. To create a class with this
 *  interface, at least the preDraw() method must
 *  be implemented.
 */
class SkyRenderer
{
  public:
    /// The destructor
    virtual ~SkyRenderer() {};
  
    /// Update the camera position and cloud movement
    virtual void update(sgVec3 *campos, double dt) {};
    
    /// Hook for the first drawing stage
    virtual void preDraw() = 0;
  
    /// Hook for the second drawing stage
    virtual void postDraw(float altitude) {};
  
};


/* ******************************************************** */

/** \brief A skybox class for use with PLIB's SSG.
 *
 *  This class creates a cubic skybox.
 */
class SkyBox : public SkyRenderer
{
  private:
    ssgRoot       *skyroot;
    ssgTransform  *skyboxtrans;
    float         boxsize;
  
  public:
    /// Create a skybox
    SkyBox(const char **textures, float size = 20.0, float texoffset = 0.0);
  
    /// Delete a skybox
    ~SkyBox();
  
    /// Update the camera position and cloud movement
    void update(sgVec3 *campos, double dt);
  
    /// Draw the skybox
    void preDraw();
};


/* ******************************************************** */

/** \brief A sky dome class based on PLIB's ssgaSky.
 *
 *  This class renders a sky dome using the ssgaSky class.
 */
#if 0
class SkyDome : public SkyRenderer
{
  private:
    ssgaSky   *theSky;
    float     radius;
    sgdVec3   planetsDummy;
    sgdVec3   starsDummy;
    sgVec4    skyDomeCol;
    sgVec4    skyFogCol;
    sgVec4    skyCloudCol;

    void      repaint();

  public:
    SkyDome(float r,
            sgVec4 domecol, sgVec4 fogcol, sgVec4 cloudcol);
    ~SkyDome();

    /// Draw the sky dome
    void preDraw() {theSky->preDraw();}
  
    /// Draw the clouds
    void postDraw(float altitude) {theSky->postDraw(altitude);}
    
    /// Update the camera position and cloud movement
    void update(sgVec3 *campos, double dt);
};
#endif

/* ******************************************************** */

/** \brief CRRCsim's original sky dome
 *
 *  This class creates a textured sky dome using SSG.
 */
class CRRCSkyDome : public SkyRenderer
{
  private:
    ssgRoot   *skyroot;
    float     radius;
  
    void      repaint(const char *texture);

  public:
    CRRCSkyDome(const char *cloud_texture, float r = 8000.0);
    ~CRRCSkyDome();
  
    /// Draw the sky dome
    void preDraw();
  
};


/* ******************************************************** */

#define PANO_NUM_RINGS     10
#define PANO_NUM_SLICES    20
#define PANO_RENDER_MODE   1
#define PANO_CAM_HEIGHT    4
/** \brief A panorama renderer for CRRCsim
 *
 *  This class renders a 360*135 deg panorama dome.
 */
class CRRCPanoDome : public SkyRenderer
{
  private:
    void init_textures();
    void init_vertices();
    void render();

    std::string     texfilename;
    ssgTexture      *panotex;
    ssgSimpleState  *state;
    float           radius;
    sgVec3 vertices[PANO_NUM_RINGS][PANO_NUM_SLICES];
    sgVec2 texco[PANO_NUM_RINGS][PANO_NUM_SLICES];
  
  public:
    CRRCPanoDome(const char *texture, float r = 10.0);
  ~CRRCPanoDome();
  
  /// Draw the panorama
  void preDraw();
};



#endif  // _CRRC_SKY_H
