#ifndef CRRC_FDM_H
# define CRRC_FDM_H

#include <string>
#include <iostream>
#include <fstream>

#include "crrc_inputs.h"
#include "SimpleXMLTransfer.h"
#include "math/vector3.h"
#include "math/matrix33.h"
#include "fdm/fdm_env.h"

#define FDM_LOG             0
#define FDM_LOG_POS         1
#define FDM_LOG_WIND_IN     0
#define FDM_LOG_AERO_OUT    0

/**
 * Base class for handling the physics of an airplane.
 * You should not call any of its methods (with the exception of get-Methods). 
 * CRRC_FDM does it for you.
 * 
 * @author Jens Wilhelm Wulf
 */
class CRRC_AirplaneSim
{
   friend class CRRC_FDM;
  public:
   
   virtual CRRCMath::Vector3 getPos()=0;
   
   virtual double getPhi()     = 0;
   virtual double getTheta()   = 0;
   virtual double getPsi()     = 0;
   virtual bool   isStalling() = 0;

   /**
    * returns velocity w.r.t. earth surface
    */
   virtual CRRCMath::Vector3 getVel() { CRRCMath::Vector3 r; r = CRRCMath::Vector3(); return(r); };
   virtual CRRCMath::Vector3 getAccel() { CRRCMath::Vector3 r; r = CRRCMath::Vector3(); return(r); };
   
   virtual CRRCMath::Vector3 getPQR() { CRRCMath::Vector3 r; r = CRRCMath::Vector3(); return(r); };

   virtual double getLon() { return(-1.0); };
   virtual double getLat() { return(-1.0); };
   virtual double getAlt() { return(-1.0); };
   
   CRRC_AirplaneSim(const char* logfilename, FDMEnviroment* myEnv);
   
   virtual ~CRRC_AirplaneSim();
   
   /**
    * Used for sound calculation. It returns the prop's number of revolutions
    * per second [1/s].
    */
   virtual double getPropFreq() = 0;
   
   /**
    * Returns velocity relative to airmass [ft/s].
    */
   virtual double getVRelAirmass() = 0;
   
   /**
    * Returns relative battery capacity/fuel left (0..1).
    */
   virtual double getBatCapLeft() { return(0.68); };
   
   /**
    * the longest distance from any of the aircrafts points to the CG
    */
   virtual double getAircraftSize() = 0;
   
   /**
    * Wingspan of the aircraft in feet
    */
   virtual double getWingspan() = 0;
   
   /**
    * computed velocity for trimmed flight in dead air
    */
   virtual double getTrimmedFlightVelocity() = 0;
   
   /**
    * returns Z coordinate of lowest point
    */
   virtual double getZLow() = 0;
   
  private:
   
   virtual void update(TSimInputs* inputs,
                       double      dt,
                       int         multiloop) = 0;
   
   virtual void initAirplaneState(double dRelVel,
                                  double dTheta,
                                  double dPsi,
                                  double X,
                                  double Y,
                                  double Z,
                                  double R_X = 0.0,
                                  double R_Y = 0.0,
                                  double R_Z = 0.0) = 0;

  protected:
  
  /**
   * This is where I get information from the outside
   */
  FDMEnviroment* env;
  
#if FDM_LOG != 0
   std::ofstream logfile;
   
   int nLogCnt;
   
   int nStep;
#endif
   
   /**
    * Inline function which becomes nothing without logging.
    */
   void logNewline() 
   {
#if FDM_LOG != 0
     if (nLogCnt != 0)
       logfile << "\n";
     
     logfile << nLogCnt++ << " ";
#endif
   };
   
   /**
    * Inline function which becomes nothing without logging.
    */
   void logVal(double val)
   {
#if FDM_LOG != 0
     logfile << val << " ";
#endif
   };
      
   /**
    * Inline function which becomes nothing without logging.
    */
   void logVal(CRRCMath::Vector3 val)
   {
#if FDM_LOG != 0
     if (fabs(val.r[0]) > 2E6)
       val.r[0] = 0;
     if (fabs(val.r[1]) > 2E6)
       val.r[1] = 0;
     if (fabs(val.r[2]) > 2E6)
       val.r[2] = 0;
     
     logfile << val.r[0] << " " << val.r[1] << " " << val.r[2] << " ";
#endif
   };
      
   /**
    * Inline function which becomes nothing without logging.
    */
   void logVal(CRRCMath::Matrix33 val)
   {
#if FDM_LOG != 0
     
     for (unsigned int m=0; m<3; m++)
       for (unsigned int n=0; n<3; n++)
       {
         if (fabs(val.v[m][n]) > 2E6)
         {
           logfile << val.v[m][n] << " ";
//           logfile << "0 ";
         }
         else
           logfile << val.v[m][n] << " ";
       }
#endif
   };
      
};

/**
 * The purpose of this class is to hide the simulation core from the rest of
 * the application and to provide a lean interface. The long term goal is to be
 * able to use different models (beside larcsim).
 *
 * @author Jens Wilhelm Wulf
 */
class CRRC_FDM
{
  public:
   
   CRRC_FDM();
   
   ~CRRC_FDM();
   
   void Clean();  
  
   /**
    * Loads an airplane with fdm_testmode, which is a special FDM only
    * mapping control inputs to simple changes of position/orientation.
    * 
    * Parameters are position coordinates in feet.
    */
   void loadAirplaneTestmode(double dNorth, double dEast, double dDown);
   
   /**
    * Load airplane description from file. This method should be able to handle
    * different filetypes (create an instance of the correct CRRC_AirplaneSim).
    */
   void loadAirplane(const char* filename, 
                     FDMEnviroment* myEnv,
                     SimpleXMLTransfer* cfg);

   /**
    * Load airplane from xml description. This method should be able to handle
    * different filetypes (create an instance of the correct CRRC_AirplaneSim).
    */
   void loadAirplane(SimpleXMLTransfer* xml, 
                     FDMEnviroment* myEnv,
                     SimpleXMLTransfer* cfg);

   /**
    * Init state of airplane (after it has been loaded from a file using loadAirplane).
    * 
    * @param dRelVel velocity (relative to trimmed flight)
    */
   void initAirplaneState(double dRelVel,
                          double dTheta,
                          double dPsi,
                          double X,
                          double Y,
                          double Z,
                          double R_X = 0.0,
                          double R_Y = 0.0,
                          double R_Z = 0.0);
   
   /**
    * Update timestep.
    */
   void update(TSimInputs* inputs,
               double      dt,
               int         multiloop);

   /**
    * Return the launch presets
    */
   SimpleXMLTransfer* getLaunchPresets() const {return launch_presets;};

   /**
    * Pointer to airplane in use.
    */
   CRRC_AirplaneSim* airplane;
   
  private:
   /**
    * Pointer to airplane-specific launch presets
    */
   SimpleXMLTransfer* launch_presets;
};

#endif
