// -*- mode: c; mode: fold -*-
//
#ifndef FDM_002_H
# define FDM_002_H

# include <stdexcept>
# include <vector>
# include "../ls_types.h"
# include "../fdm.h"

# include "../math/vector3.h"
# include "../math/matrix33.h"
# include "../power/power.h"
# include "../physics/eom.h"

/**
 * Jens Wilhelm Wulf, September 2005
 * 
 * This even simpler than CRRC_AirplaneSim_Larcsim, as it does model a flat earth only. 
 * We don't need more for usual model airplane use.
 *
 */  
class CRRC_AirplaneSim_002 : public CRRC_AirplaneSim
{
   friend class CRRC_FDM;
  public:

   virtual bool   isStalling() { return(stalling != 0); };
   virtual CRRCMath::Vector3 getPos();
   virtual double getPhi();
   virtual double getTheta();
   virtual double getPsi();

   /**
    * Used for sound calculation. It returns the prop's number of revolutions
    * per second [1/s].
    */
   virtual double getPropFreq();

   /**
    * Returns velocity relative to airmass [ft/s].
    */
   virtual double getVRelAirmass() { return(v_V_body.length()); };

   /**
    * Returns relative battery capacity/fuel left (0..1).
    */
   virtual double getBatCapLeft() { return(power->getBatteryMin()); };

   /**
    * the longest distance from any of the aircrafts points to the CG
    */
   virtual double getAircraftSize() { return(dMaxSize); }; 

   /**
    * Wingspan of the aircraft in feet
    */
   virtual double getWingspan() { return(span_ft); };
   
   /**
    * computed velocity for trimmed flight in dead air
    */
   virtual double getTrimmedFlightVelocity() {return(trimmedFlightVelocity);}; 

   /**
    * returns Z coordinate of lowest point
    */
   virtual double getZLow() { return(dZLow); };

  private:
   
   void LoadFromXML(SimpleXMLTransfer* xml);
  
   void update(TSimInputs* inputs,
               double      dt,
               int         multiloop);

   virtual void initAirplaneState(double dRelVel,
                                  double dTheta,
                                  double dPsi,
                                  double X,
                                  double Y,
                                  double Z);
   
   /**
    * read from file
    */
   CRRC_AirplaneSim_002(const char* filename, FDMEnviroment* myEnv);

   /**
    * read from xml description
    */
   CRRC_AirplaneSim_002(SimpleXMLTransfer* xml, FDMEnviroment* myEnv);
   
   virtual ~CRRC_AirplaneSim_002();

  private:
   
   EOM_6DOF   eom;

   /// @name Aerodynamic data
   //@{
   DATA  C_ref;    //  reference chord (ft)
   DATA  B_ref;    //  reference span  (ft)
   DATA  S_ref;    //  reference area (ft^2)
   DATA  U_ref;    //  reference speed for Re-scaling of CD_prof  (ft/s)
   
   DATA  Alpha_0;  //  baseline alpha (rad)
   
   DATA  Cm_0   ;  //  baseline Cm at Alpha_0
   DATA  CL_0   ;  //  baseline CL at Alpha_0   
   DATA  CL_max ;  //  positive stall limit
   DATA  CL_min ;  //  negative stall limit
   DATA  CD_prof;  //  profile CD at U_ref
   
   DATA  Uexp_CD;  //  for Re-scaling of CD_prof  ~ (U/U_ref)^Uexp_CD
   
   DATA  CL_a;     // lift-force   / alpha    ~  2 pi / (1 + 2/AR)
   DATA  Cm_a;     // pitch-moment / alpha    (pitch stability)
   DATA  CY_b;     // side-force  / sideslip
   DATA  Cl_b;     // roll-moment / sideslip (crucial for rudder-only turns)   
   DATA  Cn_b;     //  yaw-moment  / sideslip (yaw stability)
   
   DATA  CL_q;     //  lift-force   / pitch-rate
   DATA  Cm_q;     //  pitch-moment / pitch-rate  (pitch damping)
   DATA  CY_p;     //  side-force  / roll-rate
   DATA  Cl_p;     //  roll-moment / roll-rate   (roll damping)
   DATA  Cn_p;     //  yaw-moment  / roll-rate   (yaw-roll coupling)   
   DATA  CY_r;     //  side-force  / yaw-rate
   DATA  Cl_r;     //  roll-moment / yaw-rate
   DATA  Cn_r;     //  yaw-moment  / yaw-rate  (yaw damping)
   
   DATA  CL_de;    //  lift-force   / elevator
   DATA  Cm_de;    //  pitch-moment / elevator
   DATA  CY_dr;    // side-force  / rudder
   DATA  Cl_dr;    // roll-moment / rudder
   DATA  Cn_dr;    // yaw-moment  / rudder
   DATA  CY_da;    // side-force  / aileron
   DATA  Cl_da;    // roll-moment / aileron
   DATA  Cn_da;    // yaw-moment  / aileron
   
   DATA eta_loc;
   DATA CG_arm;
   DATA CL_drop;
   DATA CD_stall;
   
   DATA span_eff;  // span efficiency: Effective span  0.95 for most planes, 0.85 flying wing
   DATA CL_CD0;    // CL at minimum profile CD: 0.30 for 7037, 0.15 MH32, 0.0 RG15, AGxx, power
   DATA CD_CLsq;   // d(CD)/d(CL^2),  curvature of parabolic profile polar: 0.01 composites, 0.015 saggy ships, 0.02 beat up ship
   DATA CD_AIsq;   // d(CD)/d(aileron^2) , curvature of ail. CD influence: 0.01/(max_aileron)^2
   DATA CD_ELsq;   // d(CD)/d(elevator^2), curvature of ele. CD influence: 0.01/(max_elevator)^2 for Zagi otherwise 0
   //@}

   /// @name Mass and inertia
   //@{
   DATA Mass;  // inertia
   DATA I_xx;  // inertia
   DATA I_yy;  // inertia
   DATA I_zz;  // inertia
   DATA I_xz;  // inertia
   //@}

   /// @name Gear and ground interaction
   //@{
   
   /**
    * This struct holds information about a hard point/wheel on the airplane.
    * todo: steering is unused
    */
   typedef struct
   {
     /**
      * body axes: x,y,z
      */
     CRRCMath::Vector3 v_P;
     
     double            spring_constant;
     double            spring_damping;
     
     int    steering_mapping;  // Indicates which RC channel controls the steering
     double steering_max_angle;  // Indicates maximum anglee of steering wheel
     double percent_brake;
     double caster_angle_rad;          
   } Wheel;
   
   /**
    * Vector containing all hard points/wheels
    */
   std::vector<Wheel> wheels;
            
   /**
    * Wingspan in feet (calculated from hardpoints)
    */
   double span_ft;
   //@}
      
  private:
   
   void gear(TSimInputs* inputs);
   void aero_init();
   void aero(SCALAR dt, TSimInputs* inputs);

  private:
   
   /// @name written by constructor
   //@{
   
   /**
    * Propulsion system: batteries, shafts, engines, propellers.
    */
   Power::Power* power;
   
   /**
    * Z coordinate of lowest point
    */
   double dZLow;
   
   /**
    * Longest distance of any of the airplanes points to the cg
    */
   double dMaxSize;
   
   /**
    * Velocity in trimmed flight; dead air.
    */
   float trimmedFlightVelocity;   
         
   //@}
   
   /// @name written by init, update
   //@{
   
   /**
    * velocity of airmass (steady winds)
    * north, east, down
    */
   CRRCMath::Vector3 v_V_local_airmass;
      
   /**
    * linear turbulence components, L frame 
    */
   CRRCMath::Vector3 v_V_gust_local;
   
   /**
    * Gradients of wind velocity, runway coordinates.
    * m_V_atmo_rwy.v[0][?] is U_atmo_?
    * m_V_atmo_rwy.v[1][?] is V_atmo_?
    * m_V_atmo_rwy.v[2][?] is W_atmo_?
    */
   CRRCMath::Matrix33 m_V_atmo_rwy;
   
   //@}
   
   /// @name written by aero
   //@{
   int stalling;
   
   /**
    * Force x/y/z
    */
   CRRCMath::Vector3 v_F_aero;
   
   /**
    * Moment
    * l/m/n <-> roll/pitch/yaw
    */
   CRRCMath::Vector3 v_M_aero;
   
   /**
    * velocity of body relative to airmass
    */
   CRRCMath::Vector3 v_V_body;
   
   //@}

   /// @name written by gear
   //@{
   
   CRRCMath::Vector3 v_F_gear;
   CRRCMath::Vector3 v_M_gear;
   
   //@}

   /// @name written by engine
   //@{
   CRRCMath::Vector3  v_F_engine;
   CRRCMath::Vector3  v_M_engine;
   //@}


};


#endif
