#include "tx_rctran.h"

#include "../lib_conversions.h"

#ifdef linux
  #include "../mbuff.h"
#endif

#include <stdio.h>

T_TX_InterfaceRCTRAN::T_TX_InterfaceRCTRAN()
{
}

T_TX_InterfaceRCTRAN::~T_TX_InterfaceRCTRAN()
{
  delete map;
  delete calib;
}

int T_TX_InterfaceRCTRAN::init(SimpleXMLTransfer* config)
{
  T_TX_Interface::init(config);

  cpu_speed = config->getDouble("inputMethod.rctran.cpu_speed", 456.507890);

#ifdef linux
  rctran = (volatile unsigned long long *)mbuff_attach("crrcsim",
                                                       sizeof(volatile unsigned long long) * 8);
#else
  rctran = 0;
#endif
  
  map = new T_AxisMapper(this, config, "inputMethod.rctran");
  calib = new T_Calibration(this, config, "inputMethod.rctran");

  if (!rctran)
  {
    errMsg = "Mapping shared memory for RC transmitter interface failed.\n";
    return(-1);
  }
  else
    return(0);
}

void T_TX_InterfaceRCTRAN::putBackIntoCfg(SimpleXMLTransfer* config)
{
  T_TX_Interface::putBackIntoCfg(config);

  config->setAttributeOverwrite("inputMethod.rctran.cpu_speed", doubleToString(cpu_speed));
  map->putBackIntoCfg(config);
}

void T_TX_InterfaceRCTRAN::getInputData(TSimInputs* inputs)
{
  float throttle = 0.0;
  
  float calibrated;
  int   axisnum;
  
  axisnum    = map->func[ELEVATOR];
  if (axisnum >= 0)
  {
    calibrated = calib->calibrate(axisnum, (float)rctran[axisnum])
                  * map->inv[ELEVATOR];
    inputs->elevator = mixer->mix_elevator(calibrated);
  }  

  axisnum    = map->func[AILERON];
  if (axisnum >= 0)
  {
    calibrated = calib->calibrate(axisnum, (float)rctran[axisnum])
                  * map->inv[AILERON];
    inputs->aileron = mixer->mix_aileron(calibrated);
  }

  axisnum    = map->func[RUDDER];
  if (axisnum >= 0)
  {
    calibrated = calib->calibrate(axisnum, (float)rctran[axisnum])
                  * map->inv[RUDDER];
    inputs->rudder = mixer->mix_rudder(calibrated);
  }
  
  axisnum    = map->func[THROTTLE];
  if (axisnum >= 0)
  {
    calibrated = calib->calibrate(axisnum, (float)rctran[axisnum])
                  * map->inv[THROTTLE];
    throttle = mixer->mix_throttle(calibrated + 0.5);
  }
  if (throttle > 0.95)
  {
    inputs->throttle = 1.0;
  }
  else if (throttle < 0.05)
  {
    inputs->throttle = 0.0;
  } 
  else
  {
    inputs->throttle = throttle;
  }
}


void T_TX_InterfaceRCTRAN::getRawData(float *dest)
{
  int axes = getNumAxes();
  
  for (int i = 0; i < axes; i++)
  {
    *(dest + i) = (float)rctran[i];
  }
}


int T_TX_InterfaceRCTRAN::getNumAxes()
{
  return 4;
}
