#include "tx_interface_serpic.h"

#if SERPIC_USES_INTERFACESERIAL < 1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

#ifndef WIN32
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#endif

#include "../lib_conversions.h"
#include "../global.h"

#ifdef WIN32
#define DEFAULT_PORT_NAME   "\\\\.\\COM1"
#define SLEEP Sleep
#else
#define DEFAULT_PORT_NAME   "/dev/ttyS0"
#define INVALID_HANDLE_VALUE (-1)
// serial port handling, all unixoid systems
#define serial_read   read
#define serial_close  close
#define SLEEP(_msec)  usleep(1000 * (_msec))
#endif

#define DEFAULT_BAUDRATE        (19200)
#define DEFAULT_SYNC_BYTE       (0xFF)
#define DEFAULT_BUTTON_CHANNEL  (0)

//#define DUMMY_SERIAL

#if DEBUG_TX_INTERFACE > 0
#define DBGBUFFERSIZE   (256)
unsigned char dbgbuffer[DBGBUFFERSIZE];
int dbgbufferidx = 0;
#endif

namespace {

#ifdef WIN32
// serial port handling, Win32 systems

/** \brief Wrapper function for ReadFile()
 *
 *  This wrapper makes ReadFile() look like the Unix read()
 */
size_t serial_read(FHANDLE d, void *buf, size_t nbytes)
{
  DWORD done;
  ReadFile(d, buf, nbytes, &done, NULL);
  return done;
}

/** \brief Wrapper function for CloseHandle()
 *
 *  This wrapper makes CloseHandle() look like the Unix close().
 *  The return value has to be inverted (CloseHandle() returns
 *  != 0 on success).
 */
bool  serial_close(FHANDLE h)
{
  bool ret = CloseHandle(h);
  return !ret;
}
#endif


/** \brief Get last error as std::string
 *
 *  Converts an error code retrieved with GetLastError() to a
 *  plain text message and formats it as std::string.
 *
 *  \return string with last error message
 */
#ifdef WIN32
std::string SystemErrorMsg()
{
  std::string s;
  LPVOID  lpMsgBuf;
  DWORD   lastError = GetLastError();
  FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | 
                FORMAT_MESSAGE_FROM_SYSTEM | 
                FORMAT_MESSAGE_IGNORE_INSERTS,
                NULL,
                lastError,
                MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
                (LPTSTR) &lpMsgBuf,
                0,
                NULL);
  s = "(";
  s += itoStr(lastError, ' ', 0);
  s += ") ";
  s += (const char *)lpMsgBuf;
  
  LocalFree(lpMsgBuf);
  return s;
}
#endif


/** \brief Print the content of a DCB structure
 *
 *  Prints the content of a device control block (DCB) structure
 *  to the specified file handle. An additional information string
 *  can be included.
 *  \param fp file handle to which the information will be printed
 *  \param dp pointer to the DCB structure
 *  \param when additional information
 */
#ifdef WIN32
void PrintDCB(FILE *fp, DCB* dp, const char *when)
{
  fprintf(fp, "\nContents of DCB %s:\n", when);
  fprintf(fp, "BaudRate:    %d\n", dp->BaudRate);
  fprintf(fp, "ByteSize:    %d\n", dp->ByteSize);

  fprintf(fp, "Parity:      ");
  switch (dp->Parity)
  {
    case EVENPARITY:
      fprintf(fp, "even\n");
      break;

    case ODDPARITY:
      fprintf(fp, "odd\n");
      break;

    case MARKPARITY:
      fprintf(fp, "mark\n");
      break;

    case NOPARITY:
      fprintf(fp, "none\n");
      break;

    case SPACEPARITY:
      fprintf(fp, "space\n");
      break;

    default:
      fprintf(fp, "unclear: %d\n", dp->Parity);
      break;
  }

  fprintf(fp, "StopBits:    ");
  switch (dp->StopBits)
  {
    case ONESTOPBIT:
      fprintf(fp, "1\n");
      break;

    case ONE5STOPBITS:
      fprintf(fp, "1.5\n");
      break;

    case TWOSTOPBITS:
      fprintf(fp, "2\n");
      break;

    default:
      fprintf(fp, "unclear: %d\n", dp->StopBits);
      break;
  }

  fprintf(fp, "fDtrControl: ");
  switch(dp->fDtrControl)
  {
    case DTR_CONTROL_ENABLE:
      fprintf(fp, "enabled\n");
      break;

    case DTR_CONTROL_DISABLE:
      fprintf(fp, "disabled\n");
      break;

    case DTR_CONTROL_HANDSHAKE:
      fprintf(fp, "handshake\n");
      break;

    default:
      fprintf(fp, "unclear: %d\n", dp->fDtrControl);
      break;
  }
    
  fprintf(fp, "fRtsControl: ");
  switch(dp->fRtsControl)
  {
    case RTS_CONTROL_ENABLE:
      fprintf(fp, "enabled\n");
      break;

    case RTS_CONTROL_DISABLE:
      fprintf(fp, "disabled\n");
      break;

    case RTS_CONTROL_HANDSHAKE:
      fprintf(fp, "handshake\n");
      break;

    case RTS_CONTROL_TOGGLE:
      fprintf(fp, "toggle\n");
      break;

    default:
      fprintf(fp, "unclear: %d\n", dp->fDtrControl);
      break;
  }

  fprintf(fp, "XonLim / XoffLim: %d / %d\n", dp->XonLim, dp->XoffLim);
  fprintf(fp, "Chars: Xon Xoff Error Eof Evt: %02X %02X %02X %02X %02X\n",
              dp->XonChar, dp->XoffChar, dp->ErrorChar, dp->EofChar, dp->EvtChar);

  fprintf(fp, "fBinary fParity fNull fDsrSensitivity: %d %d %d %d\n", 
              dp->fBinary, dp->fParity, dp->fNull, dp->fDsrSensitivity);
  fprintf(fp, "fOutxCtsFlow fOutxDsrFlow fOutX fInX:  %d %d %d %d\n",
              dp->fOutxCtsFlow, dp->fOutxDsrFlow, dp->fOutX, dp->fInX);

  fprintf(fp, "fTXContinueOnXoff fAbortOnError: %d %d\n",
          dp->fTXContinueOnXoff, dp->fAbortOnError);
  fprintf(fp, "\n");
}
#endif

} // namespace


/** \brief The constructor.
 *
 *
 */
T_TX_InterfaceSerPIC::T_TX_InterfaceSerPIC()
  : iSPIC_fd(INVALID_HANDLE_VALUE),
    iSPIC_NumChannels(0), iSPIC_ButtonChannel(DEFAULT_BUTTON_CHANNEL),
    iSPIC_ChanCount(0), ucSyncByte(DEFAULT_SYNC_BYTE),
    sSerialPortExt(""), sSerialPortInt(""), uiBaudRate(DEFAULT_BAUDRATE),
    iSyncState(WAIT_SYNC), iSyncCounter(0)
{
#if DEBUG_TX_INTERFACE > 0
  printf("T_TX_InterfaceSerPIC::T_TX_InterfaceSerPIC()\n");
  printf("T_TX_InterfaceSerPIC state machine set to WAIT_SYNC\n");
#endif

  mixer = new T_TX_Mixer(this);
  map = new T_AxisMapper(this);
  calib = new T_Calibration(this);
  
  for (int i = 0; i < TX_MAXAXIS; i++)
  {
    afSPIC_RawData[i] = 0;
  }
}


/** \brief The destructor.
 *
 *
 */
T_TX_InterfaceSerPIC::~T_TX_InterfaceSerPIC()
{
#if DEBUG_TX_INTERFACE > 0
  printf("T_TX_InterfaceSerPIC::~T_TX_InterfaceSerPIC()\n");
#endif
  SPIC_Exit();
  delete mixer;
  delete map;
  delete calib;
}


int T_TX_InterfaceSerPIC::SPIC_InitSerial_TryList()
{
  int nRetCode = SPIC_InitSerial();
  
  if (nRetCode != 0)
  {
    int id = 0;
    while (serialDevs[id] != NULL && nRetCode != 0)
    {
      #if DEBUG_TX_INTERFACE > 0
      printf("SPIC_InitSerial_TryList: Trying %s\n", serialDevs[id]);
      #endif
      setDeviceName(serialDevs[id]);
      nRetCode = SPIC_InitSerial();
      id++;
    }
  }
  
  return(nRetCode);
}

/**
 *  Open the serial port and set the correct port modes,
 *  including the state of DTR and RTS.
 */
int T_TX_InterfaceSerPIC::SPIC_InitSerial()
{
  int iError = -1;

#if DEBUG_TX_INTERFACE > 0
  printf("Initializing serial PIC interface at %s\n", sSerialPortInt.c_str());
  printf("Speed %d, sync byte 0x%02X\n", uiBaudRate, ucSyncByte);
#endif

#ifdef WIN32
  DCB dcb;
  COMMTIMEOUTS  ct;
  BOOL fSuccess;

  iError = 0;
  iSPIC_fd = CreateFile( sSerialPortInt.c_str(),
                         GENERIC_READ /* | GENERIC_WRITE */,
                         0,             // comm devices must be opened w/exclusive-access
                         NULL,          // no security attributes
                         OPEN_EXISTING, // comm devices must use OPEN_EXISTING
                         0,             // overlapped I/O
                         NULL           // hTemplate must be NULL for comm devices
                         );

  if (iSPIC_fd == INVALID_HANDLE_VALUE)
  {
    // Handle the error.
    errMsg = "FMSPIC: Cannot open ";
    errMsg += sSerialPortInt + " : ";
    errMsg += SystemErrorMsg();
    fprintf (stderr, "*** %s\n", errMsg.c_str());
    iError = -1;
  }
  else
  {
    // We will build on the current configuration, and skip setting the size
    // of the input and output buffers with SetupComm.
    fSuccess = GetCommState(iSPIC_fd, &dcb);
  
    if (!fSuccess)
    {
      // Handle the error.
      errMsg = "FMSPIC: Error reading settings of ";
      errMsg += sSerialPortInt + " : ";
      errMsg += SystemErrorMsg();
      fprintf (stderr, "*** %s\n", errMsg.c_str());
      iError = -1;
    }
    else
    {
      #if DEBUG_TX_INTERFACE > 0
      PrintDCB(stdout, &dcb, "before configuring the port");
      #endif
      // save port configuration
      memcpy(&dcb_saved, &dcb, sizeof(DCB));
      
      // Fill in the DCB: 8 data bits, no parity, and 1 stop bit.
      // set the baudrate
      switch (uiBaudRate)
      {
        case 9600:
          dcb.BaudRate = CBR_9600;
          break;
    
        case 19200:
        default:
          dcb.BaudRate = CBR_19200;
          break;
      }
      dcb.ByteSize    = 8;                    // data size, xmit, and rcv
      dcb.Parity      = NOPARITY;             // no parity bit
      dcb.StopBits    = ONESTOPBIT;           // one stop bit
      dcb.fDtrControl = DTR_CONTROL_DISABLE;  // set DTR state
      dcb.fRtsControl = RTS_CONTROL_ENABLE;   // set RTS state
    
      fSuccess = SetCommState(iSPIC_fd, &dcb);

      if (!fSuccess)
      {
        // Handle the error.
        errMsg = "FMSPIC: Cannot setup ";
        errMsg += sSerialPortInt + " : ";
        errMsg += SystemErrorMsg();
        fprintf (stderr, "*** %s\n", errMsg.c_str());
        iError = -1;
      }
      else
      {
        #if DEBUG_TX_INTERFACE > 0
        PrintDCB(stdout, &dcb, "after configuring the port");
        #endif
        if (GetCommTimeouts(iSPIC_fd, &ct) == FALSE)
        {
          errMsg = "FMSPIC: Cannot read timeout settings of ";
          errMsg += sSerialPortInt + " : ";
          errMsg += SystemErrorMsg();
          fprintf (stderr, "*** %s\n", errMsg.c_str());
          iError = -1;
        }
        else
        {
          memcpy(&ct_saved, &ct, sizeof(COMMTIMEOUTS));
          ct.ReadIntervalTimeout = MAXDWORD;
          ct.ReadTotalTimeoutConstant = 0;
          ct.ReadTotalTimeoutMultiplier = 0;
          if (SetCommTimeouts(iSPIC_fd, &ct) == FALSE)
          {
            errMsg = "FMSPIC: Cannot set timeout settings of ";
            errMsg += sSerialPortInt + " : ";
            errMsg += SystemErrorMsg();
            fprintf (stderr, "*** %s\n", errMsg.c_str());
            iError = -1;
          }
        }
      }
    }
  }

#elif defined linux

  struct termios termios_p;
  int stat_;
  
  /* Try to open the serial port. */
  iError = 0;
    
# ifdef DUMMY_SERIAL
  iSPIC_fd = open("serialfifo", O_RDONLY | O_NDELAY);
# else
  iSPIC_fd = open(sSerialPortInt.c_str(), O_RDONLY | O_NOCTTY | O_NDELAY);
# endif
  
  if (iSPIC_fd == INVALID_HANDLE_VALUE)
  {
    errMsg = "FMSPIC: Cannot open ";
    errMsg += sSerialPortInt + " : ";
    errMsg += strerror(errno);
    fprintf (stderr, "*** %s\n", errMsg.c_str());
    iError = -1;
  }
  
# ifndef DUMMY_SERIAL
  if (iSPIC_fd != INVALID_HANDLE_VALUE)
  {
    /* save old settings */
    tcgetattr(iSPIC_fd, &termios_saved);
    memcpy(&termios_p, &termios_saved, sizeof(struct termios));
    
    /* successfully opened the port, setup the transfer mode */
    switch (uiBaudRate)
    {
      case 9600:
        termios_p.c_cflag = B9600;
        break;
  
      case 19200:
      default:
        termios_p.c_cflag = B19200;
        break;
    }
    termios_p.c_cflag |= CS8;     /* 8-bit-mode */
    termios_p.c_cflag |= CREAD;   /* enable receiver */
    termios_p.c_cflag |= CLOCAL;  /* no flow control */

    termios_p.c_iflag = IGNPAR | IGNBRK;
    termios_p.c_oflag = 0;
    termios_p.c_lflag = 0;
    termios_p.c_cc[VTIME] = 0;
    termios_p.c_cc[VMIN] = 1;

    tcsetattr(iSPIC_fd, TCSANOW, &termios_p);
    tcflush(iSPIC_fd, TCOFLUSH);
    tcflush(iSPIC_fd, TCIFLUSH);

    /* Set RTS signal, clear DTR signal */
    if(ioctl(iSPIC_fd, TIOCMGET, &stat_) == -1)
    {
      errMsg = "FMSPIC: Error reading settings of ";
      errMsg += sSerialPortInt + " : ";
      errMsg += strerror(errno);
      fprintf (stderr, "*** %s\n", errMsg.c_str());
      iError = -1;
    }
    else
    {
      stat_ &= ~TIOCM_DTR;
      stat_ |= TIOCM_RTS;
      if(ioctl(iSPIC_fd, TIOCMSET, &stat_) == -1)
      {
        errMsg = "FMSPIC: Cannot setup ";
        errMsg += sSerialPortInt + " : ";
        errMsg += strerror(errno);
        fprintf (stderr, "*** %s\n", errMsg.c_str());
        iError = -1;
      }
    }
  }
# endif // DUMMY_SERIAL
  
#endif

  return iError;
} /* end of SPIC_InitSerial() */


/** 
 *  Clean up: close the serial port and restore the original port settings.
 */
void T_TX_InterfaceSerPIC::SPIC_Exit(void)
{
  if (iSPIC_fd != INVALID_HANDLE_VALUE)
  {
    /* restore port settings */
    #ifndef WIN32
    tcsetattr(iSPIC_fd, TCSANOW, &termios_saved);
    #else
    SetCommTimeouts(iSPIC_fd, &ct_saved);
    SetCommState(iSPIC_fd, &dcb_saved);
    #endif
    
    /* close port */
    if(serial_close(iSPIC_fd) != 0)
    {
      fprintf(stderr, "*** T_TX_InterfaceSerPIC: Error closing port : %s\n", strerror(errno));
    }
    iSPIC_fd = INVALID_HANDLE_VALUE;
  }
} /* end of SPIC_Exit() */


/** 
 *  Wait for sync byte and reset channel counter.
 */
void T_TX_InterfaceSerPIC::SPIC_Sync(void)
{
  int iNumBytes;
  unsigned char ucByte;
  static int chancount;

  switch (iSyncState)
  {
    default:
    case WAIT_SYNC:
      // detect some sync bytes to make sure the interface
      // sends a valid stream of bytes, not any startup trash
      ucByte = 0;
      iNumBytes = serial_read(iSPIC_fd, &ucByte, 1);

      if (iNumBytes > 0)
      {
        #if DEBUG_TX_INTERFACE > 0
        if (dbgbufferidx < DBGBUFFERSIZE)
        {
          dbgbuffer[dbgbufferidx++] = ucByte;
        }
        else if (dbgbufferidx == DBGBUFFERSIZE)
        {
          dbgbufferidx++;
          for (int i = 0; i < DBGBUFFERSIZE; i++)
          {
            if (i % 32 == 0)
            {
              printf("\n");
            }
            printf("%02X ", dbgbuffer[i]);
          }
          printf("\n");
        }
        #endif
        if (ucByte >= ucSyncByte)
        {
          iSyncCounter++;
        }
        if (iSyncCounter > 4)
        {
          iSyncState = COUNT_CHANNELS;
          iSPIC_NumChannels = 0;
          chancount = 0;
          #if DEBUG_TX_INTERFACE > 0
          printf("T_TX_InterfaceSerPIC: detected sync bytes: 0x%02X\n", ucByte);
          printf("T_TX_InterfaceSerPIC state machine set to COUNT_CHANNELS\n");
          #endif
        }
      }
      break;
      

    case COUNT_CHANNELS:
      iSyncCounter = 0;

      ucByte = 0;
      iNumBytes = serial_read(iSPIC_fd, &ucByte, 1);
      if (iNumBytes > 0)
      {
        #if DEBUG_TX_INTERFACE > 0
        if (dbgbufferidx < DBGBUFFERSIZE)
        {
          dbgbuffer[dbgbufferidx++] = ucByte;
        }
        else if (dbgbufferidx == DBGBUFFERSIZE)
        {
          dbgbufferidx++;
          for (int i = 0; i < DBGBUFFERSIZE; i++)
          {
            if (i % 32 == 0)
            {
              printf("\n");
            }
            printf("%02X ", dbgbuffer[i]);
          }
          printf("\n");
        }
        #endif
        if (ucByte < ucSyncByte)
        {
          chancount++;
        }
        else
        {
          // received next sync byte, end of sync
          iSyncState = IN_SYNC;
          int tmp = (chancount - iSPIC_ButtonChannel);
          #if DEBUG_TX_INTERFACE > 0
          printf("T_TX_InterfaceSerPIC: detected %d channels\n", tmp);
          printf("T_TX_InterfaceSerPIC state machine set to IN_SYNC\n");
          #endif
          if (tmp > TX_MAXAXIS)
          {
            fprintf(stdout, 
                    "T_TX_InterfaceSerPIC: This version of T_TX_InterfaceSerPIC only \n            supports %d channels, will only use 0-7!\n",
                    TX_MAXAXIS);
            tmp = TX_MAXAXIS;
            //~ iSyncState = TRANSMITTER_ERROR;
          }
          iSPIC_NumChannels = tmp;
        }
      }
      break;
      
    case IN_SYNC:
    case TRANSMITTER_ERROR:
      break;
  }
} /* end of SPIC_Sync() */


/**
 *  Read data and store the "raw" byte in the array "channel". 
 */
void T_TX_InterfaceSerPIC::getInputData(TSimInputs *inputs)
{
  int iNumBytes;
  unsigned char ucByte;
  float *dest;
  
  if (iSyncState == IN_SYNC)
  {
    dest = static_cast<float*>(static_cast<void*>(inputs));
    
    do
    {
      iNumBytes = serial_read(iSPIC_fd, &ucByte, 1);
      if (iNumBytes > 0)
      {
        #if DEBUG_TX_INTERFACE > 0
        if (dbgbufferidx < DBGBUFFERSIZE)
        {
          dbgbuffer[dbgbufferidx++] = ucByte;
        }
        else if (dbgbufferidx == DBGBUFFERSIZE)
        {
          dbgbufferidx++;
          for (int i = 0; i < DBGBUFFERSIZE; i++)
          {
            if (i % 32 == 0)
            {
              printf("\n");
            }
            printf("%02X ", dbgbuffer[i]);
          }
          printf("\n");
        }
        #endif
        /* received one byte */
        if (ucByte >= ucSyncByte)
        {
          /* received SYNC byte */
          iSPIC_ChanCount = 0;
        }
        else
        {
          int current_channel = (iSPIC_ChanCount - iSPIC_ButtonChannel);
          if ((current_channel >= 0) && (current_channel < TX_MAXAXIS))
          {
            afSPIC_RawData[current_channel] = 
                    (float)((float)ucByte/175.0 - 0.5);
          }
          iSPIC_ChanCount++;
        }
      }
    } while (iNumBytes > 0);
    
    float calibrated;
    int   axisnum;
    
    axisnum    = map->func[ELEVATOR];
    if (axisnum >= 0)
    {
      calibrated = calib->calibrate(axisnum, afSPIC_RawData[axisnum])
                    * map->inv[ELEVATOR];
      inputs->elevator = mixer->mix_elevator(calibrated);
    }
        
    axisnum    = map->func[AILERON];
    if (axisnum >= 0)
    {
      calibrated = calib->calibrate(axisnum, afSPIC_RawData[axisnum])
                    * map->inv[AILERON];
      inputs->aileron = mixer->mix_aileron(calibrated);
    }
    
    axisnum    = map->func[RUDDER];
    if (axisnum >= 0)
    {
      calibrated = calib->calibrate(axisnum, afSPIC_RawData[axisnum])
                    * map->inv[RUDDER];
      inputs->rudder = mixer->mix_rudder(calibrated);
    }

    axisnum    = map->func[THROTTLE];
    if (axisnum >= 0)
    {
      calibrated = calib->calibrate(axisnum, afSPIC_RawData[axisnum])
                    * map->inv[THROTTLE];
      inputs->throttle = mixer->mix_throttle(calibrated + 0.5);
    }
  }
  else
  {
    // not in sync yet...
    SPIC_Sync();
  }
} /* end of SPIC_GetRawData() */


/** \brief Start the interface.
 *
 *  The complete initialization sequence rolled into one function.
 *  Initializes the serial port and synchronizes with the interface.
 *  \param config Pointer to XML config file tree.
 *  \return 0 on success, -1 on error
 */
int T_TX_InterfaceSerPIC::init(SimpleXMLTransfer *config)
{
  int iReturn = 0;
  
  SimpleXMLTransfer *port = config->getChild("inputMethod.serpic.port", true);
  
  sSerialPortInt = port->getString("name", DEFAULT_PORT_NAME);
#ifdef WIN32
  sSerialPortExt  = std::string("COM") + sSerialPortInt[7] + std::string(":");
#else
  sSerialPortExt = sSerialPortInt;
#endif
    
  uiBaudRate = port->getInt("baudrate", DEFAULT_BAUDRATE);
  ucSyncByte = port->getInt("sync", DEFAULT_SYNC_BYTE);
  iSPIC_ButtonChannel = port->getInt("button_channel", DEFAULT_BUTTON_CHANNEL);
  
  if (mixer->init(config, "inputMethod.serpic") != 0)
  {
    iReturn = -1;
  }
  else if (SPIC_InitSerial_TryList() != 0)
  {
    iReturn = -1;
  }
  else
  {
    calib->init(config, "inputMethod.serpic");
    map->init(config, "inputMethod.serpic");
  }

  return iReturn;
}


/**
 *  Store all settings.
 */
void T_TX_InterfaceSerPIC::putBackIntoCfg(SimpleXMLTransfer* config)
{
#if DEBUG_TX_INTERFACE > 0
  printf("int T_TX_InterfaceSerPIC::putBackIntoCfg(SimpleXMLTransfer* config)\n");
#endif
  T_TX_Interface::putBackIntoCfg(config);

  SimpleXMLTransfer *port = config->getChild("inputMethod.serpic.port", true);
  port->setAttributeOverwrite("name", sSerialPortInt);
  port->setAttributeOverwrite("baudrate", uiBaudRate);
  port->setAttributeOverwrite("sync", ucSyncByte);
  port->setAttributeOverwrite("button_channel", iSPIC_ButtonChannel);

  if (iSPIC_NumChannels > 0)
  {
    // only if a valid interface was detected!
    mixer->putBackIntoCfg(config);
    map->putBackIntoCfg(config);
    calib->putBackIntoCfg(config);
  }
}


void T_TX_InterfaceSerPIC::getRawData(float *dest)
{
  for (int i = 0; i < iSPIC_NumChannels; i++)
  {
    *(dest + i) = afSPIC_RawData[i];
  }
}

void T_TX_InterfaceSerPIC::setDeviceName(std::string devname)
{
  sSerialPortExt = devname;
  
  #ifdef WIN32
  sSerialPortInt  = "\\\\.\\COM";
  sSerialPortInt += devname[3];
  #else
  sSerialPortInt = devname;
  #endif
  
  #if DEBUG_TX_INTERFACE > 0
  printf("T_TX_InterfaceSerPIC::setDeviceName: external name: %s, internal name: %s\n", sSerialPortExt.c_str(), sSerialPortInt.c_str());
  #endif
}

void T_TX_InterfaceSerPIC::setDeviceSpeed(int speed)
{
  switch (speed)
  {
    case 0:
      uiBaudRate = 9600;
      ucSyncByte = 0xF0;
      iSPIC_ButtonChannel = 1;
      break;

    case 1:
      uiBaudRate = 19200;
      ucSyncByte = 0xFF;
      iSPIC_ButtonChannel = 0;
      break;

    default:
      uiBaudRate = DEFAULT_BAUDRATE;
      ucSyncByte = DEFAULT_SYNC_BYTE;
      iSPIC_ButtonChannel = DEFAULT_BUTTON_CHANNEL;
      break;
  }
    
}

int T_TX_InterfaceSerPIC::getDeviceSpeed()
{
  if (uiBaudRate == 9600)
  {
    return 0;
  }
  else
  {
    return 1;
  }
}

int T_TX_InterfaceSerPIC::getDeviceNumber(const char* list[])
{
  int id = 0;
  
  while (list[id] != NULL)
  {
    if (sSerialPortExt.compare(list[id]) == 0)
      return(id);
    
    id++;
  }
  
  return(-1);
}

// ---------------------------------------------------------------------------

#else // SERPIC_USES_INTERFACESERIAL

#define DEFAULT_SYNC_BYTE       (0xFF)
#define DEFAULT_BUTTON_CHANNEL  (0)


/**
 *  The constructor.
 *
 */
T_TX_InterfaceSerPIC::T_TX_InterfaceSerPIC()
  : T_TX_InterfaceSerial(),
    iSyncState(WAIT_SYNC), iSyncCounter(0),
    iSPIC_NumChannels(0), iSPIC_ButtonChannel(DEFAULT_BUTTON_CHANNEL),
    iSPIC_ChanCount(0), ucSyncByte(DEFAULT_SYNC_BYTE)
{
#if DEBUG_TX_INTERFACE > 0
	std::cout << "T_TX_InterfaceSerPIC::T_TX_InterfaceSerPIC ()\n";
#endif
}


/**
 *  The destructor.
 *
 */
T_TX_InterfaceSerPIC::~T_TX_InterfaceSerPIC()
{
#if DEBUG_TX_INTERFACE > 0
	std::cout << "T_TX_InterfaceSerPIC::~T_TX_InterfaceSerPIC ()\n";
#endif
}


/**
 *  Initialize the interface.
 *
 *  The base class handles all hardware initialization for us, so we only
 *  have to set up the correct control line states to power the interface.
 */
int T_TX_InterfaceSerPIC::init (SimpleXMLTransfer *config)
{
	int ret=T_TX_InterfaceSerial::init (config);

	setRts (true);
	setDtr (false);

	return ret;
}


/**
 *  Process the next data byte.
 *
 *  This method processes the next byte that was received on the serial
 *  port.
 *
 *  \param byte   The byte to be processed.
 */
void T_TX_InterfaceSerPIC::processDataByte (unsigned char ucByte)
{
  static int chancount;

  switch (iSyncState)
  {
    default:
    case WAIT_SYNC:
      // detect some sync bytes to make sure the interface
      // sends a valid stream of bytes, not any startup trash
      if (ucByte >= ucSyncByte)
      {
        iSyncCounter++;
      }
      if (iSyncCounter > 4)
      {
        iSyncState = COUNT_CHANNELS;
        iSPIC_NumChannels = 0;
        chancount = 0;
        #if DEBUG_TX_INTERFACE > 0
        printf("T_TX_InterfaceSerPIC: detected sync bytes: 0x%02X\n", ucByte);
        printf("T_TX_InterfaceSerPIC state machine set to COUNT_CHANNELS\n");
        #endif
      }
      break;
      

    case COUNT_CHANNELS:
      iSyncCounter = 0;
      if (ucByte < ucSyncByte)
      {
        chancount++;
      }
      else
      {
        // received next sync byte, end of sync
        iSyncState = IN_SYNC;
        int tmp = (chancount - iSPIC_ButtonChannel);
        #if DEBUG_TX_INTERFACE > 0
        printf("T_TX_InterfaceSerPIC: detected %d channels\n", tmp);
        printf("T_TX_InterfaceSerPIC state machine set to IN_SYNC\n");
        #endif
        if (tmp > TX_MAXAXIS)
        {
          fprintf(stdout, 
                  "T_TX_InterfaceSerPIC: This version of T_TX_InterfaceSerPIC only \n            supports %d channels, will only use 0-7!\n",
                  TX_MAXAXIS);
          tmp = TX_MAXAXIS;
          //~ iSyncState = TRANSMITTER_ERROR;
        }
        iSPIC_NumChannels = tmp;
      }
      break;
      
    case IN_SYNC:
      if (ucByte >= ucSyncByte)
      {
        /* received SYNC byte */
        iSPIC_ChanCount = 0;
      }
      else
      {
        int current_channel = (iSPIC_ChanCount - iSPIC_ButtonChannel);
        if ((current_channel >= 0) && (current_channel < TX_MAXAXIS))
        {
          setRawData(iSPIC_ChanCount, (float)((float)ucByte/175.0 - 0.5));
        }
        iSPIC_ChanCount++;
      }
      break;

    case TRANSMITTER_ERROR:
      break;
  }
}


/**
 *  Get the device speed.
 *  
 *  This method overrides T_TX_InterfaceSerial::getDeviceSpeed().
 *  The reason for this is that T_TX_InterfaceSerPIC only supports
 *  two speeds, 9600 baud and 19200 baud. Therefore the GUI only
 *  offers these two values. If the translation from the speed value to the
 *  GUI's combo box item is done in the GUI, the GUI needs to know too much
 *  about this interface, so it's much cleaner to do the translation here.
 */
int T_TX_InterfaceSerPIC::getDeviceSpeed()
{
  int speed;

  // speed as reported by the base class
  int speed_serial = T_TX_InterfaceSerial::getDeviceSpeed();
  
  if (speed_serial == 6)
  {
    // 9600 baud
    speed = 0;
  }
  else
  {
    // 19200 baud, default
    speed = 1;
  }
  return speed;
}

/**
 *  Set the device speed.
 *
 *  This method overrides T_TX_InterfaceSerial::setDeviceSpeed(int).
 *  The reason for this is that T_TX_InterfaceSerPIC only supports
 *  two speeds, 9600 baud and 19200 baud. Therefore the GUI only
 *  offers these two values. If the translation from the speed value to the
 *  GUI's combo box item is done in the GUI, the GUI needs to know too much
 *  about this interface, so it's much cleaner to do the translation here.
 */
void T_TX_InterfaceSerPIC::setDeviceSpeed(int speed)
{
  int speed_serial;
  
  if (speed == 0)
  {
    // 9600 baud
    speed_serial = 6;
    ucSyncByte = 0xF0;
    iSPIC_ButtonChannel = 1;
  }
  else
  {
    // 19200 baud, default
    speed_serial = 7;
    ucSyncByte = 0xFF;
    iSPIC_ButtonChannel = 0;
  }
  T_TX_InterfaceSerial::setDeviceSpeed(speed_serial);
}


#endif // SERPIC_USES_INTERFACESERIAL

