#ifndef TX_INTERFACE_SERPIC
#define TX_INTERFACE_SERPIC

/* 
 * 0: use old implementation with integrated serial port handling
 * 1: use new implementation with common base class T_TX_InterfaceSerial
 */
#define SERPIC_USES_INTERFACESERIAL 1

#if SERPIC_USES_INTERFACESERIAL < 1
# include "../tx_interface.h"
# include "../SimpleXMLTransfer.h"

#ifdef WIN32
#include <windows.h>
typedef HANDLE FHANDLE;
#else
#include <termios.h>
typedef int FHANDLE;
#endif

#define READ_BUFFER_SIZE  (256)   /* max. line length for config files  */
#define SYNC_TIMEOUT      (500)   /* try to sync for approx. 5 seconds  */



/** \brief The serial PIC interface.
 *  
 *  This interface handles a transmitter connected to a
 *  serial port via an FMS-style "PIC" cable.
 *
 *  Interface hardware:
 *
 *  There are several schematics floating around the WWW.
 *  A good example can be found on
 *  http://myweb.absa.co.za/eric.brouwer/interface.htm
 *
 *  The interfaces are designed to work with the FMS r/c
 *  simulator which is unfortunately a Windows-only project
 *  (http://fms.pathbot.com). This simulator supports three
 *  flavours of interfaces: parallel, simple serial and
 *  PIC serial. The simple serial interface only works with
 *  OSes which allow direct polling of the serial ports,
 *  like DOS or Win98. The PIC serial interface is designed
 *  to work with any OS by independently transforming the
 *  r/c transmitter's PPM waveform into simple 8-bit values.
 *  Those values are then sent via the serial port to the
 *  computer. The start of a transmission is marked by a
 *  unique sync byte. After that "header" byte, each byte
 *  represents the value of one channel.
 *
 *  The serial transmission runs without flow
 *  control. To activate the interface's power supply, RTS
 *  has to be set while DTR must be cleared.
 *
 *  There are two speed variants of this interface:
 *  - 9600 baud, sync byte 0xF0 + channel count (0xF1 ... 0xF8)
 *  - 19200 baud, sync byte 0xFF
 *
 *  Some interface designs will allow you to switch between
 *  both modes. CRRCsim supports both, so choose whatever
 *  you like.
 */
class T_TX_InterfaceSerPIC : public T_TX_Interface
{
  public:
    T_TX_InterfaceSerPIC();
    ~T_TX_InterfaceSerPIC();
  
   /**
    * Initialize interface. Read further config data from a file, if necessary.
    * \param config Pointer to configuration file's XML tree
    * \return error code (0: no error)
    */
    int init(SimpleXMLTransfer* config);

   /**
    * Get current input data. If some value is not available, the value 
    * is not overwritten.
    * \param inputs Pointer to channel value structure.
    */
    void getInputData(TSimInputs* inputs);

   /**
    * Write configuration back
    * \param config Pointer to configuration file's XML tree
    */
    void putBackIntoCfg(SimpleXMLTransfer* config);

   /**
    * Get input method
    * \return the value <code>T_TX_Interface::eIM_SerPIC</code>
    */
    int inputMethod() { return(T_TX_Interface::eIM_serpic); };
    
   /** \brief Get number of axis.
    * 
    *  Returns the number of detected axis.
    */
    int getNumAxes()  { return iSPIC_NumChannels; };
    
   /** \brief Get raw interface data as float values.
    *
    *  Fills the memory pointed to by <code>dest</code> with
    *  raw data values in float format
    *  \param dest Memory area to be filled with raw data. Make
    *              sure it can hold getNumAxes() values!
    */
    void getRawData(float *dest);

   /**
    * Return the device number of the interface.
    * \return 0 for ttyS0/COM1, 1 for ttyS1/COM2 and so on
    */
   int getDeviceNumber(const char* list[]);
   
   /**
    * Return the speed of the interface.
    * \return 0 for 9600 baud, 1 for 19200 baud
    */
   int getDeviceSpeed();

   /**
    * Set the device number of the interface.
    *
    * \param device name of the serial port (ttyS0, ttyS1, ttyUSB0, COM1, COM2, ...)
    */
   void setDeviceName(std::string devname);

   /**
    * Set the speed of the interface.
    *
    * \param speed Speed of the serial port (0 = 9600 baud, 1 = 19200 baud)
    */
   void setDeviceSpeed(int speed);

  private:
    FHANDLE iSPIC_fd;             ///< file descriptor for serial port
    //~ int iSPIC_sync;               ///< received a sync byte yet?
    int iSPIC_NumChannels;        ///< How many channels are transmitted?
    int iSPIC_ButtonChannel;      ///< Does channel 0 contain button states?
    int iSPIC_ChanCount;          ///< which channel is to be received next?
    unsigned char ucSyncByte;     ///< Which value is used as SYNC marker?
    float afSPIC_RawData[TX_MAXAXIS];
  
    #ifndef WIN32
    struct  termios termios_saved;
    #else
    DCB           dcb_saved;
    COMMTIMEOUTS  ct_saved;
    #endif
    
    std::string   sSerialPortExt;
    std::string   sSerialPortInt;
    unsigned int  uiBaudRate;
    int           iSyncState;
    int           iSyncCounter;

    int   SPIC_InitSerial();
    int   SPIC_InitSerial_TryList();
    void  SPIC_Exit();
    void  SPIC_Sync();

    enum {WAIT_SYNC = 0, COUNT_CHANNELS, IN_SYNC, TRANSMITTER_ERROR};
};

#else // SERPIC_USES_INTERFACESERIAL

#include "../T_TX_InterfaceSerial.h"

class T_TX_InterfaceSerPIC: public T_TX_InterfaceSerial
{
	public:
		T_TX_InterfaceSerPIC();
		~T_TX_InterfaceSerPIC();

		virtual int init (SimpleXMLTransfer *config);
		virtual std::string getInterfaceName () { return "FMSPIC"; }
		virtual void processDataByte (unsigned char ucByte);
		int inputMethod () { return T_TX_Interface::eIM_serpic; }
		int getNumAxes() { return iSPIC_NumChannels; };
		virtual int   getDeviceSpeed ();
		virtual void  setDeviceSpeed (int speed);
  
	private:
    int iSyncState;
    int iSyncCounter;
    int iSPIC_NumChannels;        ///< How many channels are transmitted?
    int iSPIC_ButtonChannel;      ///< Does channel 0 contain button states?
    int iSPIC_ChanCount;          ///< which channel is to be received next?
    unsigned char ucSyncByte;     ///< Which value is used as SYNC marker?

    enum {WAIT_SYNC = 0, COUNT_CHANNELS, IN_SYNC, TRANSMITTER_ERROR};
};

#endif // SERPIC_USES_INTERFACESERIAL

#endif    // TX_INTERFACE_SERPIC
