#include "tx_software.h"

#include "../lib_conversions.h"
#include "../crrc_main_cpp.h"

T_TX_InterfaceSoftware::T_TX_InterfaceSoftware(int method)
  : keyb_aileron_input(0.0), keyb_rudder_input(0.0), 
    keyb_elevator_input(0.0), keyb_throttle_input(-0.5),
    input_method(method)
{
#if DEBUG_TX_INTERFACE > 0
  printf("T_TX_InterfaceSoftware::T_TX_InterfaceSoftware(int method)\n");
#endif

  mixer = new T_TX_Mixer(this);

  switch (input_method)
  {
    case eIM_joystick:
      map = new T_AxisMapper(this);
      calib = new T_Calibration(this);
      break;
    
    case eIM_mouse:
      map = new T_AxisMapper(this);
      break;
    
    default:
      break;
  }
  
  for (int i = 0; i < TX_MAXAXIS; i++)
  {
    input_raw_data[i] = 0.0;
  }
}

T_TX_InterfaceSoftware::~T_TX_InterfaceSoftware()
{
#if DEBUG_TX_INTERFACE > 0
  printf("T_TX_InterfaceSoftware::~T_TX_InterfaceSoftware()\n");
#endif
  delete map;
  delete calib;
  delete mixer;
}

int T_TX_InterfaceSoftware::init(SimpleXMLTransfer* config)
{
#if DEBUG_TX_INTERFACE > 0
  printf("int T_TX_InterfaceSoftware::init(SimpleXMLTransfer* config)\n");
#endif

  int ret = 0;

  // initialize generic stuff
  T_TX_Interface::init(config);

  // initialize mixer
  switch (input_method)
  {
    case eIM_mouse:
      ret = mixer->init(config, "inputMethod.mouse");
      map->init(config, "inputMethod.mouse");
      break;

    case eIM_joystick:
      ret = mixer->init(config, "inputMethod.joystick");
      map->init(config, "inputMethod.joystick");
      calib->init(config, "inputMethod.joystick");
      break;

    case eIM_keyboard:
    default:
      ret = mixer->init(config, "inputMethod.keyboard");
      break;
  }

  return ret;
}

void T_TX_InterfaceSoftware::putBackIntoCfg(SimpleXMLTransfer* config)
{
#if DEBUG_TX_INTERFACE > 0
  printf("int T_TX_InterfaceSoftware::putBackIntoCfg(SimpleXMLTransfer* config)\n");
#endif
  T_TX_Interface::putBackIntoCfg(config);

  if (usesMixer())
  {
    mixer->putBackIntoCfg(config);
  }
  if (usesMapper())
  {
    map->putBackIntoCfg(config);
  }
  if (usesCalibration())
  {
    calib->putBackIntoCfg(config);
  }
}

void T_TX_InterfaceSoftware::getInputData(TSimInputs* inputs)
{
#if DEBUG_TX_INTERFACE > 1
  printf("void T_TX_InterfaceSoftware::getInputData(TSimInputs* inputs)\n");
#endif
  //~ inputs->aileron = aileron_input;
  //~ inputs->rudder  = rudder_input;
  //~ inputs->elevator= elevator_input;
  //~ inputs->throttle= throttle_input;

  float calibrated;
  int   axisnum;

  if (input_method == eIM_keyboard)
  {
    inputs->elevator = mixer->mix_elevator(keyb_elevator_input);
    inputs->aileron  = mixer->mix_aileron(keyb_aileron_input);
    inputs->rudder   = mixer->mix_rudder(keyb_rudder_input);
    inputs->throttle = mixer->mix_throttle(keyb_throttle_input + 0.5);
  }
  else
  {
    if (usesCalibration())
    {
      axisnum    = map->func[ELEVATOR];
      if (axisnum >= 0)
      {
        calibrated = calib->calibrate(axisnum, input_raw_data[axisnum])
                      * map->inv[ELEVATOR];
        inputs->elevator = mixer->mix_elevator(calibrated);
      }
      
      axisnum    = map->func[AILERON];
      if (axisnum >= 0)
      {
        calibrated = calib->calibrate(axisnum, input_raw_data[axisnum])
                      * map->inv[AILERON];
        inputs->aileron = mixer->mix_aileron(calibrated);
      }
      
      axisnum    = map->func[RUDDER];
      if (axisnum >= 0)
      {
        calibrated = calib->calibrate(axisnum, input_raw_data[axisnum])
                      * map->inv[RUDDER];
        inputs->rudder = mixer->mix_rudder(calibrated);
      }
    
      axisnum    = map->func[THROTTLE];
      if (axisnum >= 0)
      {
        calibrated = calib->calibrate(axisnum, input_raw_data[axisnum])
                      * map->inv[THROTTLE];
        inputs->throttle = mixer->mix_throttle(calibrated + 0.5);
      }
      else
      {
        inputs->throttle = mixer->mix_throttle(keyb_throttle_input + 0.5);
      }
    }
    else
    {
      axisnum    = map->func[ELEVATOR];
      if (axisnum >= 0)
      {
        calibrated = input_raw_data[axisnum] * map->inv[ELEVATOR];
        inputs->elevator = mixer->mix_elevator(calibrated);
      }
      
      axisnum    = map->func[AILERON];
      if (axisnum >= 0)
      {
        calibrated = input_raw_data[axisnum] * map->inv[AILERON];
        inputs->aileron = mixer->mix_aileron(calibrated);
      }
      
      axisnum    = map->func[RUDDER];
      if (axisnum >= 0)
      {
        calibrated = input_raw_data[axisnum] * map->inv[RUDDER];
        inputs->rudder = mixer->mix_rudder(calibrated);
      }
      
      axisnum    = map->func[THROTTLE];
      if (axisnum >= 0)
      {
        calibrated = input_raw_data[axisnum] * map->inv[THROTTLE];
        inputs->throttle = mixer->mix_throttle(calibrated + 0.5);
      }
      else
      {
        inputs->throttle = mixer->mix_throttle(keyb_throttle_input + 0.5);
      }
    }
  }
}


void T_TX_InterfaceSoftware::increase_throttle()
{
  float tmp = keyb_throttle_input + 0.1;
  keyb_throttle_input = limit(tmp);
}

void T_TX_InterfaceSoftware::decrease_throttle()
{
  float tmp = keyb_throttle_input - 0.1;
  keyb_throttle_input = limit(tmp);
}

void T_TX_InterfaceSoftware::move_aileron(const float x)
{
  float tmp = keyb_aileron_input + x;
  keyb_aileron_input = limit(tmp);
}

void T_TX_InterfaceSoftware::move_rudder(const float x)
{
  float tmp = keyb_rudder_input + x;
  keyb_rudder_input = limit(tmp);
}

void T_TX_InterfaceSoftware::move_elevator(const float x)
{
  float tmp = keyb_elevator_input + x;
  keyb_elevator_input = limit(tmp);
}

int T_TX_InterfaceSoftware::getNumAxes()
{
  int axes = 0;
  
  switch (input_method)
  {
    case eIM_mouse:
      axes = 2;
      break;
    
    case eIM_joystick:
      axes = inputDev.getJoystickNumAxes();
      break;
    
    case eIM_keyboard:
      axes = 4;
      break;
    
    default:
      break;
  }
  
  return axes;
}

void T_TX_InterfaceSoftware::getRawData(float *dest)
{
  int axes = getNumAxes();
  
  for (int i = 0; i < axes; i++)
  {
    *(dest + i) = input_raw_data[i];
  }
}

void T_TX_InterfaceSoftware::setAxis(int axis, const float x)
{
  input_raw_data[axis] = limit(x);
}

void T_TX_InterfaceSoftware::centerControls()
{
  keyb_elevator_input = 0.0;
  keyb_aileron_input = 0.0;
  keyb_rudder_input = 0.0;
}
