#ifndef INTGR_H
# define INTGR_H

# include "vector3.h"

/** \brief this namespace contains pure math
 *
 * [1] Bronstein/Semendjajew, 'Taschenbuch der Mathematik', B. G. Teubner Verlagsgesellschaft.
 * 
 * [2] Friedrich Weller, 'Numerische Mathematik fr Ingenieure und Naturwissenschaftler', Vieweg Verlag.
 * 
 * jwtodo: welches Integrationsverfahren? Wie anwenden?
 * 
 */
namespace CRRCMath
{

  /**
   * Adams-Bashforth algorithm
   * 
   * Einfachste Version des Adams-Bashforth (lineares explizites Zweischrittverfahren nach 
   * der Methode von Adams-Bashforth).
   * 
   * Herleitung nach [2].
   * 
   * \f[
       y(x_{i+1}) - y(x_i) = \int_{x_i}^{x_{i+1}} f(x, y(x)) \, dx
     \f]
   * Die Ableitung im Punkt \f$(x,y)\f$ ist \f$f(x,y)\f$.
   * 
   * \f[
       f_i = f(x_i, y_i)
     \f]
   * 
   * \f[
       f_{i-1} = f(x_{i-1}, y_{i-1})
     \f]
   * 
   * Die Ableitung wird linear interpoliert/extrapoliert:
   * \f[
       f(x) = f_i + \frac{f_i - f_{i-1}}{h} \, (x - x_i)
     \f]
   * 
   * mit \f$ h = x_{i+1} - x_i = x_i - x_{i-1}\f$. Der obige Ausdruck fr \f$f(x)\f$ wird in die erste 
   * Gleichung eingesetzt und umgeformt. Man erhlt dann
   * 
   * \f[
       y(x_{i+1}) = y(x_i) + \frac{h}{2} \, (3\, f_i - f_{i-1})
     \f]
   *
   * @author Jens Wilhelm Wulf
   */
  template<class C> class Integrationsverfahren /*{{{*/
  {
    public:
     
     /**
      * virtual base class should have a virtual dtor
      */
     virtual ~Integrationsverfahren() {};

      /**
      * Festlegung von Startwert und Ableitung zum Startzeitpunkt
      */
     virtual void init(C Startwert, C AblStartwert);

     /**
      * Ausfhrung eines Integrationsschritts.
      */
     virtual void step(double dT, C AblNeu);

     /**
      * der integrierte Wert
      */
     C val;

    private:
     C AblAlt;
  };

  /**
   *
   * Trapezmethode, Verfahren von Heun. Abschnitt 9.3.3 in [2]
   * 
   * jwtodo
   * 
   * Zumindest lassen die Kommentare im originalen CRRCSim darauf schlieen. Wenn ich aber versuche, 
   * das nachzuvollziehen, komme ich (bisher) noch nicht zu dem Schluss, dass diese Implementierung 
   * dieses Verfahren korrekt abbildet. Es ist einfach etwas anderes.
   * 
   * @author Jens Wilhelm Wulf
   */
  template<class C> class IntegrationsverfahrenB /*{{{*/
  {
    public:
     
     /**
      * virtual base class should have a virtual dtor
      */
     virtual ~IntegrationsverfahrenB() {};

     /**
      * Festlegung von Startwert und Ableitung zum Startzeitpunkt
      */
     virtual void init(C Startwert, C AblStartwert);

     /**
      * Ausfhrung eines Integrationsschritts.
      */
     virtual void step(double dT, C AblNeu);

     /**
      * der integrierte Wert
      */
     C val;

    private:
     C AblAlt;
  };

  /**
   * Polygonzugverfahren von Euler
   * 
   * Abschnitt 7.1.2.9 in [1]
   * 
   * Abschnitt 9.3.2   in [2]
   * 
   * @author Jens Wilhelm Wulf
   */
  template<class C> class IntegrationsverfahrenC /*{{{*/
  {
    public:
     
     /**
      * virtual base class should have a virtual dtor
      */
     virtual ~IntegrationsverfahrenC() {};

     /**
      * Festlegung von Startwert und Ableitung zum Startzeitpunkt
      */
     virtual void init(C Startwert, C AblStartwert);

     /**
      * Ausfhrung eines Integrationsschritts.
      */
     virtual void step(double dT, C AblNeu);

     /**
      * der integrierte Wert
      */
     C val;
  };
  
};
#endif
