#include "battery.h"

#include <iostream>
#include <fstream>
#include <cstdlib>
#include "../config.h"
#include "../lib_conversions.h"

Power::Battery::Battery(SimpleXMLTransfer* xml)
{
  SimpleXMLTransfer* bat;
  bool               fExtern = true;

  throttle_min = xml->getDouble("throttle_min");
  throttle_old = 0;
  
  if (xml->indexOfAttribute("filename") >= 0)  
    bat = new SimpleXMLTransfer(T_Config::getDataPath("models/battery/" + xml->getString("filename") + ".xml", true));
  else
  {
    bat     = xml;
    fExtern = false;
  }

  nUOffStatus = 0; 
  C_0   = bat->getDouble("C")*60*60;
  R_I   = bat->getDouble("R_I");
  U_off = bat->getDouble("U_off");
  
  std::cout << "Battery: C=" << C_0/(60*60) << " Ah, R_I=" << R_I << " Ohm, U_off=" << U_off << "V\n";
  
  // voltage over capacity
  {
    double                 U_0 = bat->getDouble("U_0");
    std::string            tab = bat->getChild("U_0rel")->getContentString();
    std::string::size_type pos;
    std::string::size_type start = 0;
    double                 U;
    
    while ( (pos = tab.find(';', start)) != std::string::npos && start < tab.length())
    {
      U = U_0*atof(tab.substr(start, pos-start).c_str());
      std::cout << "         " << U << " V\n";
      voltage.push_back(U);
      start = pos+1;
    }
    
    // Calculate factor for easy interpolation:
    // There are size values:
    int size = voltage.size();
    // So I want to have an index in the range 0..(size-2). 
    // C_0 has to be converted to (size-1)*2^10.
    dInterpFact = ((size-1)<<10)/C_0;
  }
  U = voltage[0];  
  
  C.init(C_0, 0);
  
  if (fExtern)
    delete bat;
    
  SimpleXMLTransfer* shaft;
    
  for (int n=0; n<xml->getChildCount(); n++)
  {
    shaft = xml->getChildAt(n);
    if (shaft->getName().compare("shaft") == 0)
    {
      Shaft *s;
      s = new Shaft(shaft);
      shafts.push_back(s);
    }
  }
}


Power::Battery::~Battery()
{
  // deallocate shafts
  for (unsigned int i = 0; i < shafts.size(); i++)
  {
    delete shafts[i];
    shafts[i] = NULL;
  }
}


void Power::Battery::step(PowerValuesStep* values)
{
  int          size = shafts.size();
  int          idx;
  double       thr;

  thr = values->inputs->throttle;
  
  if (thr < throttle_min && throttle_old > 0)
  {
    thr                      = throttle_min;
    values->inputs->throttle = throttle_min;
  }
  throttle_old = thr;
  
  if (U < U_off || C.val <= 0 || nUOffStatus == 1)
  {
    if (nUOffStatus == 1 && thr < 0.05)
      nUOffStatus = 0;
    else
      nUOffStatus = 1;
    values->U = 0;
  }
  else  
    values->U = U;
  
  values->I = 0;  
  
  for (int n=0; n<size; n++)
  {
    shafts[n]->step(values);
  }
  
  C.step(values->dt, -1*values->I);
  if (C.val < 0)
    C.init(0, 0);  
  
  {
    double dCapLeftRel = C.val/C_0;
    
    if (values->dBatCapLeftMin > dCapLeftRel)
      values->dBatCapLeftMin = dCapLeftRel;
  }
  
  // size of table:
  size = voltage.size();
  // max. index for interpolation:
  int idxmax = ((size-1)<<10)-1;
  // calc index
  idx  = (int)(dInterpFact * (C_0-C.val));
  if (idx > idxmax)
    idx = idxmax;
  // interpolation
  int idxh = idx >> 10;
  int idxl = idx & ((1<<10)-1);
  
  U = voltage[idxh] + (idxl * (voltage[idxh+1]-voltage[idxh]))*(1.0/1024);
  
//  std::cout << idxh << " " << U << " V, " << voltage[idxh] << " V, " << values->I  << " A\n";
  
  // voltage drop because of internal resistance
  U -= R_I * values->I;
}

void Power::Battery::showCapacity()
{
  std::cout << C.val/(60*60) << " Ah\n";
}

void Power::Battery::automagic(SimpleXMLTransfer* xml)
{
  Shaft* s;
  s = new Shaft();
  s->automagic(xml);
  shafts.push_back(s);

  nUOffStatus = 0;
  R_I   = 0; 
  U_off = 1;
  U     = xml->getDouble("battery.shaft.engine.automagic.U");
  C_0   = xml->getDouble("battery.shaft.engine.automagic.I")*xml->getDouble("battery.automagic.T");
  xml->setAttribute("battery.C", doubleToString(C_0/60/60));
  xml->setAttribute("battery.U_0",   doubleToString(U));
  xml->setAttribute("battery.U_off", doubleToString(U_off));
  xml->setAttribute("battery.R_I",   doubleToString(R_I));

  throttle_min = xml->getDouble("battery.throttle_min");
  throttle_old = 0;
  
  SimpleXMLTransfer* g = new SimpleXMLTransfer();
  g->setName("U_0rel");
  g->setContent("\n   1.00;\n   1.00;\n");
  xml->getChild("battery")->addChild(g);
  voltage.push_back(U);
  voltage.push_back(U);

  // Calculate factor for easy interpolation:
  // There are size values:
  int size = voltage.size();
  // So I want to have an index in the range 0..(size-2). 
  // C_0 has to be converted to (size-1)*2^10.
  dInterpFact = ((size-1)<<10)/C_0;

  C.init(C_0, 0);    
}

void Power::Battery::reset()
{
  C.init(C_0, 0);
  throttle_old = 0;
}
