#ifndef GEARING_H
# define GEARING_H

# include "../SimpleXMLTransfer.h"
# include "values_step.h"

namespace Power
{

  /**
   * This class is part of the power system. To simply use the system, you should not
   * access or call any of its members. Please take a look at Power instead.
   *
   * This class models a gearing which connects engines/propellers to a shaft. Everything
   * connected to a shaft must be a subclass of it.
   * 
   * The xml description of a gearing must be written inside of the element which is connected to 
   * the shaft by the gearbox. It looks like this:
   \verbatim
     <gearing i="1.13" J="0" eta="0.98"/>
   \endverbatim
   * A gearbox might have an inertia \c J which is not zero. This inertia is the value seen by the shaft.
   * The inertia of the propeller/engine is translated to the shaft automatically.
   * 
   * Given \c omega is the speed of the shaft, <tt>i*omega</tt> is the speed of the device which is connected to the shaft
   * using this gearing. The efficiency of the gearing is \c eta.
   *
   * @author Jens Wilhelm Wulf
   */
  class Gearing
  {
    public:

     /**
      * only used for automagic construction
      */
     Gearing();
     
     /**
      * Read configuration from xml.
      * @param xml
      */
     Gearing(SimpleXMLTransfer* xml);
    
     /**
      * virtual base class should have a virtual dtor
      */
     virtual ~Gearing() {};

     /**
      * Read parameters for automagic construction from xml.
      */
     virtual void automagic(SimpleXMLTransfer* xml) {};

     /**
      * Returns the inertia of the gearing and the device
      * attached to it as seen by the shaft [kg m^2].
      */
     double getJ();

     /**
      * Go ahead values->dt seconds in the simulation.
      * The device, which is attached to a Shaft by this Gearing, adds its
      * forces/torque/current/... to the ones in PowerValuesStep.
      */
     virtual void step(PowerValuesStep* values) = 0;

    protected:

     /**
      * Inertia of the device attached to the gearing [Nm s = kg m^2]
      */
     double J;

     /**
      * Gear ratio. The propeller/engine rotates at a rate of i * Shaft::omega.
      */
     double i;

     /**
      * Wirkungsgrad
      */
     double eta;
     
    private:

     /**
      * Inertia of the gearing as seen by the shaft [Nm s = kg m^2]
      */
     double J_G;
  };
};
#endif
