#ifndef POWER_H
# define POWER_H

# include <vector>
# include "../SimpleXMLTransfer.h"
# include "../math/vector3.h"
# include "battery.h"

/** \brief power/propulsion system
 * 
 * The power/propulsion system is packed together in a namespace. Users of the system should only 
 * use Power::Power, which does manage everything.
 */
namespace Power
{
  
  /**
   *
   * This class manages everything related to battery, engine, propeller.
   * Throughout this class (and the classes used by it) the SI unit system
   * is used (with a small number of exceptions).
   * 
   * Values used are:
   *   - lengths \c H and \c D in m
   *   - rotational velocity \c omega in rad/s
   *   - rotational velocity \c n in 1/s
   *   - linear velocity \c V in m/s
   *   - Torque \c M in Nm
   *   - Force/thrust \c F in N
   *   - Inertia \c J in kg m^2
   *   - Resistance \c R in Ohm
   *   - Current \c I in A
   *   - Voltage \c U in V
   *   - Capacity \c C in Ah
   * 
   * One airplane needs exactly one Power system, even if the latter is empty. 
   * The Power system for one airplane can contain the following items:
   *   - Power::Battery \n
   *     At least one is needed, unless you have a glider.
   *   - Power::Shaft \n
   *     At least one is needed.
   *     One or more shafts can be connected to a battery.
   *     Every engine/propeller is connected to a shaft.
   *   - Power::Engine_DCM \n
   *     One or more engines can be connected to a shaft.
   *   - Power::Propeller or Power::SimpleThrust \n
   *     One or more of them can be connected to a shaft.
   *
   * The system is configured using a description in xml. Here's an example of a system with one
   * battery; there is a gearbox with two engines driving one prop.
   \verbatim
   <power>
     <battery filename="nicd12_30" throttle_min="0">
       <shaft J="2E-6" brake="0">
         <engine filename="astro_cobalt_10">
           <gearing eta="0.95" J="0" i="2.5" />
         </engine>
         <engine filename="astro_cobalt_10">
           <gearing eta="0.95" J="0" i="2.5" />
         </engine>
         <propeller D="0.3" H="0.2" J="3E-6" n_fold="-1" downthrust="2" />
       </shaft>
     </battery>    
   </power>
   \endverbatim
   * 
   * Take a look at the individual items for details of their description.
   * 
   * If you don't want to define every single item of the system, you can take the easy way.
   * Use a description like this:
   \verbatim
     <automagic F="12" V="15.8">
      <battery throttle_min="0">
        <automagic T="420" />
        <shaft J="0" brake="1">
          <propeller D="0.243" H="0.17" J="0" n_fold="5" downthrust="2" />
          <engine>
            <automagic omega_p="2827" eta_opt="0.78" eta="0.7" />
          </engine>
        </shaft>
      </battery>
    </automagic>
   \endverbatim
   * 
   * This will create a power system which delivers a thrust of <tt>F=12 N</tt> at a velocity of <tt>V=15.8 m/s</tt>.
   * The battery will be designed to last <tt>T=420 s</tt>.
   * You also need to fill in the dimensions of the propeller. It is best to take everything else as shown in the example.
   * 
   * If you use this \e automagic way of creating the system, the program will output the configuration it calculated from your
   * values to give you a starting point for fine grained tunings.
   * 
   * @author Jens Wilhelm Wulf
   */
  class Power
  {
    public:

     /**
      * Creates an empty power system
      */
     Power();

     /**
      * Tries to load attributes of the power system. This is the only valid
      * way of creating it!
      * Throws an exception on error.
      *
      * @param xml Contains the configuration options of the system. See ./documentation/power.htm
      */
     Power(SimpleXMLTransfer* xml);
     
     /**
      * Deallocates a power system.
      */
     ~Power();
    
     /**
      * Go ahead dt seconds in the simulation. Calls Battery::step() for each battery in the list.
      * Note that force and moment must be initialized (to zero or something else) before calling this!
      *
      * @param dt     timestep [s]
      * @param inputs pointer to inputs
      * @param V_X    Velocity of airplane body relative to airmass, X-axis [m/s].
      * @param force  Force created by power system is added to this vector [N] (body axes).
      * @param moment Torque created by power system is added to this vector [Nm] (body axes).
      */
     void step(double                dt,
               TSimInputs*           inputs,
               double                V_X,
               CRRCMath::Vector3*    force,
               CRRCMath::Vector3*    moment);
     
     /**
      * Returns revolutions per second of a propeller. If there is more than one,
      * the one of the last in the list is used [1/s].
      */
     double getPropFreq() { return(dPropFreq); };
     
     /**
      * Returns lowest relative battery capacity left (0..1).
      */
     double getBatteryMin() { return(dBatCapLeftMin); };
     
     /**
      * Resets battery status
      */
     void reset();
     
    private:
     
     /**
      * List of batteries in use.
      */
     std::vector<Battery*> batteries;
     
     /**
      * Rotational speed of the last propeller calculated [1/s].
      */
     double dPropFreq;
     
     /**
      * Lowest relative battery capacity left (0..1).
      */
     double dBatCapLeftMin;
  };
  
};

#endif
