#ifndef SIMPLETHRUST_H
# define SIMPLETHRUST_H

# include "gearing.h"
# include "values_step.h"

namespace Power
{

  /**
   * This class is part of the power system. To simply use the system, you should not
   * access or call any of its members. Please take a look at Power instead.
   *
   * This is a very simple propeller to model CRRCSims first
   * engine: thrust, no matter at which airspeed.
   * 
   * This simple propeller is connected to a shaft via some gearing. 
   * The gearing is described by
   *   - \c i \n
   *     gear ratio
   *   - \c eta \n
   *     efficiency of gearing
   *   
   * Given \c omega is the speed of the shaft this simple propeller is connected to,
   * it will rotate at <tt>(omega_p = i * omega)</tt> and provide a thrust of <tt>(F = k_F * omega_p)</tt>.
   * 
   * The torque applied to the shaft is <tt>(M = -1 * k_M * omega_p * i / eta)</tt>.
   * 
   * Example for an xml description for direct connection to shaft:
   * 
   \verbatim
   <simplethrust k_F="0.004" k_M="0.00001" />
   \endverbatim
   * 
   * Example for an xml description, connection to shaft via a gearbox:
   * 
   \verbatim
   <simplethrust k_F="0.005" k_M="0.00002" >
     <gearing i="1.13" J="0" eta="0.98"/>
   </simplethrust>
   \endverbatim
   * 
   * See Power::Gearing for a description of a gearbox.
   *
   * @author Jens Wilhelm Wulf
   */
  class SimpleThrust : public Gearing
  {
    public:
     
     /**
      * Read configuration from xml.
      * @param xml
      */
     SimpleThrust(SimpleXMLTransfer* xml);

     /**
      * virtual base class should have a virtual dtor
      */
     virtual ~SimpleThrust() {};

     /**
      * Go ahead values->dt seconds in the simulation.
      */
     virtual void step(PowerValuesStep* values);

    private:

     /**
      * Thrust provided is (k_F * i * Shaft::omega)
      */
     double k_F;

     /**
      * Torque applied to shaft is (-1 * k_M * i * Shaft::omega * i / eta)
      */
     double k_M;
  };
};
#endif
