#include "tx_interface.h"

#include "lib_conversions.h"

#include <math.h>

// 1/exp(exp(1)), needed for mix_exp
#ifndef M_1_E_E
# define M_1_E_E 0.065988
#endif

// exp(1), needed for mix_exp
#ifndef M_E
# define M_E 2.7182818284590452354
#endif

const char* ButtonStrings[] =
{
  "NOTHING", "RESUME", "PAUSE", "RESET", "INCTHROTTLE",
    "DECTHROTTLE", "ZOOMIN", "ZOOMOUT"
};

const char* AxisStrings[]   = {"NOTHING", "AILERON", "ELEVATOR", "RUDDER", "THROTTLE"};


/**
 * Strings ordered like radio type enums in T_TX_Interface
 */
const char* RadioTypeStrings[] = {"Futaba", "Airtronics", "Hitec", "JR", "Cockpit", "Walkera", "Custom"};

/**
 * Strings ordered like input method enums in T_TX_Interface
 */
const char* InputMethodStrings[] = {"Keyboard", "Mouse",   "Joystick",
                                    "RCTran", "Audio",
                                    "Parallel", "Serial2", "RCTran2",
                                    "FMSPIC", "MNAV", "ZhenHua" };


int T_TX_Interface::init(SimpleXMLTransfer* config)
{
#if DEBUG_TX_INTERFACE > 0
  printf("int T_TX_Interface::init(SimpleXMLTransfer* config)\n");
#endif

  return(0);
}

std::string T_TX_Interface::getErrMsg() 
{
  return(errMsg);
};

void T_TX_Interface::getInputData(TSimInputs* inputs)
{
#if DEBUG_TX_INTERFACE > 1
  printf("void T_TX_Interface::getInputData(TSimInputs* inputs)\n");
#endif
}

T_TX_Interface::T_TX_Interface()
  : mixer(NULL), calib(NULL), map(NULL), errMsg("")
{
#if DEBUG_TX_INTERFACE > 0
  printf("T_TX_Interface::T_TX_Interface\n");
#endif
}

T_TX_Interface::~T_TX_Interface()
{
#if DEBUG_TX_INTERFACE > 0
  printf("T_TX_Interface::~T_TX_Interface\n");
#endif
}

void T_TX_Interface::putBackIntoCfg(SimpleXMLTransfer* config)
{
}

float T_TX_Interface::limit(float in)
{
  if (in < -0.5)
  {
    return -0.5;
  }
  else if (in > 0.5)
  {
    return 0.5;
  }
  else
  {
    return in;
  }
}


// --- Implementation of class T_TX_Mixer -------------------------------------

/** \brief Create a mixer object
 *
 *  The mixer will be initialized with default values: scaling of all
 *  axes is set to 1.0, all other parameters are set to 0.0.
 */
T_TX_Mixer::T_TX_Mixer(T_TX_Interface *parent) 
    : 
      trim_aileron(0.0),  trim_elevator(0.0),  trim_rudder(0.0),  trim_throttle(0.0),
      nrate_aileron(0.0), nrate_elevator(0.0), nrate_rudder(0.0), nrate_throttle(0.0),
      exp_aileron(0.0),   exp_elevator(0.0),   exp_rudder(0.0),   exp_throttle(0.0),
      enabled(1), iface(parent)
{
#if DEBUG_TX_INTERFACE > 0
  printf("T_TX_Mixer::T_TX_Mixer()\n");
#endif
}


/** \brief Create and initialize a mixer object
 *
 *  The mixer will be initialized from the given child
 *  in the SimpleXMLTransfer. This does the same as if you
 *  would call <code>T_TX_Mixer::init(cfg, child)</code>
 *  after calling the empty constructor.
 *
 *  \param cfg Pointer to an XML tree containing the config info.
 *  \param child The name of the child which holds the mixer info.
 */
T_TX_Mixer::T_TX_Mixer(T_TX_Interface *parent, 
                       SimpleXMLTransfer* cfg, std::string child)
    : 
      trim_aileron(0.0),  trim_elevator(0.0),  trim_rudder(0.0),  trim_throttle(0.0),
      nrate_aileron(0.0), nrate_elevator(0.0), nrate_rudder(0.0), nrate_throttle(0.0),
      exp_aileron(0.0),   exp_elevator(0.0),   exp_rudder(0.0),   exp_throttle(0.0),
      enabled(1), iface(parent)
{
#if DEBUG_TX_INTERFACE > 0
  printf("T_TX_Mixer::T_TX_Mixer(cfg, child)\n");
#endif
  init(cfg, child);
}


/** \brief Destroy the mixer.
 *
 *  Right now the dtor is only provided for debugging purposes.
 */
T_TX_Mixer::~T_TX_Mixer()
{
#if DEBUG_TX_INTERFACE > 0
  printf("T_TX_Mixer::~T_TX_Mixer()\n");
#endif
}


/** \brief Initialize the mixer from a config file.
 *
 *  The mixer object will be initialized from the given config file.
 *  This file may contain more than one branch with interface settings,
 *  so the name of the child has to be specified.
 */
int T_TX_Mixer::init(SimpleXMLTransfer* cfg, std::string child)
{
#if DEBUG_TX_INTERFACE > 0
  printf("T_TX_Mixer::init(cfg, child)\n");
  printf(" <-- %s\n", child.c_str());
#endif

  SimpleXMLTransfer* mixer;
  SimpleXMLTransfer* item;
  SimpleXMLTransfer* inter;

  // store the child's name for writing back the settings
  child_in_cfg = child;

  try
  {
    inter = cfg->getChild(child, true);
    mixer = inter->getChild("mixer", true);

    enabled = mixer->attributeAsInt("enabled", 1);
  
    item = mixer->getChild("aileron", true);
    trim_aileron   = item->getDouble("trim",   0.0);
    exp_aileron    = item->getDouble("exp",    0.0);
    nrate_aileron  = item->getDouble("nrate",  1.1);

    item = mixer->getChild("elevator", true);
    trim_elevator   = item->getDouble("trim",   0.0);
    exp_elevator    = item->getDouble("exp",    0.0);
    nrate_elevator  = item->getDouble("nrate",  1.0);
    
    item = mixer->getChild("rudder", true);
    trim_rudder   = item->getDouble("trim",   0.0);
    exp_rudder    = item->getDouble("exp",    0.0);
    nrate_rudder  = item->getDouble("nrate",  1.0);
    
    item = mixer->getChild("throttle", true);
    trim_throttle   = item->getDouble("trim",   0.0);
    exp_throttle    = item->getDouble("exp",    0.0);
    nrate_throttle  = item->getDouble("nrate",  1.0);
  }
  catch (XMLException e)
  {
    errMsg = e.what();
    return 1;
  }

  return 0;
}


/** \brief Transfers all settings back to the config file.
 *
 *  The mixer settings will be stored in the same branch of the config
 *  file that they were read from on initialization.
 *
 *  \param config Pointer to the config file's SimpleXMLTransfer.
 */
void T_TX_Mixer::putBackIntoCfg(SimpleXMLTransfer* config)
{
#if DEBUG_TX_INTERFACE > 0
  printf("T_TX_Mixer::putBackIntoCfg(SimpleXMLTransfer* config)\n");
  printf(" --> %s\n", child_in_cfg.c_str());
#endif

  SimpleXMLTransfer* inter = config->getChild(child_in_cfg);
  SimpleXMLTransfer* mixer = inter->getChild("mixer");
  SimpleXMLTransfer* item;

  mixer->setAttributeOverwrite("enabled", enabled);

  item = mixer->getChild("aileron");
  item->setAttributeOverwrite("trim",   doubleToString(trim_aileron));
  item->setAttributeOverwrite("exp",    doubleToString(exp_aileron));
  item->setAttributeOverwrite("nrate",  doubleToString(nrate_aileron));
                                                  
  item = mixer->getChild("elevator");
  item->setAttributeOverwrite("trim",   doubleToString(trim_elevator));
  item->setAttributeOverwrite("exp",    doubleToString(exp_elevator));
  item->setAttributeOverwrite("nrate",  doubleToString(nrate_elevator));
                                                  
  item = mixer->getChild("rudder");
  item->setAttributeOverwrite("trim",   doubleToString(trim_rudder));
  item->setAttributeOverwrite("exp",    doubleToString(exp_rudder));
  item->setAttributeOverwrite("nrate",  doubleToString(nrate_rudder));
                                                  
  item = mixer->getChild("throttle");
  item->setAttributeOverwrite("trim",   doubleToString(trim_throttle));
  item->setAttributeOverwrite("exp",    doubleToString(exp_throttle));
  item->setAttributeOverwrite("nrate",  doubleToString(nrate_throttle));    
}


/** \brief Apply exponential component to input.
 *
 *  \param x Linear input signal (-1.0 ... 1.0).
 *  \param p Percentage of exponential component to be applied to input (0.0 ... 1.0).
 */
float T_TX_Mixer::mix_exp(const float x, const float p) const
{
  return x * (1.0 + p * (exp(fabs(x) * M_E) * M_1_E_E - 1.0));
}


/** \brief Mix the aileron channel.
 *
 *  Apply the mixer settings to the aileron channel.
 */
float T_TX_Mixer::mix_aileron(float in)
{
  if (enabled)
  {
    float tmp, exp_in;
    exp_in = 2.0 * in + trim_aileron;
    tmp = 0.5 * nrate_aileron * mix_exp(exp_in, exp_aileron);
    return tmp;
  }
  else
  {
    return in;
  }
}


/** \brief Mix the elevator channel.
 *
 *  Apply the mixer settings to the elevator channel.
 */
float T_TX_Mixer::mix_elevator(float in)
{
  if (enabled)
  {
    float tmp, exp_in;
    exp_in = 2.0 * in  + trim_elevator;
    tmp = 0.5 * nrate_elevator * mix_exp(exp_in, exp_elevator);
    return tmp;
  }
  else
  {
    return in;
  }

}


/** \brief Mix the rudder channel.
 *
 *  Apply the mixer settings to the rudder channel.
 */
float T_TX_Mixer::mix_rudder(float in)
{
  if (enabled)
  {
    float tmp, exp_in;
    exp_in = 2.0 * in + trim_rudder;
    tmp = 0.5 * nrate_rudder * mix_exp(exp_in, exp_rudder);
    return tmp;
  }
  else
  {
    return in;
  }
}


/** \brief Mix the throttle channel.
 *
 *  Apply the mixer settings to the throttle channel.
 */
float T_TX_Mixer::mix_throttle(float in)
{
  if (enabled)
  {
    float tmp, exp_in;
    exp_in =  in + trim_throttle;
    tmp = nrate_throttle * mix_exp(exp_in, exp_throttle);
    return tmp;
  }
  else
  {
    return in;
  }
}


/** \brief Get current error message
 *
 *  Returns the current error message (if any).
 */
std::string T_TX_Mixer::getErrMsg() 
{
  return(errMsg);
};


// --- Implementation of class T_AxisMapper -------------------------

T_AxisMapper::T_AxisMapper(T_TX_Interface *parent)
  : child_in_cfg(""), radio_type(0), iface(parent)
{
  for (int i = 0; i < NUM_AXISFUNCS; i++)
  {
    func[i]   = i;
    inv[i]    = 1;
    c_func[i] = i;
    c_inv[i]  = 1;
  }
}

T_AxisMapper::T_AxisMapper(T_TX_Interface *parent,
                           SimpleXMLTransfer* cfgfile,
                           std::string child)
  : child_in_cfg(""), radio_type(0), iface(parent)
{
  for (int i = 0; i < NUM_AXISFUNCS; i++)
  {
    func[i]   = i;
    inv[i]    = 1;
    c_func[i] = i;
    c_inv[i]  = 1;
  }
  init(cfgfile, child);
}

void T_AxisMapper::init(SimpleXMLTransfer* cfgfile, std::string childname)
{
#if DEBUG_TX_INTERFACE > 0
  printf("T_AxisMapper::init(cfg, child)\n");
  printf(" <-- %s\n", childname.c_str());
#endif
  SimpleXMLTransfer* inter;
  SimpleXMLTransfer* bindings;
  SimpleXMLTransfer* group;
  SimpleXMLTransfer* item;
  
  child_in_cfg = childname;

  // try to load config
  try
  {
    float elevator_default_polarity = -1.0;
    inter = cfgfile->getChild(childname, true);
    bindings  = inter->getChild("bindings", true);
    group     = bindings->getChild("axes", true);
  
    item = group->getChild("aileron", true);
    c_func[AILERON] = item->attributeAsInt("axis", 0);
    c_inv[AILERON]  = item->attributeAsDouble("polarity", 1.0);

    item = group->getChild("elevator", true);
    c_func[ELEVATOR] = item->attributeAsInt("axis", 1);

    if (iface->inputMethod() == T_TX_Interface::eIM_joystick)
    {
      elevator_default_polarity = 1.0;
    }
    c_inv[ELEVATOR]  = item->attributeAsDouble("polarity", elevator_default_polarity);

    item = group->getChild("rudder", true);
    c_func[RUDDER] = item->attributeAsInt("axis", -1);
    c_inv[RUDDER]  = item->attributeAsDouble("polarity", 1.0);

    item = group->getChild("throttle", true);
    c_func[THROTTLE] = item->attributeAsInt("axis", -1);
    c_inv[THROTTLE]  = item->attributeAsDouble("polarity", 1.0);

    std::string radio = 
      strU(bindings->attribute("radio_type", RadioTypeStrings[CUSTOM]));

    for (int n=0; n < NR_OF_RADIO_TYPES; n++)
    {
      if (radio.compare(strU(RadioTypeStrings[n])) == 0)
      {
        setRadioType(n);
      }
    }
  }
  catch (XMLException e)
  {
    fprintf(stderr, "*** T_AxisMapper: XMLException: %s\n", e.what());
  }
}


void T_AxisMapper::putBackIntoCfg(SimpleXMLTransfer* cfgfile)
{
#if DEBUG_TX_INTERFACE > 0
  printf("T_AxisMapper::putBackIntoCfg(SimpleXMLTransfer* config)\n");
  printf(" --> %s\n", child_in_cfg.c_str());
#endif
  SimpleXMLTransfer* item;
  SimpleXMLTransfer* group;
  SimpleXMLTransfer* item2;

  try
  {
    item = cfgfile->getChild(child_in_cfg);
    group = item->getChild("bindings.axes");

    item2 = group->getChild("elevator");
    item2->setAttributeOverwrite("axis", c_func[ELEVATOR]);
    item2->setAttributeOverwrite("polarity", doubleToString(c_inv[ELEVATOR]));
  
    item2 = group->getChild("aileron");
    item2->setAttributeOverwrite("axis", c_func[AILERON]);
    item2->setAttributeOverwrite("polarity", doubleToString(c_inv[AILERON]));
  
    item2 = group->getChild("rudder");
    item2->setAttributeOverwrite("axis", c_func[RUDDER]);
    item2->setAttributeOverwrite("polarity", doubleToString(c_inv[RUDDER]));
  
    item2 = group->getChild("throttle");
    item2->setAttributeOverwrite("axis", c_func[THROTTLE]);
    item2->setAttributeOverwrite("polarity", doubleToString(c_inv[THROTTLE]));
  
    item2 = item->getChild("bindings");
    item2->setAttributeOverwrite("radio_type", RadioTypeStrings[radio_type]);
  }
  catch (XMLException e)
  {
    fprintf(stderr, "*** T_AxisMapper: XMLException: %s\n", e.what());
  }
}


void T_AxisMapper::setRadioType(int rtype)
{
printf ("mapper set to radio type %d\n", rtype);
  int   saved[NUM_AXISFUNCS];
  float saved_i[NUM_AXISFUNCS];
  
  for (int i = 0; i < NUM_AXISFUNCS; i++)
  {
    saved[i] = func[i];
    saved_i[i] = inv[i];
    func[i] = i;
    inv[i] = 1.0;
  }

  switch (rtype)
  {
    case AIRTRONICS:
      func[ELEVATOR]  = 0;
      func[AILERON]   = 1;
      func[THROTTLE]  = 2;
      func[RUDDER]    = 3;
      radio_type = rtype;
      break;
    
    case JR:
      func[ELEVATOR]  = 2;
      func[AILERON]   = 1;
      func[THROTTLE]  = 0;
      func[RUDDER]    = 3;
      radio_type = rtype;
      break;
    
    case COCKPIT:
      func[ELEVATOR]  = 1;
      func[AILERON]   = 0;
      func[THROTTLE]  = 3;
      func[RUDDER]    = 2;
      radio_type = rtype;
      break;
    
    case FUTABA:
    case HITEC:
      func[ELEVATOR]  = 1;
      func[AILERON]   = 0;
      func[THROTTLE]  = 2;
      func[RUDDER]    = 3;
      radio_type = rtype;
      break;

	case WALKERA:
	  func[AILERON] =2; inv[AILERON] =-1;
	  func[ELEVATOR]=1; inv[ELEVATOR]=-1;
	  func[RUDDER]  =3; inv[RUDDER  ]= 1;
	  func[THROTTLE]=0; inv[THROTTLE]= 1;
      radio_type = rtype;
	  break;
    
    case CUSTOM:
      for (int i = 0; i < NUM_AXISFUNCS; i++)
      {
        func[i] = c_func[i];
        inv[i]  = c_inv[i];
      }
      radio_type = rtype;
      break;
    
    default:
      // don't change anything
      for (int i = 0; i < NUM_AXISFUNCS; i++)
      {
        func[i] = saved[i];
        inv[i]  = saved_i[i];
      }
      break;
  }
}

void T_AxisMapper::saveToCustom()
{
#if DEBUG_TX_INTERFACE > 0
  printf("T_AxisMapper::saveToCustom()\n");
#endif
  int numAxis = iface->getNumAxes();
  for (int i = 0; i < NUM_AXISFUNCS; i++)
  {
    // only store a value if it is in the valid range
    // for the current interface
    if (func[i] < numAxis)
    {
      c_func[i] = func[i];
      c_inv[i]  = inv[i];
    }
  }
}


// --- Implementation of class T_Calibration -------------------------
T_Calibration::T_Calibration(T_TX_Interface *parent)
  : child_in_cfg(""), iface(parent)
{
#if DEBUG_TX_INTERFACE > 0
  printf("T_Calibration::T_Calibration()\n");
#endif
  for (int i = 0; i < TX_MAXAXIS; i++)
  {
    scale[i] = 1.0;
    offset[i] = 0.0;
  }
}

T_Calibration::T_Calibration(T_TX_Interface *parent,
                             SimpleXMLTransfer* cfg,
                             std::string child)
  : child_in_cfg(child), iface(parent)
{
#if DEBUG_TX_INTERFACE > 0
  printf("T_Calibration::T_Calibration(cfg, child)\n");
#endif
  for (int i = 0; i < TX_MAXAXIS; i++)
  {
    scale[i] = 1.0;
    offset[i] = 0.0;
  }
  init(cfg, child);
}

void T_Calibration::init(SimpleXMLTransfer* cfgfile,
                         std::string childname)
{
#if DEBUG_TX_INTERFACE > 0
  printf("T_Calibration::init(cfg, child)\n");
  printf(" <-- %s\n", childname.c_str());
#endif
  int size;
  SimpleXMLTransfer* item;
  SimpleXMLTransfer* group;
  SimpleXMLTransfer* item2;
  
  child_in_cfg = childname;

  // try to load config
  try
  {
    item = cfgfile->getChild(childname, true);
    group = item->getChild("calibration", true);
  
    size  = group->getChildCount();
    if (size > TX_MAXAXIS)
      size = TX_MAXAXIS;
    for (int n = 0; n < size; n++)
    {
      item2 = group->getChildAt(n);
      scale[n]  = item2->getDouble("scale", 1.0);
      offset[n] = item2->getDouble("offset", 0.0);
    }
  }
  catch (XMLException e)
  {
    fprintf(stderr, "*** T_Calibration::init(): %s\n", e.what());
  }
}


void T_Calibration::putBackIntoCfg(SimpleXMLTransfer* cfgfile)
{
#if DEBUG_TX_INTERFACE > 0
  printf("T_Calibration::putBackIntoCfg(cfg)\n");
  printf(" --> %s\n", child_in_cfg.c_str());
#endif
  int size;
  SimpleXMLTransfer* item;
  SimpleXMLTransfer* group;
  SimpleXMLTransfer* item2;

  item = cfgfile->getChild(child_in_cfg);
  group = item->getChild("calibration");

  // clean list
  size = group->getChildCount();
  for (int n = 0; n < size; n++)
  {
    item2 = group->getChildAt(0);
    group->removeChildAt(0);
    delete item2;
  }
  // create new list
  for (int n = 0; n < TX_MAXAXIS; n++)
  {
    item2 = new SimpleXMLTransfer();
    item2->setName("axis");
    item2->addAttribute("scale", doubleToString(scale[n]));
    item2->addAttribute("offset", doubleToString(offset[n]));
    group->addChild(item2);
  }
}


float T_Calibration::calibrate(int axis, float raw)
{
  return (raw * scale[axis] + offset[axis]);
}
